@@ -27,6 +27,7 @@
 #include "Flow.h"
 #include "IpHelper.h"
 #include "Oid.h"
+#include "IpFragment.h"
 
 #ifdef OVS_DBG_MOD
 #undef OVS_DBG_MOD
@@ -142,7 +143,7 @@ OvsExtAttach(NDIS_HANDLE ndisFilterHandle,
     KeMemoryBarrier();
 
 cleanup:
-    gOvsInAttach = FALSE;
+    InterlockedExchange(&gOvsInAttach, 0);
     if (status != NDIS_STATUS_SUCCESS) {
         if (switchContext != NULL) {
             OvsDeleteSwitch(switchContext);
@@ -229,6 +230,12 @@ OvsCreateSwitch(NDIS_HANDLE ndisFilterHandle,
     if (status != STATUS_SUCCESS) {
         OvsUninitSwitchContext(switchContext);
         OVS_LOG_ERROR("Exit: Failed to initialize Connection tracking");
+    }
+
+    status = OvsInitIpFragment(switchContext);
+    if (status != STATUS_SUCCESS) {
+        OvsUninitSwitchContext(switchContext);
+        OVS_LOG_ERROR("Exit: Failed to initialize Ip Fragment");
         goto create_switch_done;
     }
 
@@ -246,7 +253,6 @@ create_switch_done:
  *  Implements filter driver's FilterDetach function.
  * --------------------------------------------------------------------------
  */
-_Use_decl_annotations_
 VOID
 OvsExtDetach(NDIS_HANDLE filterModuleContext)
 {
@@ -265,6 +271,8 @@ OvsExtDetach(NDIS_HANDLE filterModuleContext)
     OvsCleanupSttDefragmentation();
     OvsCleanupConntrack();
     OvsCleanupCtRelated();
+    OvsCleanupIpFragment();
+
     /* This completes the cleanup, and a new attach can be handled now. */
 
     OVS_LOG_TRACE("Exit: OvsDetach Successfully");
@@ -300,7 +308,6 @@ OvsDeleteSwitch(POVS_SWITCH_CONTEXT switchContext)
  *  Implements filter driver's FilterRestart function.
  * --------------------------------------------------------------------------
  */
-_Use_decl_annotations_
 NDIS_STATUS
 OvsExtRestart(NDIS_HANDLE filterModuleContext,
               PNDIS_FILTER_RESTART_PARAMETERS filterRestartParameters)
@@ -509,7 +516,7 @@ OvsReleaseSwitchContext(POVS_SWITCH_CONTEXT switchContext)
     LONG icxRef = 0;
 
     do {
-        ref = gOvsSwitchContextRefCount;
+        ref = InterlockedAdd(&gOvsSwitchContextRefCount, 0);
         newRef = (0 == ref) ? 0 : ref - 1;
         icxRef = InterlockedCompareExchange(&gOvsSwitchContextRefCount,
                                             newRef,
@@ -531,7 +538,7 @@ OvsAcquireSwitchContext(VOID)
     BOOLEAN ret = FALSE;
 
     do {
-        ref = gOvsSwitchContextRefCount;
+        ref = InterlockedAdd(&gOvsSwitchContextRefCount, 0);
         newRef = (0 == ref) ? 0 : ref + 1;
         icxRef = InterlockedCompareExchange(&gOvsSwitchContextRefCount,
                                             newRef,
