@@ -25,10 +25,10 @@
 
 extern NDIS_HANDLE gOvsExtDriverHandle;
 
+_Use_decl_annotations_
 VOID*
 OvsAllocateMemoryWithTag(size_t size, ULONG tag)
 {
-    OVS_VERIFY_IRQL_LE(DISPATCH_LEVEL);
     return NdisAllocateMemoryWithTagPriority(gOvsExtDriverHandle,
         (UINT32)size, tag, NormalPoolPriority);
 }
@@ -40,19 +40,18 @@ OvsFreeMemoryWithTag(VOID *ptr, ULONG tag)
     NdisFreeMemoryWithTagPriority(gOvsExtDriverHandle, ptr, tag);
 }
 
+_Use_decl_annotations_
 VOID *
 OvsAllocateMemory(size_t size)
 {
-    OVS_VERIFY_IRQL_LE(DISPATCH_LEVEL);
     return NdisAllocateMemoryWithTagPriority(gOvsExtDriverHandle,
         (UINT32)size, OVS_MEMORY_TAG, NormalPoolPriority);
 }
 
+_Use_decl_annotations_
 VOID *
 OvsAllocateAlignedMemory(size_t size, UINT16 align)
 {
-    OVS_VERIFY_IRQL_LE(DISPATCH_LEVEL);
-
     ASSERT((align == 8) || (align == 16));
 
     if ((align == 8) || (align == 16)) {
@@ -60,7 +59,7 @@ OvsAllocateAlignedMemory(size_t size, UINT16 align)
          * XXX: NdisAllocateMemory*() functions don't talk anything about
          * alignment. Hence using ExAllocatePool*();
          */
-        return (VOID *)ExAllocatePoolWithTagPriority(NonPagedPool, size,
+        return (VOID *)ExAllocatePoolWithTagPriority(NonPagedPoolNx, size,
                                                      OVS_MEMORY_TAG,
                                                      NormalPoolPriority);
     }
