@@ -100,6 +100,12 @@ static NTSTATUS GetNICAlias(PNDIS_SWITCH_NIC_PARAMETERS nicParam,
 static NTSTATUS OvsConvertIfCountedStrToAnsiStr(PIF_COUNTED_STRING wStr,
                                                 CHAR *str,
                                                 UINT16 maxStrLen);
+_Requires_lock_held_(switchContext->dispatchLock)
+static VOID OvsBindVportWithIpHelper(POVS_VPORT_ENTRY vport,
+                                     POVS_SWITCH_CONTEXT switchContext);
+_Requires_lock_held_(switchContext->dispatchLock)
+static VOID OvsUnBindVportWithIpHelper(POVS_VPORT_ENTRY vport,
+                                       POVS_SWITCH_CONTEXT switchContext);
 
 /*
  * --------------------------------------------------------------------------
@@ -135,7 +141,7 @@ HvCreatePort(POVS_SWITCH_CONTEXT switchContext,
      * Lookup by port name to see if this port with this name had been added
      * (and deleted) previously.
      */
-    vport = OvsFindVportByHvNameW(gOvsSwitchContext,
+    vport = OvsFindVportByHvNameW(switchContext,
                                   portParam->PortFriendlyName.String,
                                   portParam->PortFriendlyName.Length);
     if (vport && vport->isAbsentOnHv == FALSE) {
@@ -452,12 +458,11 @@ HvConnectNic(POVS_SWITCH_CONTEXT switchContext,
     vport->ovsState = OVS_STATE_CONNECTED;
     vport->nicState = NdisSwitchNicStateConnected;
 
-    NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
-
     if (nicParam->NicType == NdisSwitchNicTypeInternal) {
-        OvsInternalAdapterUp(vport->portNo, &vport->netCfgInstanceId);
+        OvsBindVportWithIpHelper(vport, switchContext);
     }
 
+    NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
 done:
     VPORT_NIC_EXIT(nicParam);
 }
@@ -623,6 +628,7 @@ HvDisconnectNic(POVS_SWITCH_CONTEXT switchContext,
     event.upcallPid = vport->upcallPid;
     RtlCopyMemory(&event.ovsName, &vport->ovsName, sizeof event.ovsName);
     event.type = OVS_EVENT_LINK_DOWN;
+    OvsPostVportEvent(&event);
 
     /*
      * Delete the port from the hash tables accessible to userspace. After this
@@ -630,15 +636,20 @@ HvDisconnectNic(POVS_SWITCH_CONTEXT switchContext,
      */
     if (OvsIsRealExternalVport(vport)) {
         OvsRemoveAndDeleteVport(NULL, switchContext, vport, FALSE, TRUE);
-        OvsPostVportEvent(&event);
     }
-    NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
 
     if (isInternalPort) {
-        OvsInternalAdapterDown(vport->portNo, vport->netCfgInstanceId);
-        OvsRemoveAndDeleteVport(NULL, switchContext, vport, TRUE, TRUE);
-        OvsPostVportEvent(&event);
+        OvsUnBindVportWithIpHelper(vport, switchContext);
+        /*
+         * Don't delete the port from the hash tables here for internal port
+         * because the internal port cannot be recreated in HvCreateNic(). It
+         * only can be created in HvCreatePort() by issuing
+         * OID_SWITCH_PORT_CREATE. We should wait extensible switch interface
+         * to issue OID_SWITCH_PORT_TEARDOWN and OID_SWITCH_PORT_DELETE to
+         * delete the internal port.
+         */
     }
+    NdisReleaseRWLock(switchContext->dispatchLock, &lockState);
 
 done:
     VPORT_NIC_EXIT(nicParam);
@@ -693,6 +704,7 @@ done:
 /*
  * OVS Vport related functionality.
  */
+_Use_decl_annotations_
 POVS_VPORT_ENTRY
 OvsFindVportByPortNo(POVS_SWITCH_CONTEXT switchContext,
                      UINT32 portNo)
@@ -787,6 +799,7 @@ OvsFindTunnelVportByPortType(POVS_SWITCH_CONTEXT switchContext,
     return NULL;
 }
 
+_Use_decl_annotations_
 POVS_VPORT_ENTRY
 OvsFindVportByOvsName(POVS_SWITCH_CONTEXT switchContext,
                       PSTR name)
@@ -810,6 +823,7 @@ OvsFindVportByOvsName(POVS_SWITCH_CONTEXT switchContext,
 }
 
 /* OvsFindVportByHvName: "name" is assumed to be null-terminated */
+_Use_decl_annotations_
 POVS_VPORT_ENTRY
 OvsFindVportByHvNameW(POVS_SWITCH_CONTEXT switchContext,
                       PWSTR wsName, SIZE_T wstrSize)
@@ -862,6 +876,7 @@ Cleanup:
     return vport;
 }
 
+_Use_decl_annotations_
 POVS_VPORT_ENTRY
 OvsFindVportByHvNameA(POVS_SWITCH_CONTEXT switchContext,
                       PSTR name)
@@ -884,6 +899,7 @@ OvsFindVportByHvNameA(POVS_SWITCH_CONTEXT switchContext,
     return vport;
 }
 
+_Use_decl_annotations_
 POVS_VPORT_ENTRY
 OvsFindVportByPortIdAndNicIndex(POVS_SWITCH_CONTEXT switchContext,
                                 NDIS_SWITCH_PORT_ID portId,
@@ -1144,8 +1160,13 @@ GetNICAlias(PNDIS_SWITCH_NIC_PARAMETERS nicParam,
         if (status == STATUS_SUCCESS) {
             RtlStringCbPrintfW(portFriendlyName->String,
                                IF_MAX_STRING_SIZE, L"%s", interfaceName);
-            RtlStringCbLengthW(portFriendlyName->String, IF_MAX_STRING_SIZE,
-                               &len);
+            status = RtlStringCbLengthW(portFriendlyName->String,
+                                        IF_MAX_STRING_SIZE, &len);
+            if (!NT_SUCCESS(status)) {
+                OVS_LOG_ERROR("Failed to get the length of the string,"
+                              "status: %x", status);
+                return status;
+            }
             portFriendlyName->Length = (USHORT)len;
         } else {
             OVS_LOG_ERROR("Fail to convert interface LUID to alias, status: %x",
@@ -1214,7 +1235,6 @@ UpdateSwitchCtxWithVport(POVS_SWITCH_CONTEXT switchContext,
     if (newPort) {
         switchContext->numHvVports++;
     }
-    return;
 }
 
 /*
@@ -1300,7 +1320,7 @@ OvsRemoveAndDeleteVport(PVOID usrParamsContext,
         if (hvDelete && vport->isAbsentOnHv == FALSE) {
             switchContext->countInternalVports--;
             ASSERT(switchContext->countInternalVports >= 0);
-            OvsInternalAdapterDown(vport->portNo, vport->netCfgInstanceId);
+            OvsUnBindVportWithIpHelper(vport, switchContext);
         }
         hvSwitchPort = TRUE;
         break;
@@ -2806,3 +2826,25 @@ OvsTunnelVportPendingInit(PVOID context,
         ASSERT(*replyLen != 0);
     }
 }
+
+_Use_decl_annotations_
+static VOID
+OvsBindVportWithIpHelper(POVS_VPORT_ENTRY vport,
+                         POVS_SWITCH_CONTEXT switchContext)
+{
+    OVS_LOG_TRACE("OvsBindVportWithIpHelper: %d", vport->portNo);
+    ASSERT(switchContext->ipHelperBoundVportNo == 0);
+    switchContext->ipHelperBoundVportNo = vport->portNo;
+    OvsInternalAdapterUp(vport->portNo, &vport->netCfgInstanceId);
+}
+
+_Use_decl_annotations_
+static VOID
+OvsUnBindVportWithIpHelper(POVS_VPORT_ENTRY vport,
+                           POVS_SWITCH_CONTEXT switchContext)
+{
+    OVS_LOG_TRACE("OvsUnBindVportWithIpHelper: %d", vport->portNo);
+    ASSERT(switchContext->ipHelperBoundVportNo == vport->portNo);
+    switchContext->ipHelperBoundVportNo = 0;
+    OvsInternalAdapterDown(vport->portNo, vport->netCfgInstanceId);
+}
