@@ -15,6 +15,10 @@
 #define rcu_dereference_raw(p) rcu_dereference_check(p, 1)
 #endif
 
+#ifndef rcu_access_pointer
+#define rcu_access_pointer(p)	rcu_dereference(p)
+#endif
+
 #ifndef HAVE_RCU_READ_LOCK_HELD
 static inline int rcu_read_lock_held(void)
 {
@@ -22,4 +26,16 @@ static inline int rcu_read_lock_held(void)
 }
 #endif
 
+#ifndef RCU_INITIALIZER
+#define RCU_INITIALIZER(v) (typeof(*(v)) __force __rcu *)(v)
+#endif
+
+#ifndef RCU_INIT_POINTER
+#define RCU_INIT_POINTER(p, v) \
+        do { \
+                p = RCU_INITIALIZER(v); \
+        } while (0)
+
+#endif
+
 #endif /* linux/rcupdate.h wrapper */
