@@ -38,7 +38,6 @@
 #include "datapath.h"
 #include "gso.h"
 #include "vport.h"
-#include "gso.h"
 #include "vport-netdev.h"
 
 #define LISP_UDP_PORT		4341
@@ -457,6 +456,7 @@ static int lisp_open(struct net_device *dev)
 
 	rcu_assign_pointer(lisp->sock, sock);
 	/* Mark socket as an encapsulation socket */
+	memset(&tunnel_cfg, 0, sizeof(tunnel_cfg));
 	tunnel_cfg.sk_user_data = dev;
 	tunnel_cfg.encap_type = 1;
 	tunnel_cfg.encap_rcv = lisp_rcv;
@@ -598,10 +598,8 @@ static void lisp_setup(struct net_device *dev)
 	dev->features    |= NETIF_F_RXCSUM;
 	dev->features    |= NETIF_F_GSO_SOFTWARE;
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
 	dev->hw_features |= NETIF_F_SG | NETIF_F_HW_CSUM | NETIF_F_RXCSUM;
 	dev->hw_features |= NETIF_F_GSO_SOFTWARE;
-#endif
 #ifdef USE_UPSTREAM_TUNNEL
 	netif_keep_dst(dev);
 #endif
@@ -613,7 +611,7 @@ static const struct nla_policy lisp_policy[IFLA_LISP_MAX + 1] = {
 	[IFLA_LISP_PORT]              = { .type = NLA_U16 },
 };
 
-#ifdef HAVE_EXT_ACK_IN_RTNL_LINKOPS
+#ifdef HAVE_RTNLOP_VALIDATE_WITH_EXTACK
 static int lisp_validate(struct nlattr *tb[], struct nlattr *data[],
 			 struct netlink_ext_ack __always_unused *extack)
 #else
