@@ -103,81 +103,18 @@ enum nx_hash_fields {
      *  - NXM_OF_TCP_SRC / NXM_OF_TCP_DST
      *  - NXM_OF_UDP_SRC / NXM_OF_UDP_DST
      */
-    NX_HASH_FIELDS_SYMMETRIC_L3L4_UDP
+    NX_HASH_FIELDS_SYMMETRIC_L3L4_UDP,
 
+    /* Network source address (NXM_OF_IP_SRC) only. */
+    NX_HASH_FIELDS_NW_SRC,
 
+    /* Network destination address (NXM_OF_IP_DST) only. */
+    NX_HASH_FIELDS_NW_DST,
+
+    /* Both network destination and source destination addresses. */
+    NX_HASH_FIELDS_SYMMETRIC_L3
 };
 
-/* This command enables or disables an Open vSwitch extension that allows a
- * controller to specify the OpenFlow table to which a flow should be added,
- * instead of having the switch decide which table is most appropriate as
- * required by OpenFlow 1.0.  Because NXM was designed as an extension to
- * OpenFlow 1.0, the extension applies equally to ofp10_flow_mod and
- * nx_flow_mod.  By default, the extension is disabled.
- *
- * When this feature is enabled, Open vSwitch treats struct ofp10_flow_mod's
- * and struct nx_flow_mod's 16-bit 'command' member as two separate fields.
- * The upper 8 bits are used as the table ID, the lower 8 bits specify the
- * command as usual.  A table ID of 0xff is treated like a wildcarded table ID.
- *
- * The specific treatment of the table ID depends on the type of flow mod:
- *
- *    - OFPFC_ADD: Given a specific table ID, the flow is always placed in that
- *      table.  If an identical flow already exists in that table only, then it
- *      is replaced.  If the flow cannot be placed in the specified table,
- *      either because the table is full or because the table cannot support
- *      flows of the given type, the switch replies with an OFPFMFC_TABLE_FULL
- *      error.  (A controller can distinguish these cases by comparing the
- *      current and maximum number of entries reported in ofp_table_stats.)
- *
- *      If the table ID is wildcarded, the switch picks an appropriate table
- *      itself.  If an identical flow already exist in the selected flow table,
- *      then it is replaced.  The choice of table might depend on the flows
- *      that are already in the switch; for example, if one table fills up then
- *      the switch might fall back to another one.
- *
- *    - OFPFC_MODIFY, OFPFC_DELETE: Given a specific table ID, only flows
- *      within that table are matched and modified or deleted.  If the table ID
- *      is wildcarded, flows within any table may be matched and modified or
- *      deleted.
- *
- *    - OFPFC_MODIFY_STRICT, OFPFC_DELETE_STRICT: Given a specific table ID,
- *      only a flow within that table may be matched and modified or deleted.
- *      If the table ID is wildcarded and exactly one flow within any table
- *      matches, then it is modified or deleted; if flows in more than one
- *      table match, then none is modified or deleted.
- */
-struct nx_flow_mod_table_id {
-    uint8_t set;                /* Nonzero to enable, zero to disable. */
-    uint8_t pad[7];
-};
-OFP_ASSERT(sizeof(struct nx_flow_mod_table_id) == 8);
-
-enum nx_packet_in_format {
-    NXPIF_STANDARD = 0,         /* OFPT_PACKET_IN for this OpenFlow version. */
-    NXPIF_NXT_PACKET_IN = 1,    /* NXT_PACKET_IN (since OVS v1.1). */
-    NXPIF_NXT_PACKET_IN2 = 2,   /* NXT_PACKET_IN2 (since OVS v2.6). */
-};
-
-/* NXT_SET_PACKET_IN_FORMAT request.
- *
- * For any given OpenFlow version, Open vSwitch supports multiple formats for
- * "packet-in" messages.  The default is always the standard format for the
- * OpenFlow version in question, but NXT_SET_PACKET_IN_FORMAT can be used to
- * set an alternative format.
- *
- * From OVS v1.1 to OVS v2.5, this request was only honored for OpenFlow 1.0.
- * Requests to set format NXPIF_NXT_PACKET_IN were accepted for OF1.1+ but they
- * had no effect.  (Requests to set formats other than NXPIF_STANDARD or
- * NXPIF_NXT_PACKET_IN were rejected with OFPBRC_EPERM.)
- *
- * From OVS v2.6 onward, this request is honored for all OpenFlow versions.
- */
-struct nx_set_packet_in_format {
-    ovs_be32 format;            /* One of NXPIF_*. */
-};
-OFP_ASSERT(sizeof(struct nx_set_packet_in_format) == 4);
-
 /* NXT_PACKET_IN (analogous to OFPT_PACKET_IN).
  *
  * NXT_PACKET_IN is similar to the OpenFlow 1.2 OFPT_PACKET_IN.  The
@@ -359,16 +296,16 @@ enum nx_packet_in2_prop_type {
  *
  * The other possible roles are a related pair:
  *
- *    - Master (NX_ROLE_MASTER) is equivalent to Other, except that there may
- *      be at most one Master controller at a time: when a controller
- *      configures itself as Master, any existing Master is demoted to the
- *      Slave role.
+ *    - Primary (NX_ROLE_PRIMARY) is equivalent to Other, except that there may
+ *      be at most one Primary controller at a time: when a controller
+ *      configures itself as Primary, any existing Primary is demoted to the
+ *      Secondary role.
  *
- *    - Slave (NX_ROLE_SLAVE) allows the controller read-only access to
+ *    - Secondary (NX_ROLE_SECONDARY) allows the controller read-only access to
  *      OpenFlow features.  In particular attempts to modify the flow table
  *      will be rejected with an OFPBRC_EPERM error.
  *
- *      Slave controllers do not receive OFPT_PACKET_IN or OFPT_FLOW_REMOVED
+ *      Secondary controllers do not receive OFPT_PACKET_IN or OFPT_FLOW_REMOVED
  *      messages, but they do receive OFPT_PORT_STATUS messages.
  */
 struct nx_role_request {
@@ -378,23 +315,23 @@ OFP_ASSERT(sizeof(struct nx_role_request) == 4);
 
 enum nx_role {
     NX_ROLE_OTHER,              /* Default role, full access. */
-    NX_ROLE_MASTER,             /* Full access, at most one. */
-    NX_ROLE_SLAVE               /* Read-only access. */
+    NX_ROLE_PRIMARY,            /* Full access, at most one. */
+    NX_ROLE_SECONDARY           /* Read-only access. */
 };
 
 /* NXT_SET_ASYNC_CONFIG.
  *
  * Sent by a controller, this message configures the asynchronous messages that
  * the controller wants to receive.  Element 0 in each array specifies messages
- * of interest when the controller has an "other" or "master" role; element 1,
- * when the controller has a "slave" role.
+ * of interest when the controller has an "other" or "primary" role; element 1,
+ * when the controller has a "secondary" role.
  *
  * Each array element is a bitmask in which a 0-bit disables receiving a
  * particular message and a 1-bit enables receiving it.  Each bit controls the
  * message whose 'reason' corresponds to the bit index.  For example, the bit
  * with value 1<<2 == 4 in port_status_mask[1] determines whether the
  * controller will receive OFPT_PORT_STATUS messages with reason OFPPR_MODIFY
- * (value 2) when the controller has a "slave" role.
+ * (value 2) when the controller has a "secondary" role.
  *
  * As a side effect, for service controllers, this message changes the
  * miss_send_len from default of zero to OFP_DEFAULT_MISS_SEND_LEN (128).
@@ -613,17 +550,6 @@ OFP_ASSERT(sizeof(struct nx_async_config) == 24);
 /* ## Requests and replies. ## */
 /* ## --------------------- ## */
 
-enum nx_flow_format {
-    NXFF_OPENFLOW10 = 0,         /* Standard OpenFlow 1.0 compatible. */
-    NXFF_NXM = 2                 /* Nicira extended match. */
-};
-
-/* NXT_SET_FLOW_FORMAT request. */
-struct nx_set_flow_format {
-    ovs_be32 format;            /* One of NXFF_*. */
-};
-OFP_ASSERT(sizeof(struct nx_set_flow_format) == 4);
-
 /* NXT_FLOW_MOD (analogous to OFPT_FLOW_MOD).
  *
  * It is possible to limit flow deletions and modifications to certain
@@ -632,8 +558,8 @@ OFP_ASSERT(sizeof(struct nx_set_flow_format) == 4);
  */
 struct nx_flow_mod {
     ovs_be64 cookie;              /* Opaque controller-issued identifier. */
-    ovs_be16 command;             /* OFPFC_* + possibly a table ID (see comment
-                                   * on struct nx_flow_mod_table_id). */
+    ovs_be16 command;             /* OFPFC_*, and table ID if flow_mod_table_id
+                                   * is enabled. */
     ovs_be16 idle_timeout;        /* Idle time before discarding (seconds). */
     ovs_be16 hard_timeout;        /* Max time before discarding (seconds). */
     ovs_be16 priority;            /* Priority level of flow entry. */
