@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012, 2013, 2014, 2015, 2016, 2017 Nicira, Inc.
+ * Copyright (c) 2012-2017, 2019-2020 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -22,9 +22,14 @@
 #include "openflow/openflow.h"
 #include "openflow/nicira-ext.h"
 #include "openvswitch/meta-flow.h"
-#include "openvswitch/ofp-util.h"
 #include "openvswitch/ofp-errors.h"
+#include "openvswitch/ofp-protocol.h"
 #include "openvswitch/types.h"
+#include "openvswitch/ofp-ed-props.h"
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 struct vl_mff_map;
 
@@ -61,14 +66,14 @@ struct vl_mff_map;
     OFPACT(CONTROLLER,      ofpact_controller,  userdata, "controller") \
     OFPACT(ENQUEUE,         ofpact_enqueue,     ofpact, "enqueue")      \
     OFPACT(OUTPUT_REG,      ofpact_output_reg,  ofpact, "output_reg")   \
-    OFPACT(BUNDLE,          ofpact_bundle,      slaves, "bundle")       \
+    OFPACT(BUNDLE,          ofpact_bundle,      members, "bundle")      \
                                                                         \
     /* Header changes. */                                               \
     OFPACT(SET_FIELD,       ofpact_set_field,   ofpact, "set_field")    \
     OFPACT(SET_VLAN_VID,    ofpact_vlan_vid,    ofpact, "set_vlan_vid") \
     OFPACT(SET_VLAN_PCP,    ofpact_vlan_pcp,    ofpact, "set_vlan_pcp") \
     OFPACT(STRIP_VLAN,      ofpact_null,        ofpact, "strip_vlan")   \
-    OFPACT(PUSH_VLAN,       ofpact_null,        ofpact, "push_vlan")    \
+    OFPACT(PUSH_VLAN,       ofpact_push_vlan,   ofpact, "push_vlan")    \
     OFPACT(SET_ETH_SRC,     ofpact_mac,         ofpact, "mod_dl_src")   \
     OFPACT(SET_ETH_DST,     ofpact_mac,         ofpact, "mod_dl_dst")   \
     OFPACT(SET_IPV4_SRC,    ofpact_ipv4,        ofpact, "mod_nw_src")   \
@@ -88,6 +93,12 @@ struct vl_mff_map;
     OFPACT(DEC_MPLS_TTL,    ofpact_null,        ofpact, "dec_mpls_ttl") \
     OFPACT(PUSH_MPLS,       ofpact_push_mpls,   ofpact, "push_mpls")    \
     OFPACT(POP_MPLS,        ofpact_pop_mpls,    ofpact, "pop_mpls")     \
+    OFPACT(DEC_NSH_TTL,     ofpact_null,        ofpact, "dec_nsh_ttl")  \
+    OFPACT(DELETE_FIELD,    ofpact_delete_field, ofpact, "delete_field") \
+                                                                        \
+    /* Generic encap & decap */                                         \
+    OFPACT(ENCAP,           ofpact_encap,       props, "encap")         \
+    OFPACT(DECAP,           ofpact_decap,       ofpact, "decap")        \
                                                                         \
     /* Metadata. */                                                     \
     OFPACT(SET_TUNNEL,      ofpact_tunnel,      ofpact, "set_tunnel")   \
@@ -113,14 +124,17 @@ struct vl_mff_map;
     OFPACT(NAT,             ofpact_nat,         ofpact, "nat")          \
     OFPACT(OUTPUT_TRUNC,    ofpact_output_trunc,ofpact, "output_trunc") \
     OFPACT(CLONE,           ofpact_nest,        actions, "clone")       \
+    OFPACT(CHECK_PKT_LARGER, ofpact_check_pkt_larger, ofpact,           \
+           "check_pkt_larger")                                          \
                                                                         \
     /* Debugging actions.                                               \
      *                                                                  \
      * These are intentionally undocumented, subject to change, and     \
      * only accepted if ovs-vswitchd is started with --enable-dummy. */ \
     OFPACT(DEBUG_RECIRC, ofpact_null,           ofpact, "debug_recirc") \
+    OFPACT(DEBUG_SLOW,   ofpact_null,           ofpact, "debug_slow")   \
                                                                         \
-    /* Instructions. */                                                 \
+    /* Instructions ("meter" is an action in OF1.5+). */                \
     OFPACT(METER,           ofpact_meter,       ofpact, "meter")        \
     OFPACT(CLEAR_ACTIONS,   ofpact_null,        ofpact, "clear_actions") \
     OFPACT(WRITE_ACTIONS,   ofpact_nest,        actions, "write_actions") \
@@ -195,7 +209,8 @@ BUILD_ASSERT_DECL(sizeof(struct ofpact) == 4);
 static inline struct ofpact *
 ofpact_next(const struct ofpact *ofpact)
 {
-    return (void *) ((uint8_t *) ofpact + OFPACT_ALIGN(ofpact->len));
+    return ALIGNED_CAST(struct ofpact *,
+                        (uint8_t *) ofpact + OFPACT_ALIGN(ofpact->len));
 }
 
 struct ofpact *ofpact_next_flattened(const struct ofpact *);
@@ -203,7 +218,23 @@ struct ofpact *ofpact_next_flattened(const struct ofpact *);
 static inline struct ofpact *
 ofpact_end(const struct ofpact *ofpacts, size_t ofpacts_len)
 {
-    return (void *) ((uint8_t *) ofpacts + ofpacts_len);
+    return ofpacts
+           ? ALIGNED_CAST(struct ofpact *, (uint8_t *) ofpacts + ofpacts_len)
+           : NULL;
+}
+
+static inline bool
+ofpact_last(const struct ofpact *a, const struct ofpact *ofpacts,
+            size_t ofpact_len)
+{
+    return ofpact_next(a) == ofpact_end(ofpacts, ofpact_len);
+}
+
+static inline size_t
+ofpact_remaining_len(const struct ofpact *a, const struct ofpact *ofpacts,
+                     size_t ofpact_len)
+{
+    return ofpact_len - ((uint8_t *)a - (uint8_t *)ofpacts);
 }
 
 static inline const struct ofpact *
@@ -256,18 +287,24 @@ ofpact_find_type_flattened(const struct ofpact *a, enum ofpact_type type,
  * Action structure for actions that do not have any extra data beyond the
  * action type. */
 struct ofpact_null {
-    struct ofpact ofpact;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+    );
 };
 
 /* OFPACT_OUTPUT.
  *
  * Used for OFPAT10_OUTPUT. */
 struct ofpact_output {
-    struct ofpact ofpact;
-    ofp_port_t port;            /* Output port. */
-    uint16_t max_len;           /* Max send len, for port OFPP_CONTROLLER. */
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ofp_port_t port;        /* Output port. */
+        uint16_t max_len;       /* Max send len, for port OFPP_CONTROLLER. */
+    );
 };
 
+#define NX_CTLR_NO_METER 0
+
 /* OFPACT_CONTROLLER.
  *
  * Used for NXAST_CONTROLLER. */
@@ -286,6 +323,12 @@ struct ofpact_controller {
         /* Arbitrary data to include in the packet-in message (currently,
          * only in NXT_PACKET_IN2). */
         uint16_t userdata_len;
+
+        /* Meter to which this controller action should be associated.
+         * If requested, this will override a "controller" virtual meter.
+         * A value of NX_CTLR_NO_METER means no meter is requested. */
+        uint32_t meter_id;
+        uint32_t provider_meter_id;
     );
     uint8_t userdata[0];
 };
@@ -294,47 +337,53 @@ struct ofpact_controller {
  *
  * Used for OFPAT10_ENQUEUE. */
 struct ofpact_enqueue {
-    struct ofpact ofpact;
-    ofp_port_t port;
-    uint32_t queue;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ofp_port_t port;
+        uint32_t queue;
+    );
 };
 
 /* OFPACT_OUTPUT_REG.
  *
  * Used for NXAST_OUTPUT_REG. */
 struct ofpact_output_reg {
-    struct ofpact ofpact;
-    uint16_t max_len;
-    struct mf_subfield src;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint16_t max_len;
+        struct mf_subfield src;
+    );
 };
 
 /* OFPACT_OUTPUT_TRUNC.
  *
  * Used for NXAST_OUTPUT_TRUNC. */
 struct ofpact_output_trunc {
-    struct ofpact ofpact;
-    ofp_port_t port;            /* Output port. */
-    uint32_t max_len;           /* Max send len. */
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ofp_port_t port;        /* Output port. */
+        uint32_t max_len;       /* Max send len. */
+    );
 };
 
-/* Bundle slave choice algorithm to apply.
+/* Bundle member choice algorithm to apply.
  *
- * In the descriptions below, 'slaves' is the list of possible slaves in the
+ * In the descriptions below, 'members' is the list of possible members in the
  * order they appear in the OpenFlow action. */
 enum nx_bd_algorithm {
-    /* Chooses the first live slave listed in the bundle.
+    /* Chooses the first live member listed in the bundle.
      *
-     * O(n_slaves) performance. */
+     * O(n_members) performance. */
     NX_BD_ALG_ACTIVE_BACKUP = 0,
 
     /* Highest Random Weight.
      *
-     * for i in [0,n_slaves):
+     * for i in [0,n_members):
      *   weights[i] = hash(flow, i)
-     * slave = { slaves[i] such that weights[i] >= weights[j] for all j != i }
+     * member = { members[i] such that weights[i] >= weights[j] for all j != i }
      *
-     * Redistributes 1/n_slaves of traffic when a slave's liveness changes.
-     * O(n_slaves) performance.
+     * Redistributes 1/n_members of traffic when a member's liveness changes.
+     * O(n_members) performance.
      *
      * Uses the 'fields' and 'basis' parameters. */
     NX_BD_ALG_HRW = 1
@@ -344,20 +393,24 @@ enum nx_bd_algorithm {
  *
  * Used for NXAST_BUNDLE. */
 struct ofpact_bundle {
-    struct ofpact ofpact;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
 
-    /* Slave choice algorithm to apply to hash value. */
-    enum nx_bd_algorithm algorithm;
+        /* Member choice algorithm to apply to hash value. */
+        enum nx_bd_algorithm algorithm;
 
-    /* What fields to hash and how. */
-    enum nx_hash_fields fields;
-    uint16_t basis;             /* Universal hash parameter. */
+        /* What fields to hash and how. */
+        enum nx_hash_fields fields;
+        uint16_t basis;         /* Universal hash parameter. */
 
-    struct mf_subfield dst;
+        struct mf_subfield dst;
 
-    /* Slaves for output. */
-    unsigned int n_slaves;
-    ofp_port_t slaves[];
+        bool compat_syntax;
+
+        /* Members for output. */
+        unsigned int n_members;
+    );
+    ofp_port_t members[];
 };
 
 /* OFPACT_SET_VLAN_VID.
@@ -369,10 +422,12 @@ struct ofpact_bundle {
  *
  * Used for OFPAT10_SET_VLAN_VID and OFPAT11_SET_VLAN_VID. */
 struct ofpact_vlan_vid {
-    struct ofpact ofpact;
-    uint16_t vlan_vid;          /* VLAN VID in low 12 bits, 0 in other bits. */
-    bool push_vlan_if_needed;   /* OF 1.0 semantics if true. */
-    bool flow_has_vlan;         /* VLAN present at action validation time? */
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint16_t vlan_vid;        /* VLAN VID in low 12 bits, other bits 0. */
+        bool push_vlan_if_needed; /* OF 1.0 semantics if true. */
+        bool flow_has_vlan;       /* VLAN present at action validation time? */
+    );
 };
 
 /* OFPACT_SET_VLAN_PCP.
@@ -384,76 +439,104 @@ struct ofpact_vlan_vid {
  *
  * Used for OFPAT10_SET_VLAN_PCP and OFPAT11_SET_VLAN_PCP. */
 struct ofpact_vlan_pcp {
-    struct ofpact ofpact;
-    uint8_t vlan_pcp;           /* VLAN PCP in low 3 bits, 0 in other bits. */
-    bool push_vlan_if_needed;   /* OF 1.0 semantics if true. */
-    bool flow_has_vlan;         /* VLAN present at action validation time? */
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t vlan_pcp;           /* VLAN PCP in low 3 bits, other bits 0. */
+        bool push_vlan_if_needed;   /* OF 1.0 semantics if true. */
+        bool flow_has_vlan;         /* VLAN present at action validation? */
+    );
+};
+
+/* OFPACT_PUSH_VLAN.
+ *
+ * Used for OFPAT11_PUSH_VLAN. */
+struct ofpact_push_vlan {
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ovs_be16 ethertype;
+    );
 };
 
 /* OFPACT_SET_ETH_SRC, OFPACT_SET_ETH_DST.
  *
  * Used for OFPAT10_SET_DL_SRC, OFPAT10_SET_DL_DST. */
 struct ofpact_mac {
-    struct ofpact ofpact;
-    struct eth_addr mac;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        struct eth_addr mac;
+    );
 };
 
 /* OFPACT_SET_IPV4_SRC, OFPACT_SET_IPV4_DST.
  *
  * Used for OFPAT10_SET_NW_SRC, OFPAT10_SET_NW_DST. */
 struct ofpact_ipv4 {
-    struct ofpact ofpact;
-    ovs_be32 ipv4;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ovs_be32 ipv4;
+    );
 };
 
 /* OFPACT_SET_IP_DSCP.
  *
  * Used for OFPAT10_SET_NW_TOS. */
 struct ofpact_dscp {
-    struct ofpact ofpact;
-    uint8_t dscp;               /* DSCP in high 6 bits, rest ignored. */
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t dscp;           /* DSCP in high 6 bits, rest ignored. */
+    );
 };
 
 /* OFPACT_SET_IP_ECN.
  *
  * Used for OFPAT11_SET_NW_ECN. */
 struct ofpact_ecn {
-    struct ofpact ofpact;
-    uint8_t ecn;               /* ECN in low 2 bits, rest ignored. */
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t ecn;            /* ECN in low 2 bits, rest ignored. */
+    );
 };
 
 /* OFPACT_SET_IP_TTL.
  *
  * Used for OFPAT11_SET_NW_TTL. */
 struct ofpact_ip_ttl {
-    struct ofpact ofpact;
-    uint8_t ttl;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t ttl;
+    );
 };
 
 /* OFPACT_SET_L4_SRC_PORT, OFPACT_SET_L4_DST_PORT.
  *
  * Used for OFPAT10_SET_TP_SRC, OFPAT10_SET_TP_DST. */
 struct ofpact_l4_port {
-    struct ofpact ofpact;
-    uint16_t port;              /* TCP, UDP or SCTP port number. */
-    uint8_t  flow_ip_proto;     /* IP proto from corresponding match, or 0 */
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint16_t port;          /* TCP, UDP or SCTP port number. */
+        uint8_t  flow_ip_proto; /* IP proto from corresponding match, or 0 */
+    );
 };
 
 /* OFPACT_REG_MOVE.
  *
  * Used for NXAST_REG_MOVE. */
 struct ofpact_reg_move {
-    struct ofpact ofpact;
-    struct mf_subfield src;
-    struct mf_subfield dst;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        struct mf_subfield src;
+        struct mf_subfield dst;
+    );
 };
 
-/* OFPACT_STACK_PUSH.
+/* OFPACT_STACK_PUSH, OFPACT_STACK_POP.
  *
  * Used for NXAST_STACK_PUSH and NXAST_STACK_POP. */
 struct ofpact_stack {
-    struct ofpact ofpact;
-    struct mf_subfield subfield;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        struct mf_subfield subfield;
+    );
 };
 
 /* OFPACT_SET_FIELD.
@@ -482,60 +565,95 @@ BUILD_ASSERT_DECL(offsetof(struct ofpact_set_field, value)
  *
  * Used for NXAST_PUSH_MPLS, OFPAT11_PUSH_MPLS. */
 struct ofpact_push_mpls {
-    struct ofpact ofpact;
-    ovs_be16 ethertype;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ovs_be16 ethertype;
+    );
 };
 
 /* OFPACT_POP_MPLS
  *
  * Used for NXAST_POP_MPLS, OFPAT11_POP_MPLS.. */
 struct ofpact_pop_mpls {
-    struct ofpact ofpact;
-    ovs_be16 ethertype;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ovs_be16 ethertype;
+    );
+};
+
+/* OFPACT_DELETE_FIELD.
+ *
+ * Used for NXAST_DELETE_FIELD. */
+struct ofpact_delete_field {
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        const struct mf_field *field;
+    );
 };
 
 /* OFPACT_SET_TUNNEL.
  *
  * Used for NXAST_SET_TUNNEL, NXAST_SET_TUNNEL64. */
 struct ofpact_tunnel {
-    struct ofpact ofpact;
-    uint64_t tun_id;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint64_t tun_id;
+    );
 };
 
 /* OFPACT_SET_QUEUE.
  *
  * Used for NXAST_SET_QUEUE. */
 struct ofpact_queue {
-    struct ofpact ofpact;
-    uint32_t queue_id;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint32_t queue_id;
+    );
 };
 
 /* OFPACT_FIN_TIMEOUT.
  *
  * Used for NXAST_FIN_TIMEOUT. */
 struct ofpact_fin_timeout {
-    struct ofpact ofpact;
-    uint16_t fin_idle_timeout;
-    uint16_t fin_hard_timeout;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint16_t fin_idle_timeout;
+        uint16_t fin_hard_timeout;
+    );
 };
 
 /* OFPACT_WRITE_METADATA.
  *
  * Used for NXAST_WRITE_METADATA. */
 struct ofpact_metadata {
-    struct ofpact ofpact;
-    ovs_be64 metadata;
-    ovs_be64 mask;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ovs_be64 metadata;
+        ovs_be64 mask;
+    );
 };
 
 /* OFPACT_METER.
  *
  * Used for OFPIT13_METER. */
 struct ofpact_meter {
-    struct ofpact ofpact;
-    uint32_t meter_id;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint32_t meter_id;
+        uint32_t provider_meter_id;
+    );
 };
 
+/* OFPACT_CHECK_PKT_LARGER.
+ *
+ * Used for NXAST_CHECK_PKT_LARGER. */
+struct ofpact_check_pkt_larger {
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        struct mf_subfield dst;
+        uint16_t pkt_len;
+    );
+};
 /* OFPACT_WRITE_ACTIONS, OFPACT_CLONE.
  *
  * Used for OFPIT11_WRITE_ACTIONS, NXAST_CLONE. */
@@ -556,9 +674,14 @@ ofpact_nest_get_action_len(const struct ofpact_nest *on)
 /* Bits for 'flags' in struct nx_action_conntrack.
  *
  * If NX_CT_F_COMMIT is set, then the connection entry is moved from the
- * unconfirmed to confirmed list in the tracker. */
+ * unconfirmed to confirmed list in the tracker.
+ * If NX_CT_F_FORCE is set, in addition to NX_CT_F_COMMIT, then the conntrack
+ * entry is replaced with a new one in case the original direction of the
+ * existing entry is opposite of the current packet direction.
+ */
 enum nx_conntrack_flags {
     NX_CT_F_COMMIT = 1 << 0,
+    NX_CT_F_FORCE  = 1 << 1,
 };
 
 /* Magic value for struct nx_action_conntrack 'recirc_table' field, to specify
@@ -616,35 +739,41 @@ enum nx_nat_flags {
  *
  * Used for NXAST_NAT. */
 struct ofpact_nat {
-    struct ofpact ofpact;
-    uint8_t range_af; /* AF_UNSPEC, AF_INET, or AF_INET6 */
-    uint16_t flags;  /* NX_NAT_F_* */
-    struct {
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t range_af; /* AF_UNSPEC, AF_INET, or AF_INET6 */
+        uint16_t flags;  /* NX_NAT_F_* */
         struct {
-            uint16_t min;
-            uint16_t max;
-        } proto;
-        union {
             struct {
-                ovs_be32 min;
-                ovs_be32 max;
-            } ipv4;
-            struct {
-                struct in6_addr min;
-                struct in6_addr max;
-            } ipv6;
-        } addr;
-    } range;
+                uint16_t min;
+                uint16_t max;
+            } proto;
+            union {
+                struct {
+                    ovs_be32 min;
+                    ovs_be32 max;
+                } ipv4;
+                struct {
+                    struct in6_addr min;
+                    struct in6_addr max;
+                } ipv6;
+            } addr;
+        } range;
+    );
 };
 
 
 /* OFPACT_RESUBMIT.
  *
- * Used for NXAST_RESUBMIT, NXAST_RESUBMIT_TABLE. */
+ * Used for NXAST_RESUBMIT, NXAST_RESUBMIT_TABLE, NXAST_RESUBMIT_TABLE_CT. */
 struct ofpact_resubmit {
-    struct ofpact ofpact;
-    ofp_port_t in_port;
-    uint8_t table_id;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ofp_port_t in_port;
+        uint8_t table_id;
+        bool with_ct_orig;  /* Resubmit with Conntrack original direction tuple
+                             * fields in place of IP header fields. */
+    );
 };
 
 /* Bits for 'flags' in struct nx_action_learn.
@@ -652,6 +781,10 @@ struct ofpact_resubmit {
  * If NX_LEARN_F_SEND_FLOW_REM is set, then the learned flows will have their
  * OFPFF_SEND_FLOW_REM flag set.
  *
+ * If NX_LEARN_F_WRITE_RESULT is set, then the actions will write whether the
+ * learn operation succeded on a bit.  If the learn is successful the bit will
+ * be set, otherwise (e.g. if the limit is hit) the bit will be unset.
+ *
  * If NX_LEARN_F_DELETE_LEARNED is set, then removing this action will delete
  * all the flows from the learn action's 'table_id' that have the learn
  * action's 'cookie'.  Important points:
@@ -677,6 +810,7 @@ struct ofpact_resubmit {
 enum nx_learn_flags {
     NX_LEARN_F_SEND_FLOW_REM = 1 << 0,
     NX_LEARN_F_DELETE_LEARNED = 1 << 1,
+    NX_LEARN_F_WRITE_RESULT = 1 << 2,
 };
 
 #define NX_LEARN_N_BITS_MASK    0x3ff
@@ -699,7 +833,7 @@ struct ofpact_learn_spec {
                                     * NX_LEARN_DST_LOAD only. */
         uint16_t src_type;         /* One of NX_LEARN_SRC_*. */
         uint16_t dst_type;         /* One of NX_LEARN_DST_*. */
-        uint8_t n_bits;            /* Number of bits in source and dest. */
+        uint32_t n_bits;           /* Number of bits in source and dest. */
     );
     /* Followed by 'DIV_ROUND_UP(n_bits, 8)' bytes of immediate data for
      * match 'dst_type's NX_LEARN_DST_MATCH and NX_LEARN_DST_LOAD when
@@ -740,6 +874,13 @@ struct ofpact_learn {
         ovs_be64 cookie;           /* Cookie for new flow. */
         uint16_t fin_idle_timeout; /* Idle timeout after FIN, if nonzero. */
         uint16_t fin_hard_timeout; /* Hard timeout after FIN, if nonzero. */
+        /* If the number of flows on 'table_id' with 'cookie' exceeds this,
+         * the action will not add a new flow. 0 indicates unlimited. */
+        uint32_t limit;
+        /* Used only if 'flags' has NX_LEARN_F_WRITE_RESULT.  If the execution
+         * failed to install a new flow because 'limit' is exceeded,
+         * result_dst will be set to 0, otherwise to 1. */
+        struct mf_subfield result_dst;
     );
 
     struct ofpact_learn_spec specs[];
@@ -815,37 +956,43 @@ enum nx_mp_algorithm {
  *
  * Used for NXAST_CONJUNCTION. */
 struct ofpact_conjunction {
-    struct ofpact ofpact;
-    uint8_t clause;
-    uint8_t n_clauses;
-    uint32_t id;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t clause;
+        uint8_t n_clauses;
+        uint32_t id;
+    );
 };
 
 /* OFPACT_MULTIPATH.
  *
  * Used for NXAST_MULTIPATH. */
 struct ofpact_multipath {
-    struct ofpact ofpact;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
 
-    /* What fields to hash and how. */
-    enum nx_hash_fields fields;
-    uint16_t basis;             /* Universal hash parameter. */
+        /* What fields to hash and how. */
+        enum nx_hash_fields fields;
+        uint16_t basis;         /* Universal hash parameter. */
 
-    /* Multipath link choice algorithm to apply to hash value. */
-    enum nx_mp_algorithm algorithm;
-    uint16_t max_link;          /* Number of output links, minus 1. */
-    uint32_t arg;               /* Algorithm-specific argument. */
+        /* Multipath link choice algorithm to apply to hash value. */
+        enum nx_mp_algorithm algorithm;
+        uint16_t max_link;      /* Number of output links, minus 1. */
+        uint32_t arg;           /* Algorithm-specific argument. */
 
-    /* Where to store the result. */
-    struct mf_subfield dst;
+        /* Where to store the result. */
+        struct mf_subfield dst;
+    );
 };
 
 /* OFPACT_NOTE.
  *
  * Used for NXAST_NOTE. */
 struct ofpact_note {
-    struct ofpact ofpact;
-    size_t length;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        size_t length;
+    );
     uint8_t data[];
 };
 
@@ -867,23 +1014,25 @@ enum nx_action_sample_direction {
  *
  * Used for NXAST_SAMPLE, NXAST_SAMPLE2, and NXAST_SAMPLE3. */
 struct ofpact_sample {
-    struct ofpact ofpact;
-    uint16_t probability;  /* Always positive. */
-    uint32_t collector_set_id;
-    uint32_t obs_domain_id;
-    uint32_t obs_point_id;
-    ofp_port_t sampling_port;
-    enum nx_action_sample_direction direction;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint16_t probability;   /* Always positive. */
+        uint32_t collector_set_id;
+        uint32_t obs_domain_id;
+        uint32_t obs_point_id;
+        ofp_port_t sampling_port;
+        enum nx_action_sample_direction direction;
+    );
 };
 
 /* OFPACT_DEC_TTL.
  *
  * Used for OFPAT11_DEC_NW_TTL, NXAST_DEC_TTL and NXAST_DEC_TTL_CNT_IDS. */
 struct ofpact_cnt_ids {
-    struct ofpact ofpact;
-
-    /* Controller ids. */
-    unsigned int n_controllers;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        unsigned int n_controllers;
+    );
     uint16_t cnt_ids[];
 };
 
@@ -891,54 +1040,94 @@ struct ofpact_cnt_ids {
  *
  * Used for OFPAT11_SET_MPLS_LABEL and NXAST_SET_MPLS_LABEL */
 struct ofpact_mpls_label {
-    struct ofpact ofpact;
-
-    ovs_be32 label;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ovs_be32 label;
+    );
 };
 
 /* OFPACT_SET_MPLS_TC.
  *
  * Used for OFPAT11_SET_MPLS_TC and NXAST_SET_MPLS_TC */
 struct ofpact_mpls_tc {
-    struct ofpact ofpact;
-
-    uint8_t tc;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t tc;
+    );
 };
 
 /* OFPACT_SET_MPLS_TTL.
  *
  * Used for OFPAT11_SET_MPLS_TTL and NXAST_SET_MPLS_TTL */
 struct ofpact_mpls_ttl {
-    struct ofpact ofpact;
-
-    uint8_t ttl;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t ttl;
+    );
 };
 
 /* OFPACT_GOTO_TABLE
  *
  * Used for OFPIT11_GOTO_TABLE */
 struct ofpact_goto_table {
-    struct ofpact ofpact;
-    uint8_t table_id;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint8_t table_id;
+    );
 };
 
 /* OFPACT_GROUP.
  *
  * Used for OFPAT11_GROUP. */
 struct ofpact_group {
-    struct ofpact ofpact;
-    uint32_t group_id;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        uint32_t group_id;
+    );
 };
 
 /* OFPACT_UNROLL_XLATE.
  *
  * Used only internally. */
 struct ofpact_unroll_xlate {
-    struct ofpact ofpact;
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
 
-    /* Metadata in xlate context, visible to controller via PACKET_INs. */
-    uint8_t  rule_table_id;       /* 0xFF if none. */
-    ovs_be64 rule_cookie;         /* OVS_BE64_MAX if none. */
+        /* Metadata in xlate context, visible to controller via PACKET_INs. */
+        uint8_t  rule_table_id; /* 0xFF if none. */
+        ovs_be64 rule_cookie;   /* OVS_BE64_MAX if none. */
+    );
+};
+
+/* OFPACT_ENCAP.
+ *
+ * Used for NXAST_ENCAP. */
+
+struct ofpact_encap {
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+        ovs_be32 new_pkt_type;         /* Packet type of the header to add. */
+        uint16_t hdr_size;             /* New header size in bytes. */
+        uint16_t n_props;              /* Number of encap properties. */
+    );
+    struct ofpact_ed_prop props[]; /* Properties in internal format. */
+};
+
+/* OFPACT_DECAP.
+ *
+ * Used for NXAST_DECAP. */
+struct ofpact_decap {
+    OFPACT_PADDED_MEMBERS(
+        struct ofpact ofpact;
+
+        /* New packet type.
+         *
+         * The special value (0,0xFFFE) "Use next proto" is used to request OVS
+         * to automatically set the new packet type based on the decap'ed
+         * header's next protocol.
+         */
+        ovs_be32 new_pkt_type;
+    );
 };
 
 /* Converting OpenFlow to ofpacts. */
@@ -946,21 +1135,31 @@ enum ofperr ofpacts_pull_openflow_actions(struct ofpbuf *openflow,
                                           unsigned int actions_len,
                                           enum ofp_version version,
                                           const struct vl_mff_map *,
+                                          uint64_t *ofpacts_tlv_bitmap,
                                           struct ofpbuf *ofpacts);
 enum ofperr
 ofpacts_pull_openflow_instructions(struct ofpbuf *openflow,
                                    unsigned int instructions_len,
                                    enum ofp_version version,
                                    const struct vl_mff_map *vl_mff_map,
+                                   uint64_t *ofpacts_tlv_bitmap,
                                    struct ofpbuf *ofpacts);
+
+struct ofpact_check_params {
+    /* Input. */
+    struct match *match;
+    ofp_port_t max_ports;
+    uint8_t table_id;
+    uint8_t n_tables;
+
+    /* Output. */
+    enum ofputil_protocol usable_protocols;
+};
 enum ofperr ofpacts_check(struct ofpact[], size_t ofpacts_len,
-                          struct flow *, ofp_port_t max_ports,
-                          uint8_t table_id, uint8_t n_tables,
-                          enum ofputil_protocol *usable_protocols);
+                          struct ofpact_check_params *);
 enum ofperr ofpacts_check_consistency(struct ofpact[], size_t ofpacts_len,
-                                      struct flow *, ofp_port_t max_ports,
-                                      uint8_t table_id, uint8_t n_tables,
-                                      enum ofputil_protocol usable_protocols);
+                                      enum ofputil_protocol needed_protocols,
+                                      struct ofpact_check_params *);
 enum ofperr ofpact_check_output_port(ofp_port_t port, ofp_port_t max_ports);
 
 /* Converting ofpacts to OpenFlow. */
@@ -983,19 +1182,44 @@ bool ofpacts_output_to_group(const struct ofpact[], size_t ofpacts_len,
                              uint32_t group_id);
 bool ofpacts_equal(const struct ofpact a[], size_t a_len,
                    const struct ofpact b[], size_t b_len);
+bool ofpacts_equal_stringwise(const struct ofpact a[], size_t a_len,
+                              const struct ofpact b[], size_t b_len);
 const struct mf_field *ofpact_get_mf_dst(const struct ofpact *ofpact);
 uint32_t ofpacts_get_meter(const struct ofpact[], size_t ofpacts_len);
 
-/* Formatting and parsing ofpacts. */
-void ofpacts_format(const struct ofpact[], size_t ofpacts_len, struct ds *);
-char *ofpacts_parse_actions(const char *, struct ofpbuf *ofpacts,
-                            enum ofputil_protocol *usable_protocols)
-    OVS_WARN_UNUSED_RESULT;
-char *ofpacts_parse_instructions(const char *, struct ofpbuf *ofpacts,
-                                 enum ofputil_protocol *usable_protocols)
-    OVS_WARN_UNUSED_RESULT;
+/* Formatting ofpacts. */
+struct ofpact_format_params {
+    /* Input. */
+    const struct ofputil_port_map *port_map;
+    const struct ofputil_table_map *table_map;
+
+    /* Output. */
+    struct ds *s;
+};
+void ofpacts_format(const struct ofpact[], size_t ofpacts_len,
+                    const struct ofpact_format_params *);
 const char *ofpact_name(enum ofpact_type);
 
+/* Parsing ofpacts. */
+struct ofpact_parse_params {
+    /* Input. */
+    const struct ofputil_port_map *port_map;
+    const struct ofputil_table_map *table_map;
+
+    /* Output. */
+    struct ofpbuf *ofpacts;
+    enum ofputil_protocol *usable_protocols;
+
+    /* Parse context. */
+    unsigned int depth;
+};
+#define MAX_OFPACT_PARSE_DEPTH 100
+char *ofpacts_parse_actions(const char *, const struct ofpact_parse_params *)
+    OVS_WARN_UNUSED_RESULT;
+char *ofpacts_parse_instructions(const char *,
+                                 const struct ofpact_parse_params *)
+    OVS_WARN_UNUSED_RESULT;
+
 /* Internal use by the helpers below. */
 void ofpact_init(struct ofpact *, enum ofpact_type, size_t len);
 void *ofpact_put(struct ofpbuf *, enum ofpact_type, size_t len);
@@ -1037,21 +1261,23 @@ void *ofpact_finish(struct ofpbuf *, struct ofpact *);
  *
  *     Initializes the parts of 'ofpact' that identify it as having type
  *     OFPACT_<ENUM> and length OFPACT_<ENUM>_SIZE and zeros the rest.
- *
- *   <ENUM>_SIZE
- *
- *     The size of the action structure.  For a fixed-length action, this is
- *     sizeof(struct <STRUCT>) rounded up to a multiple of OFPACT_ALIGNTO.  For
- *     a variable-length action, this is the offset to the variable-length
- *     part.
  */
 #define OFPACT(ENUM, STRUCT, MEMBER, NAME)                              \
     BUILD_ASSERT_DECL(offsetof(struct STRUCT, ofpact) == 0);            \
                                                                         \
-    enum { OFPACT_##ENUM##_SIZE                                         \
-           = (offsetof(struct STRUCT, MEMBER)                           \
-              ? offsetof(struct STRUCT, MEMBER)                         \
-              : OFPACT_ALIGN(sizeof(struct STRUCT))) };                 \
+    /* Action structures must be a multiple of OFPACT_ALIGNTO bytes. */ \
+    BUILD_ASSERT_DECL(sizeof(struct STRUCT) % OFPACT_ALIGNTO == 0);     \
+                                                                        \
+    /* Variable-length data must start at a multiple of OFPACT_ALIGNTO  \
+     * bytes. */                                                        \
+    BUILD_ASSERT_DECL(offsetof(struct STRUCT, MEMBER)                   \
+                      % OFPACT_ALIGNTO == 0);                           \
+                                                                        \
+    /* If there is variable-length data, it starts at the end of the    \
+     * structure. */                                                    \
+    BUILD_ASSERT_DECL(!offsetof(struct STRUCT, MEMBER)                  \
+                      || (offsetof(struct STRUCT, MEMBER)               \
+                          == sizeof(struct STRUCT)));                   \
                                                                         \
     static inline struct STRUCT *                                       \
     ofpact_get_##ENUM(const struct ofpact *ofpact)                      \
@@ -1070,15 +1296,15 @@ void *ofpact_finish(struct ofpbuf *, struct ofpact *);
     static inline struct STRUCT *                                       \
     ofpact_put_##ENUM(struct ofpbuf *ofpacts)                           \
     {                                                                   \
-        return ofpact_put(ofpacts, OFPACT_##ENUM,                       \
-                          OFPACT_##ENUM##_SIZE);                        \
+        return (struct STRUCT *) ofpact_put(ofpacts, OFPACT_##ENUM,     \
+                                            sizeof(struct STRUCT));     \
     }                                                                   \
                                                                         \
     static inline void                                                  \
     ofpact_init_##ENUM(struct STRUCT *ofpact)                           \
     {                                                                   \
         ofpact_init(&ofpact->ofpact, OFPACT_##ENUM,                     \
-                    OFPACT_##ENUM##_SIZE);                              \
+                    sizeof(struct STRUCT));                             \
     }                                                                   \
                                                                         \
     static inline void                                                  \
@@ -1086,7 +1312,7 @@ void *ofpact_finish(struct ofpbuf *, struct ofpact *);
     {                                                                   \
         struct ofpact *ofpact = &(*ofpactp)->ofpact;                    \
         ovs_assert(ofpact->type == OFPACT_##ENUM);                      \
-        *ofpactp = ofpact_finish(ofpbuf, ofpact);                       \
+        *ofpactp = (struct STRUCT *) ofpact_finish(ofpbuf, ofpact);     \
     }
 OFPACTS
 #undef OFPACT
@@ -1149,7 +1375,7 @@ enum {
 const char *ovs_instruction_name_from_type(enum ovs_instruction_type type);
 int ovs_instruction_type_from_name(const char *name);
 enum ovs_instruction_type ovs_instruction_type_from_ofpact_type(
-    enum ofpact_type);
+    enum ofpact_type, enum ofp_version);
 enum ofperr ovs_instruction_type_from_inst_type(
     enum ovs_instruction_type *instruction_type, const uint16_t inst_type);
 
@@ -1157,4 +1383,8 @@ ovs_be32 ovsinst_bitmap_to_openflow(uint32_t ovsinst_bitmap, enum ofp_version);
 uint32_t ovsinst_bitmap_from_openflow(ovs_be32 ofpit_bitmap,
                                       enum ofp_version);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* ofp-actions.h */
