@@ -21,6 +21,7 @@
 #include "smap.h"
 #include "ovs-thread.h"
 #include "openvswitch/vlog.h"
+#include "vswitch-idl.h"
 
 VLOG_DEFINE_THIS_MODULE(dpdk);
 
@@ -37,10 +38,15 @@ dpdk_init(const struct smap *ovs_other_config)
     }
 }
 
+bool
+dpdk_attach_thread(unsigned cpu OVS_UNUSED)
+{
+    return false;
+}
+
 void
-dpdk_set_lcore_id(unsigned cpu OVS_UNUSED)
+dpdk_detach_thread(void)
 {
-    /* Nothing */
 }
 
 const char *
@@ -48,3 +54,41 @@ dpdk_get_vhost_sock_dir(void)
 {
     return NULL;
 }
+
+bool
+dpdk_vhost_iommu_enabled(void)
+{
+    return false;
+}
+
+bool
+dpdk_vhost_postcopy_enabled(void)
+{
+    return false;
+}
+
+bool
+dpdk_per_port_memory(void)
+{
+    return false;
+}
+
+bool
+dpdk_available(void)
+{
+    return false;
+}
+
+void
+print_dpdk_version(void)
+{
+}
+
+void
+dpdk_status(const struct ovsrec_open_vswitch *cfg)
+{
+    if (cfg) {
+        ovsrec_open_vswitch_set_dpdk_initialized(cfg, false);
+        ovsrec_open_vswitch_set_dpdk_version(cfg, "none");
+    }
+}
