@@ -25,6 +25,7 @@
 #include <rte_cpuflags.h>
 #include <rte_errno.h>
 #include <rte_log.h>
+#include <rte_malloc.h>
 #include <rte_memzone.h>
 #include <rte_version.h>
 
@@ -34,6 +35,7 @@
 #include "netdev-offload-provider.h"
 #include "openvswitch/dynamic-string.h"
 #include "openvswitch/vlog.h"
+#include "ovs-atomic.h"
 #include "ovs-numa.h"
 #include "smap.h"
 #include "svec.h"
@@ -49,10 +51,11 @@ static char *vhost_sock_dir = NULL;   /* Location of vhost-user sockets */
 static bool vhost_iommu_enabled = false; /* Status of vHost IOMMU support */
 static bool vhost_postcopy_enabled = false; /* Status of vHost POSTCOPY
                                              * support. */
-static bool dpdk_initialized = false; /* Indicates successful initialization
-                                       * of DPDK. */
 static bool per_port_memory = false; /* Status of per port memory support */
 
+/* Indicates successful initialization of DPDK. */
+static atomic_bool dpdk_initialized = ATOMIC_VAR_INIT(false);
+
 static int
 process_vhost_flags(char *flag, const char *default_val, int size,
                     const struct smap *ovs_other_config,
@@ -129,33 +132,12 @@ construct_dpdk_options(const struct smap *ovs_other_config, struct svec *args)
     }
 }
 
-static char *
-construct_dpdk_socket_mem(void)
-{
-    const char *def_value = "1024";
-    int numa, numa_nodes = ovs_numa_get_n_numas();
-    struct ds dpdk_socket_mem = DS_EMPTY_INITIALIZER;
-
-    if (numa_nodes == 0 || numa_nodes == OVS_NUMA_UNSPEC) {
-        numa_nodes = 1;
-    }
-
-    ds_put_cstr(&dpdk_socket_mem, def_value);
-    for (numa = 1; numa < numa_nodes; ++numa) {
-        ds_put_format(&dpdk_socket_mem, ",%s", def_value);
-    }
-
-    return ds_cstr(&dpdk_socket_mem);
-}
-
 #define MAX_DPDK_EXCL_OPTS 10
 
 static void
 construct_dpdk_mutex_options(const struct smap *ovs_other_config,
                              struct svec *args)
 {
-    char *default_dpdk_socket_mem = construct_dpdk_socket_mem();
-
     struct dpdk_exclusive_options_map {
         const char *category;
         const char *ovs_dpdk_options[MAX_DPDK_EXCL_OPTS];
@@ -166,7 +148,7 @@ construct_dpdk_mutex_options(const struct smap *ovs_other_config,
         {"memory type",
          {"dpdk-alloc-mem", "dpdk-socket-mem", NULL,},
          {"-m",             "--socket-mem",    NULL,},
-         default_dpdk_socket_mem, 1
+         NULL, 0
         },
     };
 
@@ -210,8 +192,6 @@ construct_dpdk_mutex_options(const struct smap *ovs_other_config,
                       "dpdk-extra config", popt->eal_dpdk_options[found_pos]);
         }
     }
-
-    free(default_dpdk_socket_mem);
 }
 
 static void
@@ -356,6 +336,12 @@ dpdk_unixctl_log_set(struct unixctl_conn *conn, int argc, const char *argv[],
     unixctl_command_reply(conn, NULL);
 }
 
+static void
+malloc_dump_stats_wrapper(FILE *stream)
+{
+    rte_malloc_dump_stats(stream, NULL);
+}
+
 static bool
 dpdk_init__(const struct smap *ovs_other_config)
 {
@@ -421,21 +407,12 @@ dpdk_init__(const struct smap *ovs_other_config)
     svec_add(&args, ovs_get_program_name());
     construct_dpdk_args(ovs_other_config, &args);
 
-    if (!args_contains(&args, "--legacy-mem")
-        && !args_contains(&args, "--socket-limit")) {
-        const char *arg;
-        size_t i;
-
-        SVEC_FOR_EACH (i, arg, &args) {
-            if (!strcmp(arg, "--socket-mem")) {
-                break;
-            }
-        }
-        if (i < args.n - 1) {
-            svec_add(&args, "--socket-limit");
-            svec_add(&args, args.names[i + 1]);
-        }
+#ifdef DPDK_IN_MEMORY_SUPPORTED
+    if (!args_contains(&args, "--in-memory") &&
+            !args_contains(&args, "--legacy-mem")) {
+        svec_add(&args, "--in-memory");
     }
+#endif
 
     if (args_contains(&args, "-c") || args_contains(&args, "-l")) {
         auto_determine = false;
@@ -443,7 +420,7 @@ dpdk_init__(const struct smap *ovs_other_config)
 
     /**
      * NOTE: This is an unsophisticated mechanism for determining the DPDK
-     * lcore for the DPDK Master.
+     * main core.
      */
     if (auto_determine) {
         const struct ovs_numa_info_core *core;
@@ -502,6 +479,12 @@ dpdk_init__(const struct smap *ovs_other_config)
         return false;
     }
 
+    if (!rte_mp_disable()) {
+        VLOG_EMER("Could not disable multiprocess, DPDK won't be available.");
+        rte_eal_cleanup();
+        return false;
+    }
+
     if (VLOG_IS_DBG_ENABLED()) {
         size_t size;
         char *response = NULL;
@@ -521,10 +504,15 @@ dpdk_init__(const struct smap *ovs_other_config)
         }
     }
 
+    unixctl_command_register("dpdk/lcore-list", "", 0, 0,
+                             dpdk_unixctl_mem_stream, rte_lcore_dump);
     unixctl_command_register("dpdk/log-list", "", 0, 0,
                              dpdk_unixctl_mem_stream, rte_log_dump);
     unixctl_command_register("dpdk/log-set", "{level | pattern:level}", 0,
                              INT_MAX, dpdk_unixctl_log_set, NULL);
+    unixctl_command_register("dpdk/get-malloc-stats", "", 0, 0,
+                             dpdk_unixctl_mem_stream,
+                             malloc_dump_stats_wrapper);
 
     /* We are called from the main thread here */
     RTE_PER_LCORE(_lcore_id) = NON_PMD_CORE_ID;
@@ -567,7 +555,7 @@ dpdk_init(const struct smap *ovs_other_config)
     } else {
         VLOG_INFO_ONCE("DPDK Disabled - Use other_config:dpdk-init to enable");
     }
-    dpdk_initialized = enabled;
+    atomic_store_relaxed(&dpdk_initialized, enabled);
 }
 
 const char *
@@ -597,15 +585,40 @@ dpdk_per_port_memory(void)
 bool
 dpdk_available(void)
 {
-    return dpdk_initialized;
+    bool initialized;
+
+    atomic_read_relaxed(&dpdk_initialized, &initialized);
+    return initialized;
 }
 
-void
-dpdk_set_lcore_id(unsigned cpu)
+bool
+dpdk_attach_thread(unsigned cpu)
 {
     /* NON_PMD_CORE_ID is reserved for use by non pmd threads. */
     ovs_assert(cpu != NON_PMD_CORE_ID);
-    RTE_PER_LCORE(_lcore_id) = cpu;
+
+    if (!dpdk_available()) {
+        return false;
+    }
+
+    if (rte_thread_register() < 0) {
+        VLOG_WARN("DPDK max threads count has been reached. "
+                  "PMD thread performance may be impacted.");
+        return false;
+    }
+
+    VLOG_INFO("PMD thread uses DPDK lcore %u.", rte_lcore_id());
+    return true;
+}
+
+void
+dpdk_detach_thread(void)
+{
+    unsigned int lcore_id;
+
+    lcore_id = rte_lcore_id();
+    rte_thread_unregister();
+    VLOG_INFO("PMD thread released DPDK lcore %u.", lcore_id);
 }
 
 void
@@ -614,40 +627,11 @@ print_dpdk_version(void)
     puts(rte_version());
 }
 
-#define CHECK_CPU_FEATURE(feature, name_str, RTE_CPUFLAG)               \
-    do {                                                                \
-        if (strncmp(feature, name_str, strlen(name_str)) == 0) {        \
-            int has_isa = rte_cpu_get_flag_enabled(RTE_CPUFLAG);        \
-            VLOG_DBG("CPU flag %s, available %s\n", name_str,           \
-                      has_isa ? "yes" : "no");                          \
-            return true;                                                \
-        }                                                               \
-    } while (0)
-
-bool
-dpdk_get_cpu_has_isa(const char *arch, const char *feature)
-{
-    /* Ensure Arch is x86_64. */
-    if (strncmp(arch, "x86_64", 6) != 0) {
-        return false;
-    }
-
-#if __x86_64__
-    /* CPU flags only defined for the architecture that support it. */
-    CHECK_CPU_FEATURE(feature, "avx512f", RTE_CPUFLAG_AVX512F);
-    CHECK_CPU_FEATURE(feature, "bmi2", RTE_CPUFLAG_BMI2);
-#endif
-
-    VLOG_WARN("Unknown CPU arch,feature: %s,%s. Returning not supported.\n",
-              arch, feature);
-    return false;
-}
-
 void
 dpdk_status(const struct ovsrec_open_vswitch *cfg)
 {
     if (cfg) {
-        ovsrec_open_vswitch_set_dpdk_initialized(cfg, dpdk_initialized);
+        ovsrec_open_vswitch_set_dpdk_initialized(cfg, dpdk_available());
         ovsrec_open_vswitch_set_dpdk_version(cfg, rte_version());
     }
 }
