@@ -17,9 +17,11 @@
 #include <config.h>
 #include <stdint.h>
 
+#include "dpdk.h"
 #include "dpif-netdev-perf.h"
 #include "openvswitch/dynamic-string.h"
 #include "openvswitch/vlog.h"
+#include "ovs-numa.h"
 #include "ovs-thread.h"
 #include "timeval.h"
 
@@ -43,21 +45,60 @@ uint64_t iter_cycle_threshold;
 
 static struct vlog_rate_limit latency_rl = VLOG_RATE_LIMIT_INIT(600, 600);
 
+static uint64_t tsc_hz = 1;
+
+void
+pmd_perf_estimate_tsc_frequency(void)
+{
 #ifdef DPDK_NETDEV
-static uint64_t
-get_tsc_hz(void)
-{
-    return rte_get_tsc_hz();
-}
-#else
-/* This function is only invoked from PMD threads which depend on DPDK.
- * A dummy function is sufficient when building without DPDK_NETDEV. */
-static uint64_t
-get_tsc_hz(void)
-{
-    return 1;
-}
+    if (dpdk_available()) {
+        tsc_hz = rte_get_tsc_hz();
+    }
+    if (tsc_hz > 1) {
+        VLOG_INFO("DPDK provided TSC frequency: %"PRIu64" KHz", tsc_hz / 1000);
+        return;
+    }
 #endif
+    struct ovs_numa_dump *affinity;
+    struct pmd_perf_stats s;
+    uint64_t start, stop;
+
+    /* DPDK is not available or returned unreliable value.
+     * Trying to estimate. */
+    affinity = ovs_numa_thread_getaffinity_dump();
+    if (affinity) {
+        const struct ovs_numa_info_core *core;
+
+        FOR_EACH_CORE_ON_DUMP (core, affinity) {
+            /* Setting affinity to a single core from the affinity mask to
+             * avoid re-scheduling to another core while sleeping. */
+            ovs_numa_thread_setaffinity_core(core->core_id);
+            break;
+        }
+    }
+
+    memset(&s, 0, sizeof s);
+    start = cycles_counter_update(&s);
+    /* Using xnanosleep as it's interrupt resistant.
+     * Sleeping only 100 ms to avoid holding the main thread for too long. */
+    xnanosleep(1E8);
+    stop = cycles_counter_update(&s);
+
+    if (affinity) {
+        /* Restoring previous affinity. */
+        ovs_numa_thread_setaffinity_dump(affinity);
+        ovs_numa_dump_destroy(affinity);
+    }
+
+    if (stop <= start) {
+        VLOG_WARN("TSC source is unreliable.");
+        tsc_hz = 1;
+    } else {
+        tsc_hz = (stop - start) * 10;
+    }
+
+    VLOG_INFO("Estimated TSC frequency: %"PRIu64" KHz", tsc_hz / 1000);
+}
 
 /* Histogram functions. */
 
@@ -170,7 +211,6 @@ pmd_perf_format_overall_stats(struct ds *str, struct pmd_perf_stats *s,
                               double duration)
 {
     uint64_t stats[PMD_N_STATS];
-    uint64_t tsc_hz = get_tsc_hz();
     double us_per_cycle = 1000000.0 / tsc_hz;
 
     if (duration == 0) {
@@ -192,10 +232,10 @@ pmd_perf_format_overall_stats(struct ds *str, struct pmd_perf_stats *s,
     uint64_t busy_iter = tot_iter >= idle_iter ? tot_iter - idle_iter : 0;
 
     ds_put_format(str,
-            "  Iterations:        %12"PRIu64"  (%.2f us/it)\n"
-            "  - Used TSC cycles: %12"PRIu64"  (%5.1f %% of total cycles)\n"
-            "  - idle iterations: %12"PRIu64"  (%5.1f %% of used cycles)\n"
-            "  - busy iterations: %12"PRIu64"  (%5.1f %% of used cycles)\n",
+            "  Iterations:         %12"PRIu64"  (%.2f us/it)\n"
+            "  - Used TSC cycles:  %12"PRIu64"  (%5.1f %% of total cycles)\n"
+            "  - idle iterations:  %12"PRIu64"  (%5.1f %% of used cycles)\n"
+            "  - busy iterations:  %12"PRIu64"  (%5.1f %% of used cycles)\n",
             tot_iter, tot_cycles * us_per_cycle / tot_iter,
             tot_cycles, 100.0 * (tot_cycles / duration) / tsc_hz,
             idle_iter,
@@ -204,17 +244,26 @@ pmd_perf_format_overall_stats(struct ds *str, struct pmd_perf_stats *s,
             100.0 * stats[PMD_CYCLES_ITER_BUSY] / tot_cycles);
     if (rx_packets > 0) {
         ds_put_format(str,
-            "  Rx packets:        %12"PRIu64"  (%.0f Kpps, %.0f cycles/pkt)\n"
-            "  Datapath passes:   %12"PRIu64"  (%.2f passes/pkt)\n"
-            "  - EMC hits:        %12"PRIu64"  (%5.1f %%)\n"
-            "  - SMC hits:        %12"PRIu64"  (%5.1f %%)\n"
-            "  - Megaflow hits:   %12"PRIu64"  (%5.1f %%, %.2f "
-                                                "subtbl lookups/hit)\n"
-            "  - Upcalls:         %12"PRIu64"  (%5.1f %%, %.1f us/upcall)\n"
-            "  - Lost upcalls:    %12"PRIu64"  (%5.1f %%)\n",
+            "  Rx packets:         %12"PRIu64"  (%.0f Kpps, %.0f cycles/pkt)\n"
+            "  Datapath passes:    %12"PRIu64"  (%.2f passes/pkt)\n"
+            "  - PHWOL hits:       %12"PRIu64"  (%5.1f %%)\n"
+            "  - MFEX Opt hits:    %12"PRIu64"  (%5.1f %%)\n"
+            "  - Simple Match hits:%12"PRIu64"  (%5.1f %%)\n"
+            "  - EMC hits:         %12"PRIu64"  (%5.1f %%)\n"
+            "  - SMC hits:         %12"PRIu64"  (%5.1f %%)\n"
+            "  - Megaflow hits:    %12"PRIu64"  (%5.1f %%, %.2f "
+                                                 "subtbl lookups/hit)\n"
+            "  - Upcalls:          %12"PRIu64"  (%5.1f %%, %.1f us/upcall)\n"
+            "  - Lost upcalls:     %12"PRIu64"  (%5.1f %%)\n",
             rx_packets, (rx_packets / duration) / 1000,
             1.0 * stats[PMD_CYCLES_ITER_BUSY] / rx_packets,
             passes, rx_packets ? 1.0 * passes / rx_packets : 0,
+            stats[PMD_STAT_PHWOL_HIT],
+            100.0 * stats[PMD_STAT_PHWOL_HIT] / passes,
+            stats[PMD_STAT_MFEX_OPT_HIT],
+            100.0 * stats[PMD_STAT_MFEX_OPT_HIT] / passes,
+            stats[PMD_STAT_SIMPLE_HIT],
+            100.0 * stats[PMD_STAT_SIMPLE_HIT] / passes,
             stats[PMD_STAT_EXACT_HIT],
             100.0 * stats[PMD_STAT_EXACT_HIT] / passes,
             stats[PMD_STAT_SMC_HIT],
@@ -229,16 +278,18 @@ pmd_perf_format_overall_stats(struct ds *str, struct pmd_perf_stats *s,
             stats[PMD_STAT_LOST],
             100.0 * stats[PMD_STAT_LOST] / passes);
     } else {
-        ds_put_format(str, "  Rx packets:        %12d\n", 0);
+        ds_put_format(str,
+            "  Rx packets:         %12d\n", 0);
     }
     if (tx_packets > 0) {
         ds_put_format(str,
-            "  Tx packets:        %12"PRIu64"  (%.0f Kpps)\n"
-            "  Tx batches:        %12"PRIu64"  (%.2f pkts/batch)\n",
+            "  Tx packets:         %12"PRIu64"  (%.0f Kpps)\n"
+            "  Tx batches:         %12"PRIu64"  (%.2f pkts/batch)\n",
             tx_packets, (tx_packets / duration) / 1000,
             tx_batches, 1.0 * tx_packets / tx_batches);
     } else {
-        ds_put_format(str, "  Tx packets:        %12d\n\n", 0);
+        ds_put_format(str,
+            "  Tx packets:         %12d\n\n", 0);
     }
 }
 
@@ -498,15 +549,7 @@ pmd_perf_end_iteration(struct pmd_perf_stats *s, int rx_packets,
         cycles_per_pkt = cycles / rx_packets;
         histogram_add_sample(&s->cycles_per_pkt, cycles_per_pkt);
     }
-    if (s->current.batches > 0) {
-        histogram_add_sample(&s->pkts_per_batch,
-                             rx_packets / s->current.batches);
-    }
     histogram_add_sample(&s->upcalls, s->current.upcalls);
-    if (s->current.upcalls > 0) {
-        histogram_add_sample(&s->cycles_per_upcall,
-                             s->current.upcall_cycles / s->current.upcalls);
-    }
     histogram_add_sample(&s->max_vhost_qfill, s->current.max_vhost_qfill);
 
     /* Add iteration samples to millisecond stats. */
@@ -562,8 +605,8 @@ pmd_perf_end_iteration(struct pmd_perf_stats *s, int rx_packets,
             cum_ms = history_next(&s->milliseconds);
             cum_ms->timestamp = now;
         }
-        /* Do the next check after 10K cycles (4 us at 2.5 GHz TSC clock). */
-        s->next_check_tsc = cycles_counter_update(s) + 10000;
+        /* Do the next check after 4 us (10K cycles at 2.5 GHz TSC clock). */
+        s->next_check_tsc = cycles_counter_update(s) + tsc_hz / 250000;
     }
 }
 
@@ -593,7 +636,7 @@ pmd_perf_set_log_susp_iteration(struct pmd_perf_stats *s,
                 " duration=%"PRIu64" us\n",
                 s->log_reason,
                 susp->timestamp,
-                (1000000L * susp->cycles) / get_tsc_hz());
+                (1000000L * susp->cycles) / tsc_hz);
 
         new_end_it = history_add(s->iterations.idx, log_it_after + 1);
         new_range = history_sub(new_end_it, s->log_begin_it);
@@ -623,7 +666,7 @@ pmd_perf_log_susp_iteration_neighborhood(struct pmd_perf_stats *s)
                  " duration=%"PRIu64" us\n",
                  s->log_reason,
                  susp->timestamp,
-                 (1000000L * susp->cycles) / get_tsc_hz());
+                 (1000000L * susp->cycles) / tsc_hz);
 
     pmd_perf_format_iteration_history(&log, s, range);
     VLOG_WARN_RL(&latency_rl,
@@ -737,7 +780,7 @@ pmd_perf_log_set_cmd(struct unixctl_conn *conn,
     log_it_after = it_after;
     log_q_thr = q_thr;
     log_us_thr = us_thr;
-    iter_cycle_threshold = (log_us_thr * get_tsc_hz()) / 1000000L;
+    iter_cycle_threshold = (log_us_thr * tsc_hz) / 1000000L;
 
     unixctl_command_reply(conn, "");
 }
