@@ -24,22 +24,18 @@
 #include <string.h>
 
 #include "coverage.h"
-#include "dpctl.h"
 #include "dp-packet.h"
+#include "dpctl.h"
 #include "dpif-netdev.h"
-#include "openvswitch/dynamic-string.h"
 #include "flow.h"
+#include "netdev-provider.h"
 #include "netdev.h"
 #include "netlink.h"
 #include "odp-execute.h"
 #include "odp-util.h"
-#include "openvswitch/ofp-print.h"
-#include "openvswitch/ofpbuf.h"
 #include "packets.h"
-#include "openvswitch/poll-loop.h"
 #include "route-table.h"
 #include "seq.h"
-#include "openvswitch/shash.h"
 #include "sset.h"
 #include "timeval.h"
 #include "tnl-neigh-cache.h"
@@ -47,9 +43,14 @@
 #include "util.h"
 #include "uuid.h"
 #include "valgrind.h"
+#include "openvswitch/dynamic-string.h"
 #include "openvswitch/ofp-errors.h"
+#include "openvswitch/ofp-print.h"
+#include "openvswitch/ofpbuf.h"
+#include "openvswitch/poll-loop.h"
+#include "openvswitch/shash.h"
+#include "openvswitch/usdt-probes.h"
 #include "openvswitch/vlog.h"
-#include "lib/netdev-provider.h"
 
 VLOG_DEFINE_THIS_MODULE(dpif);
 
@@ -364,7 +365,8 @@ do_open(const char *name, const char *type, bool create, struct dpif **dpifp)
             err = netdev_open(dpif_port.name, dpif_port.type, &netdev);
 
             if (!err) {
-                netdev_ports_insert(netdev, dpif_type_str, &dpif_port);
+                netdev_set_dpif_type(netdev, dpif_type_str);
+                netdev_ports_insert(netdev, &dpif_port);
                 netdev_close(netdev);
             } else {
                 VLOG_WARN("could not open netdev %s type %s: %s",
@@ -602,10 +604,12 @@ dpif_port_add(struct dpif *dpif, struct netdev *netdev, odp_port_t *port_nop)
             const char *dpif_type_str = dpif_normalize_type(dpif_type(dpif));
             struct dpif_port dpif_port;
 
+            netdev_set_dpif_type(netdev, dpif_type_str);
+
             dpif_port.type = CONST_CAST(char *, netdev_get_type(netdev));
             dpif_port.name = CONST_CAST(char *, netdev_name);
             dpif_port.port_no = port_no;
-            netdev_ports_insert(netdev, dpif_type_str, &dpif_port);
+            netdev_ports_insert(netdev, &dpif_port);
         }
     } else {
         VLOG_WARN_RL(&error_rl, "%s: failed to add %s as port: %s",
@@ -697,13 +701,14 @@ dpif_port_set_config(struct dpif *dpif, odp_port_t port_no,
  * initializes '*port' appropriately; on failure, returns a positive errno
  * value.
  *
- * Retuns ENODEV if the port doesn't exist.
+ * Retuns ENODEV if the port doesn't exist.  Will not log a warning in this
+ * case unless 'warn_if_not_found' is true.
  *
  * The caller owns the data in 'port' and must free it with
  * dpif_port_destroy() when it is no longer needed. */
 int
 dpif_port_query_by_number(const struct dpif *dpif, odp_port_t port_no,
-                          struct dpif_port *port)
+                          struct dpif_port *port, bool warn_if_not_found)
 {
     int error = dpif->dpif_class->port_query_by_number(dpif, port_no, port);
     if (!error) {
@@ -711,8 +716,13 @@ dpif_port_query_by_number(const struct dpif *dpif, odp_port_t port_no,
                     dpif_name(dpif), port_no, port->name);
     } else {
         memset(port, 0, sizeof *port);
-        VLOG_WARN_RL(&error_rl, "%s: failed to query port %"PRIu32": %s",
-                     dpif_name(dpif), port_no, ovs_strerror(error));
+        if (error == ENODEV && !warn_if_not_found) {
+            VLOG_DBG_RL(&dpmsg_rl, "%s: failed to query port %"PRIu32": %s",
+                        dpif_name(dpif), port_no, ovs_strerror(error));
+        } else {
+            VLOG_WARN_RL(&error_rl, "%s: failed to query port %"PRIu32": %s",
+                         dpif_name(dpif), port_no, ovs_strerror(error));
+        }
     }
     return error;
 }
@@ -780,7 +790,7 @@ dpif_port_get_name(struct dpif *dpif, odp_port_t port_no,
 
     ovs_assert(name_size > 0);
 
-    error = dpif_port_query_by_number(dpif, port_no, &port);
+    error = dpif_port_query_by_number(dpif, port_no, &port, true);
     if (!error) {
         ovs_strlcpy(name, port.name, name_size);
         dpif_port_destroy(&port);
@@ -1209,7 +1219,7 @@ dpif_execute_helper_cb(void *aux_, struct dp_packet_batch *packets_,
             /* The Linux kernel datapath throws away the tunnel information
              * that we supply as metadata.  We have to use a "set" action to
              * supply it. */
-            if (md->tunnel.ip_dst) {
+            if (flow_tnl_dst_is_set(&md->tunnel)) {
                 odp_put_tunnel_action(&md->tunnel, &execute_actions, NULL);
             }
             ofpbuf_put(&execute_actions, action, NLA_ALIGN(action->nla_len));
@@ -1274,6 +1284,7 @@ dpif_execute_helper_cb(void *aux_, struct dp_packet_batch *packets_,
     case OVS_ACTION_ATTR_UNSPEC:
     case OVS_ACTION_ATTR_CHECK_PKT_LEN:
     case OVS_ACTION_ATTR_DROP:
+    case OVS_ACTION_ATTR_ADD_MPLS:
     case __OVS_ACTION_ATTR_MAX:
         OVS_NOT_REACHED();
     }
@@ -1427,6 +1438,14 @@ dpif_operate(struct dpif *dpif, struct dpif_op **ops, size_t n_ops,
     }
 }
 
+int dpif_offload_stats_get(struct dpif *dpif,
+                           struct netdev_custom_stats *stats)
+{
+    return (dpif->dpif_class->offload_stats_get
+            ? dpif->dpif_class->offload_stats_get(dpif, stats)
+            : EOPNOTSUPP);
+}
+
 /* Returns a string that represents 'type', for use in log messages. */
 const char *
 dpif_upcall_type_to_string(enum dpif_upcall_type type)
@@ -1489,6 +1508,22 @@ dpif_handlers_set(struct dpif *dpif, uint32_t n_handlers)
     return error;
 }
 
+/* Checks if a certain number of handlers are required.
+ *
+ * If a certain number of handlers are required, returns 'true' and sets
+ * 'n_handlers' to that number of handler threads.
+ *
+ * If not, returns 'false'
+ */
+bool
+dpif_number_handlers_required(struct dpif *dpif, uint32_t *n_handlers)
+{
+    if (dpif->dpif_class->number_handlers_required) {
+        return dpif->dpif_class->number_handlers_required(dpif, n_handlers);
+    }
+    return false;
+}
+
 void
 dpif_register_dp_purge_cb(struct dpif *dpif, dp_purge_callback *cb, void *aux)
 {
@@ -1586,6 +1621,12 @@ dpif_recv(struct dpif *dpif, uint32_t handler_id, struct dpif_upcall *upcall,
     if (dpif->dpif_class->recv) {
         error = dpif->dpif_class->recv(dpif, handler_id, upcall, buf);
         if (!error) {
+            OVS_USDT_PROBE(dpif_recv, recv_upcall, dpif->full_name,
+                           upcall->type,
+                           dp_packet_data(&upcall->packet),
+                           dp_packet_size(&upcall->packet),
+                           upcall->key, upcall->key_len);
+
             dpif_print_packet(dpif, upcall);
         } else if (error != EAGAIN) {
             log_operation(dpif, "recv", error);
@@ -1996,7 +2037,7 @@ dpif_meter_del(struct dpif *dpif, ofproto_meter_id meter_id,
 int
 dpif_bond_add(struct dpif *dpif, uint32_t bond_id, odp_port_t *member_map)
 {
-    return dpif->dpif_class->bond_del
+    return dpif->dpif_class->bond_add
            ? dpif->dpif_class->bond_add(dpif, bond_id, member_map)
            : EOPNOTSUPP;
 }
@@ -2042,3 +2083,41 @@ dpif_get_n_offloaded_flows(struct dpif *dpif, uint64_t *n_flows)
     }
     return n_devs ? 0 : EOPNOTSUPP;
 }
+
+int
+dpif_cache_get_supported_levels(struct dpif *dpif, uint32_t *levels)
+{
+    return dpif->dpif_class->cache_get_supported_levels
+           ? dpif->dpif_class->cache_get_supported_levels(dpif, levels)
+           : EOPNOTSUPP;
+}
+
+int
+dpif_cache_get_name(struct dpif *dpif, uint32_t level, const char **name)
+{
+    return dpif->dpif_class->cache_get_name
+           ? dpif->dpif_class->cache_get_name(dpif, level, name)
+           : EOPNOTSUPP;
+}
+
+int
+dpif_cache_get_size(struct dpif *dpif, uint32_t level, uint32_t *size)
+{
+    return dpif->dpif_class->cache_get_size
+           ? dpif->dpif_class->cache_get_size(dpif, level, size)
+           : EOPNOTSUPP;
+}
+
+int
+dpif_cache_set_size(struct dpif *dpif, uint32_t level, uint32_t size)
+{
+    return dpif->dpif_class->cache_set_size
+           ? dpif->dpif_class->cache_set_size(dpif, level, size)
+           : EOPNOTSUPP;
+}
+
+bool
+dpif_synced_dp_layers(struct dpif *dpif)
+{
+    return dpif->dpif_class->synced_dp_layers;
+}
