@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010, 2012, 2013, 2015 Nicira, Inc.
+ * Copyright (c) 2008, 2009, 2010, 2012, 2013, 2015, 2019 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -15,7 +15,7 @@
  */
 
 #include <config.h>
-#include "hmap.h"
+#include "openvswitch/hmap.h"
 #include <stdint.h>
 #include <string.h>
 #include "coverage.h"
@@ -103,6 +103,9 @@ resize(struct hmap *hmap, size_t new_mask, const char *where)
             tmp.buckets[i] = NULL;
         }
     }
+    int n_big_buckets = 0;
+    int biggest_count = 0;
+    int n_biggest_buckets = 0;
     for (i = 0; i <= hmap->mask; i++) {
         struct hmap_node *node, *next;
         int count = 0;
@@ -112,14 +115,30 @@ resize(struct hmap *hmap, size_t new_mask, const char *where)
             count++;
         }
         if (count > 5) {
-            static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(10, 10);
-            COVERAGE_INC(hmap_pathological);
-            VLOG_DBG_RL(&rl, "%s: %d nodes in bucket (%"PRIuSIZE" nodes, %"PRIuSIZE" buckets)",
-                        where, count, hmap->n, hmap->mask + 1);
+            n_big_buckets++;
+            if (count > biggest_count) {
+                biggest_count = count;
+                n_biggest_buckets = 1;
+            } else if (count == biggest_count) {
+                n_biggest_buckets++;
+            }
         }
     }
     hmap_swap(hmap, &tmp);
     hmap_destroy(&tmp);
+
+    if (n_big_buckets) {
+        static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(10, 10);
+        COVERAGE_INC(hmap_pathological);
+        VLOG_DBG_RL(&rl, "%s: %d bucket%s with 6+ nodes, "
+                    "including %d bucket%s with %d nodes "
+                    "(%"PRIuSIZE" nodes total across %"PRIuSIZE" buckets)",
+                    where,
+                    n_big_buckets, n_big_buckets > 1 ? "s" : "",
+                    n_biggest_buckets, n_biggest_buckets > 1 ? "s" : "",
+                    biggest_count,
+                    hmap->n, hmap->mask + 1);
+    }
 }
 
 static size_t
@@ -236,24 +255,22 @@ hmap_random_node(const struct hmap *hmap)
 }
 
 /* Returns the next node in 'hmap' in hash order, or NULL if no nodes remain in
- * 'hmap'.  Uses '*bucketp' and '*offsetp' to determine where to begin
- * iteration, and stores new values to pass on the next iteration into them
- * before returning.
+ * 'hmap'.  Uses '*pos' to determine where to begin iteration, and updates
+ * '*pos' to pass on the next iteration into them before returning.
  *
  * It's better to use plain HMAP_FOR_EACH and related functions, since they are
  * faster and better at dealing with hmaps that change during iteration.
  *
- * Before beginning iteration, store 0 into '*bucketp' and '*offsetp'.
- */
+ * Before beginning iteration, set '*pos' to all zeros. */
 struct hmap_node *
 hmap_at_position(const struct hmap *hmap,
-                 uint32_t *bucketp, uint32_t *offsetp)
+                 struct hmap_position *pos)
 {
     size_t offset;
     size_t b_idx;
 
-    offset = *offsetp;
-    for (b_idx = *bucketp; b_idx <= hmap->mask; b_idx++) {
+    offset = pos->offset;
+    for (b_idx = pos->bucket; b_idx <= hmap->mask; b_idx++) {
         struct hmap_node *node;
         size_t n_idx;
 
@@ -261,11 +278,11 @@ hmap_at_position(const struct hmap *hmap,
              n_idx++, node = node->next) {
             if (n_idx == offset) {
                 if (node->next) {
-                    *bucketp = node->hash & hmap->mask;
-                    *offsetp = offset + 1;
+                    pos->bucket = node->hash & hmap->mask;
+                    pos->offset = offset + 1;
                 } else {
-                    *bucketp = (node->hash & hmap->mask) + 1;
-                    *offsetp = 0;
+                    pos->bucket = (node->hash & hmap->mask) + 1;
+                    pos->offset = 0;
                 }
                 return node;
             }
@@ -273,8 +290,8 @@ hmap_at_position(const struct hmap *hmap,
         offset = 0;
     }
 
-    *bucketp = 0;
-    *offsetp = 0;
+    pos->bucket = 0;
+    pos->offset = 0;
     return NULL;
 }
 
