@@ -27,16 +27,6 @@
 #include "tun-metadata.h"
 #include "openvswitch/nsh.h"
 
-static void
-set_eth(const struct eth_addr value_src,
-        struct eth_addr *value_dst,
-        struct eth_addr *mask_dst);
-
-static void
-set_eth_masked(const struct eth_addr value_src,
-               const struct eth_addr mask_src,
-               struct eth_addr *value_dst, struct eth_addr *mask_dst);
-
 /* Converts the flow in 'flow' into a match in 'match', with the given
  * 'wildcards'. */
 void
@@ -412,50 +402,6 @@ match_set_tun_gtpu_msgtype(struct match *match, uint8_t msgtype)
     match_set_tun_gtpu_msgtype_masked(match, msgtype, UINT8_MAX);
 }
 
-void
-match_set_tun_eth_src(struct match *match, const struct eth_addr dl_src)
-{
-    set_eth(dl_src, &match->flow.tunnel.eth_src, &match->wc.masks.tunnel.eth_src);
-}
-
-void
-match_set_tun_eth_src_masked(struct match *match,
-                             const struct eth_addr dl_src,
-                             const struct eth_addr mask)
-{
-    set_eth_masked(dl_src, mask, &match->flow.tunnel.eth_src,
-                   &match->wc.masks.tunnel.eth_src);
-}
-
-void
-match_set_tun_eth_dst(struct match *match, const struct eth_addr dl_dst)
-{
-    set_eth(dl_dst, &match->flow.tunnel.eth_dst, &match->wc.masks.tunnel.eth_dst);
-}
-
-void
-match_set_tun_eth_dst_masked(struct match *match,
-                             const struct eth_addr dl_dst,
-                             const struct eth_addr mask)
-{
-    set_eth_masked(dl_dst, mask, &match->flow.tunnel.eth_dst,
-                   &match->wc.masks.tunnel.eth_dst);
-}
-
-void
-match_set_tun_type(struct match *match, uint8_t type)
-{
-    match->wc.masks.tunnel.type = UINT8_MAX;
-    match->flow.tunnel.type = type;
-}
-
-void
-match_set_tun_type_masked(struct match *match, uint8_t type, uint8_t mask)
-{
-    match->wc.masks.tunnel.type = mask;
-    match->flow.tunnel.type = type;
-}
-
 void
 match_set_in_port(struct match *match, ofp_port_t ofp_port)
 {
@@ -1452,12 +1398,6 @@ format_flow_tunnel(struct ds *s, const struct match *match)
                             FLOW_TNL_F_MASK);
         ds_put_char(s, ',');
     }
-    format_eth_masked(s, "tun_eth_src", tnl->eth_src, wc->masks.tunnel.eth_src);
-    format_eth_masked(s, "tun_eth_dst", tnl->eth_dst, wc->masks.tunnel.eth_dst);
-    if (wc->masks.tunnel.type) {
-        ds_put_format(s, "tun_type=%s,",
-                      tun_type_type2str((enum tun_type) tnl->type));
-    }
     tun_metadata_match_format(s, match);
 }
 
