@@ -23,12 +23,12 @@
 #include <fcntl.h>
 #include <sys/socket.h>
 #include <net/if.h>
+#include <sys/types.h>
 #include <netinet/in.h>
 #include <netinet/ip.h>
 #include <netinet/ip6.h>
 #include <sys/ioctl.h>
 
-#include <errno.h>
 #include <stdlib.h>
 #include <sys/time.h>
 
@@ -54,6 +54,9 @@ static struct vlog_rate_limit err_rl = VLOG_RATE_LIMIT_INIT(60, 5);
 #define GENEVE_BASE_HLEN   (sizeof(struct udp_header) +         \
                             sizeof(struct genevehdr))
 
+#define GTPU_HLEN   (sizeof(struct udp_header) +        \
+                     sizeof(struct gtpuhdr))
+
 uint16_t tnl_udp_port_min = 32768;
 uint16_t tnl_udp_port_max = 61000;
 
@@ -122,7 +125,7 @@ netdev_tnl_ip_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
         tnl->ip_tos = ntohl(tc_flow) >> 20;
         tnl->ip_ttl = ip6->ip6_hlim;
 
-        *hlen += IPV6_HEADER_LEN;
+        *hlen += packet->l4_ofs - packet->l3_ofs;
 
     } else {
         VLOG_WARN_RL(&err_rl, "ipv4 packet has invalid version (%d)",
@@ -212,9 +215,31 @@ udp_extract_tnl_md(struct dp_packet *packet, struct flow_tnl *tnl,
     return udp + 1;
 }
 
+static void
+netdev_tnl_calc_udp_csum(struct udp_header *udp, struct dp_packet *packet,
+                         int ip_tot_size)
+{
+    uint32_t csum;
+
+    if (netdev_tnl_is_header_ipv6(dp_packet_data(packet))) {
+        csum = packet_csum_pseudoheader6(netdev_tnl_ipv6_hdr(
+                                         dp_packet_data(packet)));
+    } else {
+        csum = packet_csum_pseudoheader(netdev_tnl_ip_hdr(
+                                        dp_packet_data(packet)));
+    }
+
+    csum = csum_continue(csum, udp, ip_tot_size);
+    udp->udp_csum = csum_finish(csum);
+
+    if (!udp->udp_csum) {
+        udp->udp_csum = htons(0xffff);
+    }
+}
 
 void
-netdev_tnl_push_udp_header(struct dp_packet *packet,
+netdev_tnl_push_udp_header(const struct netdev *netdev OVS_UNUSED,
+                           struct dp_packet *packet,
                            const struct ovs_action_push_tnl *data)
 {
     struct udp_header *udp;
@@ -227,19 +252,7 @@ netdev_tnl_push_udp_header(struct dp_packet *packet,
     udp->udp_len = htons(ip_tot_size);
 
     if (udp->udp_csum) {
-        uint32_t csum;
-        if (netdev_tnl_is_header_ipv6(dp_packet_data(packet))) {
-            csum = packet_csum_pseudoheader6(netdev_tnl_ipv6_hdr(dp_packet_data(packet)));
-        } else {
-            csum = packet_csum_pseudoheader(netdev_tnl_ip_hdr(dp_packet_data(packet)));
-        }
-
-        csum = csum_continue(csum, udp, ip_tot_size);
-        udp->udp_csum = csum_finish(csum);
-
-        if (!udp->udp_csum) {
-            udp->udp_csum = htons(0xffff);
-        }
+        netdev_tnl_calc_udp_csum(udp, packet, ip_tot_size);
     }
 }
 
@@ -307,7 +320,7 @@ netdev_tnl_ip_build_header(struct ovs_action_push_tnl *data,
 }
 
 static void *
-udp_build_header(struct netdev_tunnel_config *tnl_cfg,
+udp_build_header(const struct netdev_tunnel_config *tnl_cfg,
                  struct ovs_action_push_tnl *data,
                  const struct netdev_tnl_build_header_params *params)
 {
@@ -434,9 +447,11 @@ err:
 }
 
 void
-netdev_gre_push_header(struct dp_packet *packet,
+netdev_gre_push_header(const struct netdev *netdev,
+                       struct dp_packet *packet,
                        const struct ovs_action_push_tnl *data)
 {
+    struct netdev_vport *dev = netdev_vport_cast(netdev);
     struct gre_base_hdr *greh;
     int ip_tot_size;
 
@@ -446,6 +461,14 @@ netdev_gre_push_header(struct dp_packet *packet,
         ovs_be16 *csum_opt = (ovs_be16 *) (greh + 1);
         *csum_opt = csum(greh, ip_tot_size);
     }
+
+    if (greh->flags & htons(GRE_SEQ)) {
+        /* Last 4 byte is GRE seqno */
+        int seq_ofs = gre_header_len(greh->flags) - 4;
+        ovs_16aligned_be32 *seq_opt =
+            ALIGNED_CAST(ovs_16aligned_be32 *, (char *)greh + seq_ofs);
+        put_16aligned_be32(seq_opt, htonl(atomic_count_inc(&dev->gre_seqno)));
+    }
 }
 
 int
@@ -453,16 +476,11 @@ netdev_gre_build_header(const struct netdev *netdev,
                         struct ovs_action_push_tnl *data,
                         const struct netdev_tnl_build_header_params *params)
 {
-    struct netdev_vport *dev = netdev_vport_cast(netdev);
-    struct netdev_tunnel_config *tnl_cfg;
+    const struct netdev_tunnel_config *tnl_cfg;
     struct gre_base_hdr *greh;
     ovs_16aligned_be32 *options;
     unsigned int hlen;
 
-    /* XXX: RCUfy tnl_cfg. */
-    ovs_mutex_lock(&dev->mutex);
-    tnl_cfg = &dev->tnl_cfg;
-
     greh = netdev_tnl_ip_build_header(data, params, IPPROTO_GRE);
 
     if (params->flow->packet_type == htonl(PT_ETH)) {
@@ -470,8 +488,7 @@ netdev_gre_build_header(const struct netdev *netdev,
     } else if (pt_ns(params->flow->packet_type) == OFPHTN_ETHERTYPE) {
         greh->protocol = pt_ns_type_be(params->flow->packet_type);
     } else {
-        ovs_mutex_unlock(&dev->mutex);
-        return 1;
+        return EINVAL;
     }
     greh->flags = 0;
 
@@ -482,18 +499,329 @@ netdev_gre_build_header(const struct netdev *netdev,
         options++;
     }
 
+    tnl_cfg = netdev_get_tunnel_config(netdev);
+
     if (tnl_cfg->out_key_present) {
         greh->flags |= htons(GRE_KEY);
         put_16aligned_be32(options, be64_to_be32(params->flow->tunnel.tun_id));
         options++;
     }
 
-    ovs_mutex_unlock(&dev->mutex);
+    if (tnl_cfg->set_seq) {
+        greh->flags |= htons(GRE_SEQ);
+        /* seqno is updated at push header */
+        options++;
+    }
 
     hlen = (uint8_t *) options - (uint8_t *) greh;
 
     data->header_len += hlen;
-    data->tnl_type = OVS_VPORT_TYPE_GRE;
+    if (!params->is_ipv6) {
+        data->tnl_type = OVS_VPORT_TYPE_GRE;
+    } else {
+        data->tnl_type = OVS_VPORT_TYPE_IP6GRE;
+    }
+    return 0;
+}
+
+struct dp_packet *
+netdev_erspan_pop_header(struct dp_packet *packet)
+{
+    const struct gre_base_hdr *greh;
+    const struct erspan_base_hdr *ersh;
+    struct pkt_metadata *md = &packet->md;
+    struct flow_tnl *tnl = &md->tunnel;
+    int hlen = sizeof(struct eth_header);
+    unsigned int ulen;
+    uint16_t greh_protocol;
+
+    hlen += netdev_tnl_is_header_ipv6(dp_packet_data(packet)) ?
+            IPV6_HEADER_LEN : IP_HEADER_LEN;
+
+    pkt_metadata_init_tnl(md);
+    if (hlen > dp_packet_size(packet)) {
+        goto err;
+    }
+
+    greh = netdev_tnl_ip_extract_tnl_md(packet, tnl, &ulen);
+    if (!greh) {
+        goto err;
+    }
+
+    greh_protocol = ntohs(greh->protocol);
+    if (greh_protocol != ETH_TYPE_ERSPAN1 &&
+        greh_protocol != ETH_TYPE_ERSPAN2) {
+        goto err;
+    }
+
+    if (greh->flags & ~htons(GRE_SEQ)) {
+        goto err;
+    }
+
+    ersh = ERSPAN_HDR(greh);
+    tnl->tun_id = be16_to_be64(htons(get_sid(ersh)));
+    tnl->erspan_ver = ersh->ver;
+
+    if (ersh->ver == 1) {
+        ovs_16aligned_be32 *index = ALIGNED_CAST(ovs_16aligned_be32 *,
+                                                 ersh + 1);
+        tnl->erspan_idx = ntohl(get_16aligned_be32(index));
+        tnl->flags |= FLOW_TNL_F_KEY;
+        hlen = ulen + ERSPAN_GREHDR_LEN + sizeof *ersh + ERSPAN_V1_MDSIZE;
+    } else if (ersh->ver == 2) {
+        struct erspan_md2 *md2 = ALIGNED_CAST(struct erspan_md2 *, ersh + 1);
+        tnl->erspan_dir = md2->dir;
+        tnl->erspan_hwid = get_hwid(md2);
+        tnl->flags |= FLOW_TNL_F_KEY;
+        hlen = ulen + ERSPAN_GREHDR_LEN + sizeof *ersh + ERSPAN_V2_MDSIZE;
+    } else {
+        VLOG_WARN_RL(&err_rl, "ERSPAN version error %d", ersh->ver);
+        goto err;
+    }
+
+    if (hlen > dp_packet_size(packet)) {
+        goto err;
+    }
+
+    dp_packet_reset_packet(packet, hlen);
+
+    return packet;
+err:
+    dp_packet_delete(packet);
+    return NULL;
+}
+
+void
+netdev_erspan_push_header(const struct netdev *netdev,
+                          struct dp_packet *packet,
+                          const struct ovs_action_push_tnl *data)
+{
+    struct netdev_vport *dev = netdev_vport_cast(netdev);
+    struct erspan_base_hdr *ersh;
+    struct gre_base_hdr *greh;
+    struct erspan_md2 *md2;
+    int ip_tot_size;
+
+    greh = netdev_tnl_push_ip_header(packet, data->header,
+                                     data->header_len, &ip_tot_size);
+
+    /* update GRE seqno */
+    ovs_16aligned_be32 *seqno = (ovs_16aligned_be32 *) (greh + 1);
+    put_16aligned_be32(seqno, htonl(atomic_count_inc(&dev->gre_seqno)));
+
+    /* update v2 timestamp */
+    if (greh->protocol == htons(ETH_TYPE_ERSPAN2)) {
+        ersh = ERSPAN_HDR(greh);
+        md2 = ALIGNED_CAST(struct erspan_md2 *, ersh + 1);
+        put_16aligned_be32(&md2->timestamp, get_erspan_ts(ERSPAN_100US));
+    }
+}
+
+int
+netdev_erspan_build_header(const struct netdev *netdev,
+                           struct ovs_action_push_tnl *data,
+                           const struct netdev_tnl_build_header_params *params)
+{
+    const struct netdev_tunnel_config *tnl_cfg;
+    struct gre_base_hdr *greh;
+    struct erspan_base_hdr *ersh;
+    unsigned int hlen;
+    uint32_t tun_id;
+    int erspan_ver;
+    uint16_t sid;
+
+    greh = netdev_tnl_ip_build_header(data, params, IPPROTO_GRE);
+    ersh = ERSPAN_HDR(greh);
+
+    tun_id = ntohl(be64_to_be32(params->flow->tunnel.tun_id));
+    /* ERSPAN only has 10-bit session ID */
+    if (tun_id & ~ERSPAN_SID_MASK) {
+        return EINVAL;
+    } else {
+        sid = (uint16_t) tun_id;
+    }
+
+    tnl_cfg = netdev_get_tunnel_config(netdev);
+
+    if (tnl_cfg->erspan_ver_flow) {
+        erspan_ver = params->flow->tunnel.erspan_ver;
+    } else {
+        erspan_ver = tnl_cfg->erspan_ver;
+    }
+
+    if (erspan_ver == 1) {
+        greh->protocol = htons(ETH_TYPE_ERSPAN1);
+        greh->flags = htons(GRE_SEQ);
+        ersh->ver = 1;
+        set_sid(ersh, sid);
+
+        uint32_t erspan_idx = (tnl_cfg->erspan_idx_flow
+                          ? params->flow->tunnel.erspan_idx
+                          : tnl_cfg->erspan_idx);
+        put_16aligned_be32(ALIGNED_CAST(ovs_16aligned_be32 *, ersh + 1),
+                           htonl(erspan_idx));
+
+        hlen = ERSPAN_GREHDR_LEN + sizeof *ersh + ERSPAN_V1_MDSIZE;
+    } else if (erspan_ver == 2) {
+        struct erspan_md2 *md2 = ALIGNED_CAST(struct erspan_md2 *, ersh + 1);
+
+        greh->protocol = htons(ETH_TYPE_ERSPAN2);
+        greh->flags = htons(GRE_SEQ);
+        ersh->ver = 2;
+        set_sid(ersh, sid);
+
+        md2->sgt = 0; /* security group tag */
+        md2->gra = 0;
+        put_16aligned_be32(&md2->timestamp, 0);
+
+        if (tnl_cfg->erspan_hwid_flow) {
+            set_hwid(md2, params->flow->tunnel.erspan_hwid);
+        } else {
+            set_hwid(md2, tnl_cfg->erspan_hwid);
+        }
+
+        if (tnl_cfg->erspan_dir_flow) {
+            md2->dir = params->flow->tunnel.erspan_dir;
+        } else {
+            md2->dir = tnl_cfg->erspan_dir;
+        }
+
+        hlen = ERSPAN_GREHDR_LEN + sizeof *ersh + ERSPAN_V2_MDSIZE;
+    } else {
+        VLOG_WARN_RL(&err_rl, "ERSPAN version error %d", tnl_cfg->erspan_ver);
+        return EINVAL;
+    }
+
+    data->header_len += hlen;
+
+    if (params->is_ipv6) {
+        data->tnl_type = OVS_VPORT_TYPE_IP6ERSPAN;
+    } else {
+        data->tnl_type = OVS_VPORT_TYPE_ERSPAN;
+    }
+    return 0;
+}
+
+struct dp_packet *
+netdev_gtpu_pop_header(struct dp_packet *packet)
+{
+    struct pkt_metadata *md = &packet->md;
+    struct flow_tnl *tnl = &md->tunnel;
+    struct gtpuhdr *gtph;
+    unsigned int gtpu_hlen;
+    unsigned int hlen;
+
+    ovs_assert(packet->l3_ofs > 0);
+    ovs_assert(packet->l4_ofs > 0);
+
+    pkt_metadata_init_tnl(md);
+    if (GTPU_HLEN > dp_packet_l4_size(packet)) {
+        goto err;
+    }
+
+    gtph = udp_extract_tnl_md(packet, tnl, &hlen);
+    if (!gtph) {
+        goto err;
+    }
+
+    tnl->gtpu_flags = gtph->md.flags;
+    tnl->gtpu_msgtype = gtph->md.msgtype;
+    tnl->tun_id = be32_to_be64(get_16aligned_be32(&gtph->teid));
+
+    if (tnl->gtpu_msgtype == GTPU_MSGTYPE_GPDU) {
+        struct ip_header *ip;
+
+        if (gtph->md.flags & GTPU_S_MASK) {
+            gtpu_hlen = GTPU_HLEN + sizeof(struct gtpuhdr_opt);
+        } else {
+            gtpu_hlen = GTPU_HLEN;
+        }
+        ip = ALIGNED_CAST(struct ip_header *, (char *)gtph + gtpu_hlen);
+
+        if (IP_VER(ip->ip_ihl_ver) == 4) {
+            packet->packet_type = htonl(PT_IPV4);
+        } else if (IP_VER(ip->ip_ihl_ver) == 6) {
+            packet->packet_type = htonl(PT_IPV6);
+        } else {
+            VLOG_WARN_RL(&err_rl, "GTP-U: Receive non-IP packet.");
+        }
+        dp_packet_reset_packet(packet, hlen + gtpu_hlen);
+    } else {
+        /* non-GPDU GTP-U messages, ex: echo request, end marker.
+         * Users should redirect these packets to controller, or.
+         * any application that handles GTP-U messages, so keep
+         * the original packet.
+         */
+        packet->packet_type = htonl(PT_ETH);
+        VLOG_WARN_ONCE("Receive non-GPDU msgtype: %"PRIu8,
+                       gtph->md.msgtype);
+    }
+
+    return packet;
+
+err:
+    dp_packet_delete(packet);
+    return NULL;
+}
+
+void
+netdev_gtpu_push_header(const struct netdev *netdev,
+                        struct dp_packet *packet,
+                        const struct ovs_action_push_tnl *data)
+{
+    struct netdev_vport *dev = netdev_vport_cast(netdev);
+    struct udp_header *udp;
+    struct gtpuhdr *gtpuh;
+    int ip_tot_size;
+    unsigned int payload_len;
+
+    payload_len = dp_packet_size(packet);
+    udp = netdev_tnl_push_ip_header(packet, data->header,
+                                    data->header_len, &ip_tot_size);
+    udp->udp_src = netdev_tnl_get_src_port(packet);
+    udp->udp_len = htons(ip_tot_size);
+    netdev_tnl_calc_udp_csum(udp, packet, ip_tot_size);
+
+    gtpuh = ALIGNED_CAST(struct gtpuhdr *, udp + 1);
+
+    if (gtpuh->md.flags & GTPU_S_MASK) {
+        ovs_be16 *seqno = ALIGNED_CAST(ovs_be16 *, gtpuh + 1);
+        *seqno = htons(atomic_count_inc(&dev->gre_seqno));
+        payload_len += sizeof(struct gtpuhdr_opt);
+    }
+    gtpuh->len = htons(payload_len);
+}
+
+int
+netdev_gtpu_build_header(const struct netdev *netdev,
+                         struct ovs_action_push_tnl *data,
+                         const struct netdev_tnl_build_header_params *params)
+{
+    const struct netdev_tunnel_config *tnl_cfg;
+    struct gtpuhdr *gtph;
+    unsigned int gtpu_hlen;
+
+    tnl_cfg = netdev_get_tunnel_config(netdev);
+
+    gtph = udp_build_header(tnl_cfg, data, params);
+
+    /* Set to default if not set in flow. */
+    gtph->md.flags = params->flow->tunnel.gtpu_flags ?
+                     params->flow->tunnel.gtpu_flags : GTPU_FLAGS_DEFAULT;
+    gtph->md.msgtype = params->flow->tunnel.gtpu_msgtype ?
+                       params->flow->tunnel.gtpu_msgtype : GTPU_MSGTYPE_GPDU;
+    put_16aligned_be32(&gtph->teid,
+                       be64_to_be32(params->flow->tunnel.tun_id));
+
+    gtpu_hlen = sizeof *gtph;
+    if (tnl_cfg->set_seq) {
+        gtph->md.flags |= GTPU_S_MASK;
+        gtpu_hlen += sizeof(struct gtpuhdr_opt);
+    }
+
+    data->header_len += gtpu_hlen;
+    data->tnl_type = OVS_VPORT_TYPE_GTPU;
+
     return 0;
 }
 
@@ -507,6 +835,9 @@ netdev_vxlan_pop_header(struct dp_packet *packet)
     ovs_be32 vx_flags;
     enum packet_type next_pt = PT_ETH;
 
+    ovs_assert(packet->l3_ofs > 0);
+    ovs_assert(packet->l4_ofs > 0);
+
     pkt_metadata_init_tnl(md);
     if (VXLAN_HLEN > dp_packet_l4_size(packet)) {
         goto err;
@@ -569,13 +900,10 @@ netdev_vxlan_build_header(const struct netdev *netdev,
                           struct ovs_action_push_tnl *data,
                           const struct netdev_tnl_build_header_params *params)
 {
-    struct netdev_vport *dev = netdev_vport_cast(netdev);
-    struct netdev_tunnel_config *tnl_cfg;
+    const struct netdev_tunnel_config *tnl_cfg;
     struct vxlanhdr *vxh;
 
-    /* XXX: RCUfy tnl_cfg. */
-    ovs_mutex_lock(&dev->mutex);
-    tnl_cfg = &dev->tnl_cfg;
+    tnl_cfg = netdev_get_tunnel_config(netdev);
 
     vxh = udp_build_header(tnl_cfg, data, params);
 
@@ -600,10 +928,10 @@ netdev_vxlan_build_header(const struct netdev *netdev,
                 vxh->vx_gpe.next_protocol = VXLAN_GPE_NP_ETHERNET;
                 break;
             default:
-                goto drop;
+                return EINVAL;
             }
         } else {
-            goto drop;
+            return EINVAL;
         }
     } else {
         put_16aligned_be32(&vxh->vx_flags, htonl(VXLAN_FLAGS));
@@ -611,14 +939,9 @@ netdev_vxlan_build_header(const struct netdev *netdev,
                            htonl(ntohll(params->flow->tunnel.tun_id) << 8));
     }
 
-    ovs_mutex_unlock(&dev->mutex);
     data->header_len += sizeof *vxh;
     data->tnl_type = OVS_VPORT_TYPE_VXLAN;
     return 0;
-
-drop:
-    ovs_mutex_unlock(&dev->mutex);
-    return 1;
 }
 
 struct dp_packet *
@@ -682,22 +1005,14 @@ netdev_geneve_build_header(const struct netdev *netdev,
                            struct ovs_action_push_tnl *data,
                            const struct netdev_tnl_build_header_params *params)
 {
-    struct netdev_vport *dev = netdev_vport_cast(netdev);
-    struct netdev_tunnel_config *tnl_cfg;
     struct genevehdr *gnh;
     int opt_len;
     bool crit_opt;
 
-    /* XXX: RCUfy tnl_cfg. */
-    ovs_mutex_lock(&dev->mutex);
-    tnl_cfg = &dev->tnl_cfg;
-
-    gnh = udp_build_header(tnl_cfg, data, params);
+    gnh = udp_build_header(netdev_get_tunnel_config(netdev), data, params);
 
     put_16aligned_be32(&gnh->vni, htonl(ntohll(params->flow->tunnel.tun_id) << 8));
 
-    ovs_mutex_unlock(&dev->mutex);
-
     opt_len = tun_metadata_to_geneve_header(&params->flow->tunnel,
                                             gnh->options, &crit_opt);
 
