@@ -342,24 +342,6 @@ tunnel_get_status(const struct netdev *netdev_, struct smap *smap)
     return 0;
 }
 
-static int
-derive_tunnel_config(const struct netdev *upper_netdev,
-                     struct netdev *lower_netdev)
-{
-        const struct netdev_tunnel_config *upper_cfg;
-        struct netdev_tunnel_config *lower_cfg;
-
-        upper_cfg = netdev_get_tunnel_config(upper_netdev);
-        lower_cfg = CONST_CAST(struct netdev_tunnel_config *,
-                               netdev_get_tunnel_config(lower_netdev));
-        if (upper_cfg == NULL || lower_cfg == NULL) {
-            return EOPNOTSUPP;
-        }
-
-        lower_cfg->is_explicit = upper_cfg->is_explicit;
-        return 0;
-}
-
 static int
 netdev_vport_update_flags(struct netdev *netdev OVS_UNUSED,
                           enum netdev_flags off,
@@ -812,8 +794,6 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args, char **errp)
                use them, but they are read by 'ovs-monitor-ipsec'. In order to
                suppress the "unknown %s argument" warning message below, we
                handle them here by ignoring them. */
-        } else if (!strcmp(node->key, "explicit")) {
-            tnl_cfg.is_explicit = !strcasecmp("true", node->value);
         } else {
             ds_put_format(&errors, "%s: unknown %s argument '%s'\n", name,
                           type, node->key);
@@ -1057,10 +1037,6 @@ get_tunnel_config(const struct netdev *dev, struct smap *args)
         }
     }
 
-    if (tnl_cfg->is_explicit) {
-        smap_add(args, "explicit", "true");
-    }
-
     return 0;
 }
 
@@ -1202,7 +1178,7 @@ netdev_vport_get_ifindex(const struct netdev *netdev_)
     const char *name = netdev_vport_get_dpif_port(netdev_, buf, sizeof(buf));
     const char *dpif_type = netdev_get_dpif_type(netdev_);
 
-    if (dpif_type == NULL || strcmp(dpif_type, "system")) {
+    if (dpif_type && strcmp(dpif_type, "system")) {
         /* Not a system device. */
         return -ENODEV;
     }
@@ -1233,8 +1209,7 @@ netdev_vport_get_ifindex(const struct netdev *netdev_)
     .get_config = get_tunnel_config,                \
     .set_config = set_tunnel_config,                \
     .get_tunnel_config = get_netdev_tunnel_config,  \
-    .get_status = tunnel_get_status,                \
-    .derive_tunnel_config = derive_tunnel_config
+    .get_status = tunnel_get_status
 
 void
 netdev_vport_tunnel_register(void)
@@ -1249,7 +1224,6 @@ netdev_vport_tunnel_register(void)
               .build_header = netdev_geneve_build_header,
               .push_header = netdev_tnl_push_udp_header,
               .pop_header = netdev_geneve_pop_header,
-              .support_explicit_header = netdev_udp_tnl_support_explicit_header,
               .get_ifindex = NETDEV_VPORT_GET_IFINDEX,
           },
           {{NULL, NULL, 0, 0}}
@@ -1261,7 +1235,6 @@ netdev_vport_tunnel_register(void)
               .build_header = netdev_gre_build_header,
               .push_header = netdev_gre_push_header,
               .pop_header = netdev_gre_pop_header,
-              .support_explicit_header = netdev_gre_support_explicit_header,
               .get_ifindex = NETDEV_VPORT_GET_IFINDEX,
           },
           {{NULL, NULL, 0, 0}}
@@ -1273,7 +1246,6 @@ netdev_vport_tunnel_register(void)
               .build_header = netdev_vxlan_build_header,
               .push_header = netdev_tnl_push_udp_header,
               .pop_header = netdev_vxlan_pop_header,
-              .support_explicit_header = netdev_udp_tnl_support_explicit_header,
               .get_ifindex = NETDEV_VPORT_GET_IFINDEX
           },
           {{NULL, NULL, 0, 0}}
