@@ -200,14 +200,6 @@ nl_msg_put(struct ofpbuf *msg, const void *data, size_t size)
     memcpy(nl_msg_put_uninit(msg, size), data, size);
 }
 
-static void
-nl_msg_zero_pad(char *pad_ptr, size_t pad_size)
-{
-    while (pad_size--) {
-        *pad_ptr++ = 0;
-    }
-}
-
 /* Appends 'size' bytes of data, plus Netlink padding if needed, to the tail
  * end of 'msg', reallocating and copying its data if necessary.  Returns a
  * pointer to the first byte of the new data, which is left uninitialized. */
@@ -216,7 +208,9 @@ nl_msg_put_uninit(struct ofpbuf *msg, size_t size)
 {
     size_t pad = PAD_SIZE(size, NLMSG_ALIGNTO);
     char *p = ofpbuf_put_uninit(msg, size + pad);
-    nl_msg_zero_pad(p + size, pad);
+    if (pad) {
+        memset(p + size, 0, pad);
+    }
     return p;
 }
 
@@ -237,7 +231,9 @@ nl_msg_push_uninit(struct ofpbuf *msg, size_t size)
 {
     size_t pad = PAD_SIZE(size, NLMSG_ALIGNTO);
     char *p = ofpbuf_push_uninit(msg, size + pad);
-    nl_msg_zero_pad(p + size, pad);
+    if (pad) {
+        memset(p + size, 0, pad);
+    }
     return p;
 }
 
@@ -523,7 +519,7 @@ size_t
 nl_msg_start_nested(struct ofpbuf *msg, uint16_t type)
 {
     size_t offset = msg->size;
-    nl_msg_put_unspec_uninit(msg, type | NLA_F_NESTED, 0);
+    nl_msg_put_unspec_uninit(msg, type, 0);
     return offset;
 }
 
