@@ -22,6 +22,10 @@
 #include "compiler.h"
 #include "uuid.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 struct json;
 
 /* An atomic type: one that OVSDB regards as a single unit of data. */
@@ -49,6 +53,27 @@ enum ovsdb_ref_type {
     OVSDB_REF_WEAK              /* Delete reference if target disappears. */
 };
 
+struct ovsdb_integer_constraints {
+    int64_t min;        /* minInteger or INT64_MIN. */
+    int64_t max;        /* maxInteger or INT64_MAX. */
+};
+
+struct ovsdb_real_constraints {
+    double min;         /* minReal or -DBL_MAX. */
+    double max;         /* minReal or DBL_MAX. */
+};
+
+struct ovsdb_string_constraints {
+    unsigned int minLen; /* minLength or 0. */
+    unsigned int maxLen; /* maxLength or UINT_MAX. */
+};
+
+struct ovsdb_uuid_constraints {
+    char *refTableName; /* Name of referenced table, or NULL. */
+    struct ovsdb_table *refTable; /* Referenced table, if available. */
+    enum ovsdb_ref_type refType;  /* Reference type. */
+};
+
 struct ovsdb_base_type {
     enum ovsdb_atomic_type type;
 
@@ -57,41 +82,24 @@ struct ovsdb_base_type {
     struct ovsdb_datum *enum_;
 
     union {
-        struct ovsdb_integer_constraints {
-            int64_t min;        /* minInteger or INT64_MIN. */
-            int64_t max;        /* maxInteger or INT64_MAX. */
-        } integer;
-
-        struct ovsdb_real_constraints {
-            double min;         /* minReal or -DBL_MAX. */
-            double max;         /* minReal or DBL_MAX. */
-        } real;
-
+        struct ovsdb_integer_constraints integer;
+        struct ovsdb_real_constraints real;
         /* No constraints for Boolean types. */
-
-        struct ovsdb_string_constraints {
-            unsigned int minLen; /* minLength or 0. */
-            unsigned int maxLen; /* maxLength or UINT_MAX. */
-        } string;
-
-        struct ovsdb_uuid_constraints {
-            char *refTableName; /* Name of referenced table, or NULL. */
-            struct ovsdb_table *refTable; /* Referenced table, if available. */
-            enum ovsdb_ref_type refType;  /* Reference type. */
-        } uuid;
-    } u;
+        struct ovsdb_string_constraints string;
+        struct ovsdb_uuid_constraints uuid;
+    };
 };
 
 #define OVSDB_BASE_VOID_INIT    { .type = OVSDB_TYPE_VOID }
 #define OVSDB_BASE_INTEGER_INIT { .type = OVSDB_TYPE_INTEGER,           \
-                                  .u.integer = { INT64_MIN, INT64_MAX } }
+                                  .integer = { INT64_MIN, INT64_MAX } }
 #define OVSDB_BASE_REAL_INIT    { .type = OVSDB_TYPE_REAL,          \
-                                  .u.real = { -DBL_MAX, DBL_MAX } }
+                                  .real = { -DBL_MAX, DBL_MAX } }
 #define OVSDB_BASE_BOOLEAN_INIT { .type = OVSDB_TYPE_BOOLEAN }
 #define OVSDB_BASE_STRING_INIT  { .type = OVSDB_TYPE_STRING,    \
-                                  .u.string = { 0, UINT_MAX } }
+                                  .string = { 0, UINT_MAX } }
 #define OVSDB_BASE_UUID_INIT    { .type = OVSDB_TYPE_UUID,      \
-                                  .u.uuid = { NULL, NULL, 0 } }
+                                  .uuid = { NULL, NULL, 0 } }
 
 void ovsdb_base_type_init(struct ovsdb_base_type *, enum ovsdb_atomic_type);
 void ovsdb_base_type_clone(struct ovsdb_base_type *,
@@ -176,21 +184,21 @@ ovsdb_atomic_type_is_valid(enum ovsdb_atomic_type atomic_type)
 static inline bool
 ovsdb_base_type_is_ref(const struct ovsdb_base_type *base)
 {
-    return base->type == OVSDB_TYPE_UUID && base->u.uuid.refTableName;
+    return base->type == OVSDB_TYPE_UUID && base->uuid.refTableName;
 }
 
 static inline bool
 ovsdb_base_type_is_strong_ref(const struct ovsdb_base_type *base)
 {
     return (ovsdb_base_type_is_ref(base)
-            && base->u.uuid.refType == OVSDB_REF_STRONG);
+            && base->uuid.refType == OVSDB_REF_STRONG);
 }
 
 static inline bool
 ovsdb_base_type_is_weak_ref(const struct ovsdb_base_type *base)
 {
     return (ovsdb_base_type_is_ref(base)
-            && base->u.uuid.refType == OVSDB_REF_WEAK);
+            && base->uuid.refType == OVSDB_REF_WEAK);
 }
 
 static inline bool ovsdb_type_is_scalar(const struct ovsdb_type *type)
@@ -227,4 +235,20 @@ static inline bool ovsdb_type_is_map(const struct ovsdb_type *type)
     return type->value.type != OVSDB_TYPE_VOID;
 }
 
+static inline bool ovsdb_type_has_strong_refs(const struct ovsdb_type *type)
+{
+    return ovsdb_base_type_is_strong_ref(&type->key)
+           || ovsdb_base_type_is_strong_ref(&type->value);
+}
+
+static inline bool ovsdb_type_has_weak_refs(const struct ovsdb_type *type)
+{
+    return ovsdb_base_type_is_weak_ref(&type->key)
+           || ovsdb_base_type_is_weak_ref(&type->value);
+}
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* ovsdb-types.h */
