@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016 Nicira, Inc.
+ * Copyright (c) 2008-2016, 2019 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -38,7 +38,7 @@
 #include "openvswitch/ofpbuf.h"
 #include "openflow/openflow.h"
 #include "packets.h"
-#include "poll-loop.h"
+#include "openvswitch/poll-loop.h"
 #include "openvswitch/shash.h"
 #include "socket-util.h"
 #include "util.h"
@@ -193,7 +193,9 @@ static void ssl_clear_txbuf(struct ssl_stream *);
 static void interpret_queued_ssl_error(const char *function);
 static int interpret_ssl_error(const char *function, int ret, int error,
                                int *want);
+#if OPENSSL_VERSION_NUMBER < 0x3000000fL
 static DH *tmp_dh_callback(SSL *ssl, int is_export OVS_UNUSED, int keylength);
+#endif
 static void log_ca_cert(const char *file_name, X509 *cert);
 static void stream_ssl_set_ca_cert_file__(const char *file_name,
                                           bool bootstrap, bool force);
@@ -220,9 +222,19 @@ want_to_poll_events(int want)
     }
 }
 
-/* Takes ownership of 'name'. */
+/* Creates a new SSL connection based on socket 'fd', as either a client or a
+ * server according to 'type', initially in 'state'.  On success, returns 0 and
+ * stores the new stream in '*streamp', otherwise returns an errno value and
+ * doesn't bother with '*streamp'.
+ *
+ * Takes ownership of 'name', which should be the name of the connection in the
+ * format that would be used to connect to it, e.g. "ssl:1.2.3.4:5".
+ *
+ * For client connections, 'server_name' should be the host name of the server
+ * being connected to, for use with SSL SNI (server name indication).  Takes
+ * ownership of 'server_name'. */
 static int
-new_ssl_stream(char *name, int fd, enum session_type type,
+new_ssl_stream(char *name, char *server_name, int fd, enum session_type type,
                enum ssl_state state, struct stream **streamp)
 {
     struct ssl_stream *sslv;
@@ -274,6 +286,14 @@ new_ssl_stream(char *name, int fd, enum session_type type,
     if (!verify_peer_cert || (bootstrap_ca_cert && type == CLIENT)) {
         SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);
     }
+#if OPENSSL_SUPPORTS_SNI
+    if (server_name && !SSL_set_tlsext_host_name(ssl, server_name)) {
+        VLOG_ERR("%s: failed to set server name indication (%s)",
+                 server_name, ERR_error_string(ERR_get_error(), NULL));
+        retval = ENOPROTOOPT;
+        goto error;
+    }
+#endif
 
     /* Create and return the ssl_stream. */
     sslv = xmalloc(sizeof *sslv);
@@ -293,6 +313,7 @@ new_ssl_stream(char *name, int fd, enum session_type type,
     }
 
     *streamp = &sslv->stream;
+    free(server_name);
     return 0;
 
 error:
@@ -301,6 +322,7 @@ error:
     }
     closesocket(fd);
     free(name);
+    free(server_name);
     return retval;
 }
 
@@ -311,6 +333,30 @@ ssl_stream_cast(struct stream *stream)
     return CONTAINER_OF(stream, struct ssl_stream, stream);
 }
 
+/* Extracts and returns the server name from 'suffix'.  The caller must
+ * eventually free it.
+ *
+ * Returns NULL if there is no server name, and particularly if it is an IP
+ * address rather than a host name, since RFC 3546 is explicit that IP
+ * addresses are unsuitable as server name indication (SNI). */
+static char *
+get_server_name(const char *suffix_)
+{
+    char *suffix = xstrdup(suffix_);
+
+    char *host, *port;
+    inet_parse_host_port_tokens(suffix, &host, &port);
+
+    ovs_be32 ipv4;
+    struct in6_addr ipv6;
+    char *server_name = (ip_parse(host, &ipv4) || ipv6_parse(host, &ipv6)
+                         ? NULL : xstrdup(host));
+
+    free(suffix);
+
+    return server_name;
+}
+
 static int
 ssl_open(const char *name, char *suffix, struct stream **streamp, uint8_t dscp)
 {
@@ -325,7 +371,8 @@ ssl_open(const char *name, char *suffix, struct stream **streamp, uint8_t dscp)
                              dscp);
     if (fd >= 0) {
         int state = error ? STATE_TCP_CONNECTING : STATE_SSL_CONNECTING;
-        return new_ssl_stream(xstrdup(name), fd, CLIENT, state, streamp);
+        return new_ssl_stream(xstrdup(name), get_server_name(suffix),
+                              fd, CLIENT, state, streamp);
     } else {
         VLOG_ERR("%s: connect: %s", name, ovs_strerror(error));
         return error;
@@ -425,7 +472,12 @@ do_ca_cert_bootstrap(struct stream *stream)
 static char *
 get_peer_common_name(const struct ssl_stream *sslv)
 {
+    char *peer_name = NULL;
+#if OPENSSL_VERSION_NUMBER < 0x3000000fL
     X509 *peer_cert = SSL_get_peer_certificate(sslv->ssl);
+#else
+    X509 *peer_cert = SSL_get1_peer_certificate(sslv->ssl);
+#endif
     if (!peer_cert) {
         return NULL;
     }
@@ -433,28 +485,32 @@ get_peer_common_name(const struct ssl_stream *sslv)
     int cn_index = X509_NAME_get_index_by_NID(X509_get_subject_name(peer_cert),
                                               NID_commonName, -1);
     if (cn_index < 0) {
-        return NULL;
+        goto error;
     }
 
     X509_NAME_ENTRY *cn_entry = X509_NAME_get_entry(
         X509_get_subject_name(peer_cert), cn_index);
     if (!cn_entry) {
-        return NULL;
+        goto error;
     }
 
     ASN1_STRING *cn_data = X509_NAME_ENTRY_get_data(cn_entry);
     if (!cn_data) {
-        return NULL;
+        goto error;
     }
 
     const char *cn;
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined (LIBRESSL_VERSION_NUMBER)
     /* ASN1_STRING_data() is deprecated as of OpenSSL version 1.1 */
     cn = (const char *)ASN1_STRING_data(cn_data);
 #else
     cn = (const char *)ASN1_STRING_get0_data(cn_data);
  #endif
-    return xstrdup(cn);
+    peer_name = xstrdup(cn);
+
+error:
+    X509_free(peer_cert);
+    return peer_name;
 }
 
 static int
@@ -514,6 +570,14 @@ ssl_connect(struct stream *stream)
             VLOG_INFO("rejecting SSL connection during bootstrap race window");
             return EPROTO;
         } else {
+#if OPENSSL_SUPPORTS_SNI
+            const char *servername = SSL_get_servername(
+                sslv->ssl, TLSEXT_NAMETYPE_host_name);
+            if (servername) {
+                VLOG_DBG("connection indicated server name %s", servername);
+            }
+#endif
+
             char *cn = get_peer_common_name(sslv);
 
             if (cn) {
@@ -704,18 +768,22 @@ ssl_send(struct stream *stream, const void *buffer, size_t n)
     if (sslv->txbuf) {
         return -EAGAIN;
     } else {
+        struct ofpbuf buf;
         int error;
 
-        sslv->txbuf = ofpbuf_clone_data(buffer, n);
+        ofpbuf_use_const(&buf, buffer, n);
+        sslv->txbuf = &buf;
         error = ssl_do_tx(stream);
         switch (error) {
         case 0:
-            ssl_clear_txbuf(sslv);
+            sslv->txbuf = NULL;
             return n;
         case EAGAIN:
+            /* Copy remaining data. */
+            sslv->txbuf = ofpbuf_clone_data(buf.data, buf.size);
             return n;
         default:
-            ssl_clear_txbuf(sslv);
+            sslv->txbuf = NULL;
             return -error;
         }
     }
@@ -899,7 +967,7 @@ pssl_accept(struct pstream *pstream, struct stream **new_streamp)
     ds_put_cstr(&name, "ssl:");
     ss_format_address(&ss, &name);
     ds_put_format(&name, ":%"PRIu16, ss_get_port(&ss));
-    return new_ssl_stream(ds_steal_cstr(&name), new_fd, SERVER,
+    return new_ssl_stream(ds_steal_cstr(&name), NULL, new_fd, SERVER,
                           STATE_SSL_CONNECTING, new_streamp);
 }
 
@@ -947,12 +1015,14 @@ do_ssl_init(void)
 {
     SSL_METHOD *method;
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined (LIBRESSL_VERSION_NUMBER)
 #ifdef _WIN32
     /* The following call is needed if we "#include <openssl/applink.c>". */
     CRYPTO_malloc_init();
 #endif
     SSL_library_init();
     SSL_load_error_strings();
+#endif
 
     if (!RAND_status()) {
         /* We occasionally see OpenSSL fail to seed its random number generator
@@ -1005,8 +1075,18 @@ do_ssl_init(void)
         VLOG_ERR("SSL_CTX_new: %s", ERR_error_string(ERR_get_error(), NULL));
         return ENOPROTOOPT;
     }
-    SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
+
+    long options = SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3;
+#ifdef SSL_OP_IGNORE_UNEXPECTED_EOF
+    options |= SSL_OP_IGNORE_UNEXPECTED_EOF;
+#endif
+    SSL_CTX_set_options(ctx, options);
+
+#if OPENSSL_VERSION_NUMBER < 0x3000000fL
     SSL_CTX_set_tmp_dh_callback(ctx, tmp_dh_callback);
+#else
+    SSL_CTX_set_dh_auto(ctx, 1);
+#endif
     SSL_CTX_set_mode(ctx, SSL_MODE_ENABLE_PARTIAL_WRITE);
     SSL_CTX_set_mode(ctx, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
     SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
@@ -1017,6 +1097,7 @@ do_ssl_init(void)
     return 0;
 }
 
+#if OPENSSL_VERSION_NUMBER < 0x3000000fL
 static DH *
 tmp_dh_callback(SSL *ssl OVS_UNUSED, int is_export OVS_UNUSED, int keylength)
 {
@@ -1027,7 +1108,6 @@ tmp_dh_callback(SSL *ssl OVS_UNUSED, int is_export OVS_UNUSED, int keylength)
     };
 
     static struct dh dh_table[] = {
-        {1024, NULL, get_dh1024},
         {2048, NULL, get_dh2048},
         {4096, NULL, get_dh4096},
     };
@@ -1035,7 +1115,7 @@ tmp_dh_callback(SSL *ssl OVS_UNUSED, int is_export OVS_UNUSED, int keylength)
     struct dh *dh;
 
     for (dh = dh_table; dh < &dh_table[ARRAY_SIZE(dh_table)]; dh++) {
-        if (dh->keylength == keylength) {
+        if (dh->keylength >= keylength) {
             if (!dh->dh) {
                 dh->dh = dh->constructor();
                 if (!dh->dh) {
@@ -1049,6 +1129,7 @@ tmp_dh_callback(SSL *ssl OVS_UNUSED, int is_export OVS_UNUSED, int keylength)
                 keylength);
     return NULL;
 }
+#endif
 
 /* Returns true if SSL is at least partially configured. */
 bool
@@ -1155,7 +1236,7 @@ stream_ssl_set_key_and_cert(const char *private_key_file,
                             const char *certificate_file)
 {
     if (update_ssl_config(&private_key, private_key_file)
-        || update_ssl_config(&certificate, certificate_file)) {
+        && update_ssl_config(&certificate, certificate_file)) {
         stream_ssl_set_certificate_file__(certificate_file);
         stream_ssl_set_private_key_file__(private_key_file);
     }
@@ -1186,8 +1267,13 @@ stream_ssl_set_protocols(const char *arg)
     }
 
     /* Start with all the flags off and turn them on as requested. */
-    long protocol_flags = SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3 | SSL_OP_NO_TLSv1;
-    protocol_flags |= SSL_OP_NO_TLSv1_1 | SSL_OP_NO_TLSv1_2;
+#ifndef SSL_OP_NO_SSL_MASK
+    /* For old OpenSSL without this macro, this is the correct value.  */
+#define SSL_OP_NO_SSL_MASK (SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3 | \
+                            SSL_OP_NO_TLSv1 | SSL_OP_NO_TLSv1_1 | \
+                            SSL_OP_NO_TLSv1_2)
+#endif
+    long protocol_flags = SSL_OP_NO_SSL_MASK;
 
     char *s = xstrdup(arg);
     char *save_ptr = NULL;
@@ -1247,12 +1333,12 @@ read_cert_file(const char *file_name, X509 ***certs, size_t *n_certs)
     }
 
     for (;;) {
-        X509 *certificate;
+        X509 *cert;
         int c;
 
         /* Read certificate from file. */
-        certificate = PEM_read_X509(file, NULL, NULL, NULL);
-        if (!certificate) {
+        cert = PEM_read_X509(file, NULL, NULL, NULL);
+        if (!cert) {
             size_t i;
 
             VLOG_ERR("PEM_read_X509 failed reading %s: %s",
@@ -1271,7 +1357,7 @@ read_cert_file(const char *file_name, X509 ***certs, size_t *n_certs)
         if (*n_certs >= allocated_certs) {
             *certs = x2nrealloc(*certs, &allocated_certs, sizeof **certs);
         }
-        (*certs)[(*n_certs)++] = certificate;
+        (*certs)[(*n_certs)++] = cert;
 
         /* Are there additional certificates in the file? */
         do {
