@@ -212,9 +212,9 @@ struct connmgr {
      * traversals from other threads can be made safe by holding the
      * ofproto_mutex.*/
     struct ovs_list conns;       /* All ofconns. */
-    uint64_t master_election_id; /* monotonically increasing sequence number
-                                  * for master election */
-    bool master_election_id_defined;
+    uint64_t primary_election_id; /* monotonically increasing sequence number
+                                   * for primary election */
+    bool primary_election_id_defined;
 
     /* OpenFlow connection establishment. */
     struct hmap services;       /* Contains "struct ofservice"s. */
@@ -253,8 +253,8 @@ connmgr_create(struct ofproto *ofproto,
     mgr->local_port_name = xstrdup(local_port_name);
 
     ovs_list_init(&mgr->conns);
-    mgr->master_election_id = 0;
-    mgr->master_election_id_defined = false;
+    mgr->primary_election_id = 0;
+    mgr->primary_election_id_defined = false;
 
     hmap_init(&mgr->services);
     mgr->snoops = NULL;
@@ -310,8 +310,8 @@ connmgr_destroy(struct connmgr *mgr)
         return;
     }
 
-    struct ofservice *ofservice, *next_ofservice;
-    HMAP_FOR_EACH_SAFE (ofservice, next_ofservice, hmap_node, &mgr->services) {
+    struct ofservice *ofservice;
+    HMAP_FOR_EACH_SAFE (ofservice, hmap_node, &mgr->services) {
         ofservice_destroy(ofservice);
     }
     hmap_destroy(&mgr->services);
@@ -351,8 +351,8 @@ connmgr_run(struct connmgr *mgr,
         }
     }
 
-    struct ofconn *ofconn, *next_ofconn;
-    LIST_FOR_EACH_SAFE (ofconn, next_ofconn, connmgr_node, &mgr->conns) {
+    struct ofconn *ofconn;
+    LIST_FOR_EACH_SAFE (ofconn, connmgr_node, &mgr->conns) {
         ofconn_run(ofconn, handle_openflow);
     }
     ofmonitor_run(mgr);
@@ -592,8 +592,8 @@ connmgr_set_controllers(struct connmgr *mgr, struct shash *controllers)
 
     /* Delete services that are no longer configured.
      * Update configuration of all now-existing services. */
-    struct ofservice *ofservice, *next_ofservice;
-    HMAP_FOR_EACH_SAFE (ofservice, next_ofservice, hmap_node, &mgr->services) {
+    struct ofservice *ofservice;
+    HMAP_FOR_EACH_SAFE (ofservice, hmap_node, &mgr->services) {
         const char *target = ofservice->target;
         struct ofproto_controller *c = shash_find_data(controllers, target);
         if (!c) {
@@ -773,11 +773,11 @@ snoop_preference(const struct ofservice *ofservice)
     }
 
     switch (ofconn->role) {
-    case OFPCR12_ROLE_MASTER:
+    case OFPCR12_ROLE_PRIMARY:
         return 3;
     case OFPCR12_ROLE_EQUAL:
         return 2;
-    case OFPCR12_ROLE_SLAVE:
+    case OFPCR12_ROLE_SECONDARY:
         return 1;
     case OFPCR12_ROLE_NOCHANGE:
     default:
@@ -818,33 +818,33 @@ ofconn_get_type(const struct ofconn *ofconn)
     return ofconn->type;
 }
 
-/* If a master election id is defined, stores it into '*idp' and returns
+/* If a primary election id is defined, stores it into '*idp' and returns
  * true.  Otherwise, stores UINT64_MAX into '*idp' and returns false. */
 bool
-ofconn_get_master_election_id(const struct ofconn *ofconn, uint64_t *idp)
+ofconn_get_primary_election_id(const struct ofconn *ofconn, uint64_t *idp)
 {
-    *idp = (ofconn->connmgr->master_election_id_defined
-            ? ofconn->connmgr->master_election_id
+    *idp = (ofconn->connmgr->primary_election_id_defined
+            ? ofconn->connmgr->primary_election_id
             : UINT64_MAX);
-    return ofconn->connmgr->master_election_id_defined;
+    return ofconn->connmgr->primary_election_id_defined;
 }
 
-/* Sets the master election id.
+/* Sets the primary election id.
  *
  * Returns true if successful, false if the id is stale
  */
 bool
-ofconn_set_master_election_id(struct ofconn *ofconn, uint64_t id)
+ofconn_set_primary_election_id(struct ofconn *ofconn, uint64_t id)
 {
-    if (ofconn->connmgr->master_election_id_defined
+    if (ofconn->connmgr->primary_election_id_defined
         &&
         /* Unsigned difference interpreted as a two's complement signed
          * value */
-        (int64_t)(id - ofconn->connmgr->master_election_id) < 0) {
+        (int64_t)(id - ofconn->connmgr->primary_election_id) < 0) {
         return false;
     }
-    ofconn->connmgr->master_election_id = id;
-    ofconn->connmgr->master_election_id_defined = true;
+    ofconn->connmgr->primary_election_id = id;
+    ofconn->connmgr->primary_election_id_defined = true;
 
     return true;
 }
@@ -864,7 +864,7 @@ ofconn_send_role_status(struct ofconn *ofconn, uint32_t role, uint8_t reason)
     struct ofputil_role_status status;
     status.reason = reason;
     status.role = role;
-    ofconn_get_master_election_id(ofconn, &status.generation_id);
+    ofconn_get_primary_election_id(ofconn, &status.generation_id);
 
     struct ofpbuf *buf
         = ofputil_encode_role_status(&status, ofconn_get_protocol(ofconn));
@@ -873,19 +873,19 @@ ofconn_send_role_status(struct ofconn *ofconn, uint32_t role, uint8_t reason)
     }
 }
 
-/* Changes 'ofconn''s role to 'role'.  If 'role' is OFPCR12_ROLE_MASTER then
- * any existing master is demoted to a slave. */
+/* Changes 'ofconn''s role to 'role'.  If 'role' is OFPCR12_ROLE_PRIMARY then
+ * any existing primary is demoted to a secondary. */
 void
 ofconn_set_role(struct ofconn *ofconn, enum ofp12_controller_role role)
 {
-    if (role != ofconn->role && role == OFPCR12_ROLE_MASTER) {
+    if (role != ofconn->role && role == OFPCR12_ROLE_PRIMARY) {
         struct ofconn *other;
 
         LIST_FOR_EACH (other, connmgr_node, &ofconn->connmgr->conns) {
-            if (other->role == OFPCR12_ROLE_MASTER) {
-                other->role = OFPCR12_ROLE_SLAVE;
-                ofconn_send_role_status(other, OFPCR12_ROLE_SLAVE,
-                                        OFPCRR_MASTER_REQUEST);
+            if (other->role == OFPCR12_ROLE_PRIMARY) {
+                other->role = OFPCR12_ROLE_SECONDARY;
+                ofconn_send_role_status(other, OFPCR12_ROLE_SECONDARY,
+                                        OFPCRR_PRIMARY_REQUEST);
             }
         }
     }
@@ -898,9 +898,9 @@ ofconn_set_invalid_ttl_to_controller(struct ofconn *ofconn, bool enable)
     struct ofputil_async_cfg ac = ofconn_get_async_config(ofconn);
     uint32_t bit = 1u << OFPR_INVALID_TTL;
     if (enable) {
-        ac.master[OAM_PACKET_IN] |= bit;
+        ac.primary[OAM_PACKET_IN] |= bit;
     } else {
-        ac.master[OAM_PACKET_IN] &= ~bit;
+        ac.primary[OAM_PACKET_IN] &= ~bit;
     }
     ofconn_set_async_config(ofconn, &ac);
 }
@@ -910,7 +910,7 @@ ofconn_get_invalid_ttl_to_controller(struct ofconn *ofconn)
 {
     struct ofputil_async_cfg ac = ofconn_get_async_config(ofconn);
     uint32_t bit = 1u << OFPR_INVALID_TTL;
-    return (ac.master[OAM_PACKET_IN] & bit) != 0;
+    return (ac.primary[OAM_PACKET_IN] & bit) != 0;
 }
 
 /* Returns the currently configured protocol for 'ofconn', one of OFPUTIL_P_*.
@@ -1002,11 +1002,11 @@ ofconn_set_async_config(struct ofconn *ofconn,
 
     if (ofputil_protocol_to_ofp_version(ofconn_get_protocol(ofconn))
         < OFP14_VERSION) {
-        if (ofconn->async_cfg->master[OAM_PACKET_IN] & (1u << OFPR_ACTION)) {
-            ofconn->async_cfg->master[OAM_PACKET_IN] |= OFPR14_ACTION_BITS;
+        if (ofconn->async_cfg->primary[OAM_PACKET_IN] & (1u << OFPR_ACTION)) {
+            ofconn->async_cfg->primary[OAM_PACKET_IN] |= OFPR14_ACTION_BITS;
         }
-        if (ofconn->async_cfg->slave[OAM_PACKET_IN] & (1u << OFPR_ACTION)) {
-            ofconn->async_cfg->slave[OAM_PACKET_IN] |= OFPR14_ACTION_BITS;
+        if (ofconn->async_cfg->secondary[OAM_PACKET_IN] & (1u << OFPR_ACTION)) {
+            ofconn->async_cfg->secondary[OAM_PACKET_IN] |= OFPR14_ACTION_BITS;
         }
     }
 }
@@ -1137,9 +1137,9 @@ ofconn_remove_bundle(struct ofconn *ofconn, struct ofp_bundle *bundle)
 static void
 bundle_remove_all(struct ofconn *ofconn)
 {
-    struct ofp_bundle *b, *next;
+    struct ofp_bundle *b;
 
-    HMAP_FOR_EACH_SAFE (b, next, node, &ofconn->bundles) {
+    HMAP_FOR_EACH_SAFE (b, node, &ofconn->bundles) {
         ofp_bundle_remove__(ofconn, b);
     }
 }
@@ -1149,8 +1149,8 @@ bundle_remove_expired(struct ofconn *ofconn, long long int now)
 {
     long long int limit = now - bundle_idle_timeout;
 
-    struct ofp_bundle *b, *next;
-    HMAP_FOR_EACH_SAFE (b, next, node, &ofconn->bundles) {
+    struct ofp_bundle *b;
+    HMAP_FOR_EACH_SAFE (b, node, &ofconn->bundles) {
         if (b->used <= limit) {
             ofconn_send_error(ofconn, b->msg, OFPERR_OFPBFC_TIMEOUT);
             ofp_bundle_remove__(ofconn, b);
@@ -1209,7 +1209,7 @@ ofconn_create(struct ofservice *ofservice, struct rconn *rconn,
     hmap_init(&ofconn->bundles);
     ofconn->next_bundle_expiry_check = time_msec() + BUNDLE_EXPIRY_INTERVAL;
 
-    ofconn_set_rate_limit(ofconn, settings->rate_limit, settings->burst_limit);
+    ofservice_reconfigure(ofservice, settings);
 
     ovs_mutex_unlock(&ofproto_mutex);
 }
@@ -1247,8 +1247,8 @@ ofconn_destroy(struct ofconn *ofconn)
 
     free(ofconn->async_cfg);
 
-    struct ofmonitor *monitor, *next_monitor;
-    HMAP_FOR_EACH_SAFE (monitor, next_monitor, ofconn_node,
+    struct ofmonitor *monitor;
+    HMAP_FOR_EACH_SAFE (monitor, ofconn_node,
                         &ofconn->monitors) {
         ofmonitor_destroy(monitor);
     }
@@ -1441,9 +1441,9 @@ ofconn_receives_async_msg(const struct ofconn *ofconn,
     }
 
     struct ofputil_async_cfg ac = ofconn_get_async_config(ofconn);
-    uint32_t *masks = (ofconn->role == OFPCR12_ROLE_SLAVE
-                       ? ac.slave
-                       : ac.master);
+    uint32_t *masks = (ofconn->role == OFPCR12_ROLE_SECONDARY
+                       ? ac.secondary
+                       : ac.primary);
     return (masks[type] & (1u << reason)) != 0;
 }
 
@@ -1649,6 +1649,8 @@ connmgr_send_table_status(struct connmgr *mgr,
     }
 }
 
+COVERAGE_DEFINE(connmgr_async_unsent);
+
 /* Given 'pin', sends an OFPT_PACKET_IN message to each OpenFlow controller as
  * necessary according to their individual configurations. */
 void
@@ -1656,6 +1658,7 @@ connmgr_send_async_msg(struct connmgr *mgr,
                        const struct ofproto_async_msg *am)
 {
     struct ofconn *ofconn;
+    bool sent = false;
 
     LIST_FOR_EACH (ofconn, connmgr_node, &mgr->conns) {
         enum ofputil_protocol protocol = ofconn_get_protocol(ofconn);
@@ -1677,6 +1680,11 @@ connmgr_send_async_msg(struct connmgr *mgr,
                       am->pin.up.base.flow_metadata.flow.in_port.ofp_port,
                       msg, &txq);
         do_send_packet_ins(ofconn, &txq);
+        sent = true;
+    }
+
+    if (!sent) {
+        COVERAGE_INC(connmgr_async_unsent);
     }
 }
 
@@ -1907,10 +1915,7 @@ connmgr_count_hidden_rules(const struct connmgr *mgr)
 }
 
 /* Creates a new ofservice for 'target' in 'mgr'.  Returns 0 if successful,
- * otherwise a positive errno value.
- *
- * ofservice_reconfigure() must be called to fully configure the new
- * ofservice. */
+ * otherwise a positive errno value. */
 static void
 ofservice_create(struct connmgr *mgr, const char *target,
                  const struct ofproto_controller *c)
@@ -1920,7 +1925,8 @@ ofservice_create(struct connmgr *mgr, const char *target,
     struct rconn *rconn = NULL;
     if (!vconn_verify_name(target)) {
         char *name = ofconn_make_name(mgr, target);
-        rconn = rconn_create(5, 8, c->dscp, c->allowed_versions);
+        rconn = rconn_create(c->probe_interval, c->max_backoff,
+                             c->dscp, c->allowed_versions);
         rconn_connect(rconn, target, name);
         free(name);
     } else if (!pvconn_verify_name(target)) {
@@ -1943,7 +1949,6 @@ ofservice_create(struct connmgr *mgr, const char *target,
     ofservice->rconn = rconn;
     ofservice->pvconn = pvconn;
     ofservice->s = *c;
-    ofservice_reconfigure(ofservice, c);
 
     VLOG_INFO("%s: added %s controller \"%s\"",
               mgr->name, ofconn_type_to_string(ofservice->type), target);
@@ -1953,8 +1958,8 @@ static void
 ofservice_close_all(struct ofservice *ofservice)
     OVS_REQUIRES(ofproto_mutex)
 {
-    struct ofconn *ofconn, *next;
-    LIST_FOR_EACH_SAFE (ofconn, next, ofservice_node, &ofservice->conns) {
+    struct ofconn *ofconn;
+    LIST_FOR_EACH_SAFE (ofconn, ofservice_node, &ofservice->conns) {
         ofconn_destroy(ofconn);
     }
 }
@@ -2140,7 +2145,7 @@ ofmonitor_report(struct connmgr *mgr, struct rule *rule,
                  const struct rule_actions *old_actions)
     OVS_REQUIRES(ofproto_mutex)
 {
-    if (rule_is_hidden(rule)) {
+    if (!mgr || rule_is_hidden(rule)) {
         return;
     }
 
@@ -2244,6 +2249,10 @@ ofmonitor_flush(struct connmgr *mgr)
 {
     struct ofconn *ofconn;
 
+    if (!mgr) {
+        return;
+    }
+
     LIST_FOR_EACH (ofconn, connmgr_node, &mgr->conns) {
         struct rconn_packet_counter *counter = ofconn->monitor_counter;
 
