@@ -1,4 +1,4 @@
-/* Copyright (c) 2009, 2010, 2011, 2012, 2013, 2017 Nicira, Inc.
+/* Copyright (c) 2009, 2010, 2011, 2012, 2013, 2017, 2019 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -99,7 +99,8 @@ lookup_executor(const char *name, bool *read_only)
 }
 
 /* On success, returns a transaction and stores the results to return to the
- * client in '*resultsp'.
+ * client in '*resultsp'.  If 'forwarding_needed' is nonnull and transaction
+ * needs to be forwarded (in relay mode), sets '*forwarding_needed' to true.
  *
  * On failure, returns NULL.  If '*resultsp' is nonnull, then it is the results
  * to return to the client.  If '*resultsp' is null, then the execution failed
@@ -111,7 +112,8 @@ ovsdb_execute_compose(struct ovsdb *db, const struct ovsdb_session *session,
                       const struct json *params, bool read_only,
                       const char *role, const char *id,
                       long long int elapsed_msec, long long int *timeout_msec,
-                      bool *durable, struct json **resultsp)
+                      bool *durable, bool *forwarding_needed,
+                      struct json **resultsp)
 {
     struct ovsdb_execution x;
     struct ovsdb_error *error;
@@ -120,10 +122,13 @@ ovsdb_execute_compose(struct ovsdb *db, const struct ovsdb_session *session,
     size_t i;
 
     *durable = false;
+    if (forwarding_needed) {
+        *forwarding_needed = false;
+    }
     if (params->type != JSON_ARRAY
-        || !params->u.array.n
-        || params->u.array.elems[0]->type != JSON_STRING
-        || strcmp(params->u.array.elems[0]->u.string, db->schema->name)) {
+        || !params->array.n
+        || params->array.elems[0]->type != JSON_STRING
+        || strcmp(params->array.elems[0]->string, db->schema->name)) {
         if (params->type != JSON_ARRAY) {
             error = ovsdb_syntax_error(params, NULL, "array expected");
         } else {
@@ -147,10 +152,10 @@ ovsdb_execute_compose(struct ovsdb *db, const struct ovsdb_session *session,
     results = NULL;
 
     results = json_array_create_empty();
-    n_operations = params->u.array.n - 1;
+    n_operations = params->array.n - 1;
     error = NULL;
     for (i = 1; i <= n_operations; i++) {
-        struct json *operation = params->u.array.elems[i];
+        struct json *operation = params->array.elems[i];
         struct ovsdb_error *parse_error;
         struct ovsdb_parser parser;
         struct json *result;
@@ -196,6 +201,8 @@ ovsdb_execute_compose(struct ovsdb *db, const struct ovsdb_session *session,
                                     "%s operation not allowed on "
                                     "table in reserved database %s",
                                     op_name, db->schema->name);
+            } else if (db->is_relay && forwarding_needed) {
+                *forwarding_needed = true;
             }
         }
         if (error) {
@@ -240,7 +247,7 @@ ovsdb_execute(struct ovsdb *db, const struct ovsdb_session *session,
     struct json *results;
     struct ovsdb_txn *txn = ovsdb_execute_compose(
         db, session, params, read_only, role, id, elapsed_msec, timeout_msec,
-        &durable, &results);
+        &durable, NULL, &results);
     if (!txn) {
         return results;
     }
@@ -313,7 +320,7 @@ parse_row(const struct json *json, const struct ovsdb_table *table,
     }
 
     row = ovsdb_row_create(table);
-    error = ovsdb_row_from_json(row, json, symtab, columns);
+    error = ovsdb_row_from_json(row, json, symtab, columns, false);
     if (error) {
         ovsdb_row_destroy(row);
         return error;
@@ -329,11 +336,12 @@ ovsdb_execute_insert(struct ovsdb_execution *x, struct ovsdb_parser *parser,
 {
     struct ovsdb_table *table;
     struct ovsdb_row *row = NULL;
-    const struct json *uuid_name, *row_json;
+    const struct json *uuid_json, *uuid_name, *row_json;
     struct ovsdb_error *error;
     struct uuid row_uuid;
 
     table = parse_table(x, parser, "table");
+    uuid_json = ovsdb_parser_member(parser, "uuid", OP_STRING | OP_OPTIONAL);
     uuid_name = ovsdb_parser_member(parser, "uuid-name", OP_ID | OP_OPTIONAL);
     row_json = ovsdb_parser_member(parser, "row", OP_OBJECT);
     error = ovsdb_parser_get_error(parser);
@@ -341,6 +349,19 @@ ovsdb_execute_insert(struct ovsdb_execution *x, struct ovsdb_parser *parser,
         return error;
     }
 
+    if (uuid_json) {
+        if (!uuid_from_string(&row_uuid, json_string(uuid_json))) {
+            return ovsdb_syntax_error(uuid_json, NULL, "bad uuid");
+        }
+
+        if (!ovsdb_txn_may_create_row(table, &row_uuid)) {
+            return ovsdb_syntax_error(uuid_json, "duplicate uuid",
+                                      "This UUID would duplicate a UUID "
+                                      "already present within the table or "
+                                      "deleted within the same transaction.");
+        }
+    }
+
     if (uuid_name) {
         struct ovsdb_symbol *symbol;
 
@@ -350,9 +371,13 @@ ovsdb_execute_insert(struct ovsdb_execution *x, struct ovsdb_parser *parser,
                                       "This \"uuid-name\" appeared on an "
                                       "earlier \"insert\" operation.");
         }
-        row_uuid = symbol->uuid;
+        if (uuid_json) {
+            symbol->uuid = row_uuid;
+        } else {
+            row_uuid = symbol->uuid;
+        }
         symbol->created = true;
-    } else {
+    } else if (!uuid_json) {
         uuid_generate(&row_uuid);
     }
 
@@ -465,8 +490,9 @@ update_row_cb(const struct ovsdb_row *row, void *ur_)
 
     ur->n_matches++;
     if (!ovsdb_row_equal_columns(row, ur->row, ur->columns)) {
-        ovsdb_row_update_columns(ovsdb_txn_row_modify(ur->txn, row),
-                                 ur->row, ur->columns);
+        ovsdb_error_assert(ovsdb_row_update_columns(
+                               ovsdb_txn_row_modify(ur->txn, row),
+                               ur->row, ur->columns, false));
     }
 
     return true;
@@ -562,7 +588,6 @@ ovsdb_execute_mutate(struct ovsdb_execution *x, struct ovsdb_parser *parser,
     const struct json *mutations_json;
     struct ovsdb_condition condition = OVSDB_CONDITION_INITIALIZER(&condition);
     struct ovsdb_mutation_set mutations = OVSDB_MUTATION_SET_INITIALIZER;
-    struct ovsdb_row *row = NULL;
     struct mutate_row_cbdata mr;
     struct ovsdb_error *error;
 
@@ -595,7 +620,6 @@ ovsdb_execute_mutate(struct ovsdb_execution *x, struct ovsdb_parser *parser,
         json_object_put(result, "count", json_integer_create(mr.n_matches));
     }
 
-    ovsdb_row_destroy(row);
     ovsdb_mutation_set_destroy(&mutations);
     ovsdb_condition_destroy(&condition);
 
@@ -696,7 +720,7 @@ ovsdb_execute_wait(struct ovsdb_execution *x, struct ovsdb_parser *parser,
     long long int timeout_msec = 0;
     size_t i;
 
-    timeout = ovsdb_parser_member(parser, "timeout", OP_NUMBER | OP_OPTIONAL);
+    timeout = ovsdb_parser_member(parser, "timeout", OP_INTEGER | OP_OPTIONAL);
     where = ovsdb_parser_member(parser, "where", OP_ARRAY);
     columns_json = ovsdb_parser_member(parser, "columns",
                                        OP_ARRAY | OP_OPTIONAL);
@@ -714,7 +738,7 @@ ovsdb_execute_wait(struct ovsdb_execution *x, struct ovsdb_parser *parser,
     }
     if (!error) {
         if (timeout) {
-            timeout_msec = MIN(LLONG_MAX, json_real(timeout));
+            timeout_msec = json_integer(timeout);
             if (timeout_msec < 0) {
                 error = ovsdb_syntax_error(timeout, NULL,
                                            "timeout must be nonnegative");
@@ -735,12 +759,12 @@ ovsdb_execute_wait(struct ovsdb_execution *x, struct ovsdb_parser *parser,
     if (!error) {
         /* Parse "rows" into 'expected'. */
         ovsdb_row_hash_init(&expected, &columns);
-        for (i = 0; i < rows->u.array.n; i++) {
+        for (i = 0; i < rows->array.n; i++) {
             struct ovsdb_row *row;
 
             row = ovsdb_row_create(table);
-            error = ovsdb_row_from_json(row, rows->u.array.elems[i], x->symtab,
-                                        NULL);
+            error = ovsdb_row_from_json(row, rows->array.elems[i], x->symtab,
+                                        NULL, false);
             if (error) {
                 ovsdb_row_destroy(row);
                 break;
