@@ -58,6 +58,9 @@ static const char *rbac_role;
 /* --cid: Cluster ID for "join-cluster" command. */
 static struct uuid cid;
 
+/* --election-timer: Election timer for "create-cluster" command. */
+static uint64_t election_timer;
+
 static const struct ovs_cmdl_command *get_all_commands(void);
 
 OVS_NO_RETURN static void usage(void);
@@ -85,12 +88,14 @@ parse_options(int argc, char *argv[])
 {
     enum {
         OPT_RBAC_ROLE = UCHAR_MAX + 1,
-        OPT_CID
+        OPT_CID,
+        OPT_ELECTION_TIMER,
     };
     static const struct option long_options[] = {
         {"more", no_argument, NULL, 'm'},
         {"rbac-role", required_argument, NULL, OPT_RBAC_ROLE},
         {"cid", required_argument, NULL, OPT_CID},
+        {"election-timer", required_argument, NULL, OPT_ELECTION_TIMER},
         {"verbose", optional_argument, NULL, 'v'},
         {"help", no_argument, NULL, 'h'},
         {"option", no_argument, NULL, 'o'},
@@ -100,6 +105,7 @@ parse_options(int argc, char *argv[])
     char *short_options = ovs_cmdl_long_options_to_short_options(long_options);
 
     for (;;) {
+        struct ovsdb_error *error;
         int c;
 
         c = getopt_long(argc, argv, short_options, long_options, NULL);
@@ -122,6 +128,14 @@ parse_options(int argc, char *argv[])
             }
             break;
 
+        case OPT_ELECTION_TIMER:
+            election_timer = atoll(optarg);
+            error = raft_validate_election_timer(election_timer);
+            if (error) {
+                ovs_fatal(0, "%s", ovsdb_error_to_string_free(error));
+            }
+            break;
+
         case 'h':
             usage();
 
@@ -153,7 +167,7 @@ usage(void)
     printf("%s: Open vSwitch database management utility\n"
            "usage: %s [OPTIONS] COMMAND [ARG...]\n"
            "  create [DB [SCHEMA]]    create DB with the given SCHEMA\n"
-           "  create-cluster DB CONTENTS LOCAL\n"
+           "  [--election-timer=ms] create-cluster DB CONTENTS LOCAL\n"
            "    create clustered DB with given CONTENTS and LOCAL address\n"
            "  [--cid=UUID] join-cluster DB NAME LOCAL REMOTE...\n"
            "    join clustered DB with given NAME and LOCAL and REMOTE addrs\n"
@@ -173,6 +187,9 @@ usage(void)
            "  compare-versions A OP B  compare OVSDB schema version numbers\n"
            "  query [DB] TRNS         execute read-only transaction on DB\n"
            "  transact [DB] TRNS      execute read/write transaction on DB\n"
+           "  cluster-to-standalone DB DB    Convert clustered DB to\n"
+           "      standalone DB when cluster is down and cannot be\n"
+           "        revived\n"
            "  [-m]... show-log [DB]   print DB's log entries\n"
            "The default DB is %s.\n"
            "The default SCHEMA is %s.\n",
@@ -212,7 +229,7 @@ parse_json(const char *s)
 {
     struct json *json = json_from_string(s);
     if (json->type == JSON_STRING) {
-        ovs_fatal(0, "\"%s\": %s", s, json->u.string);
+        ovs_fatal(0, "\"%s\": %s", s, json->string);
     }
     return json;
 }
@@ -226,14 +243,6 @@ print_and_free_json(struct json *json)
     free(string);
 }
 
-static struct ovsdb_error *
-write_and_free_json(struct ovsdb_log *log, struct json *json)
-{
-    struct ovsdb_error *error = ovsdb_log_write(log, json);
-    json_destroy(json);
-    return error;
-}
-
 static void
 check_ovsdb_error(struct ovsdb_error *error)
 {
@@ -270,7 +279,7 @@ do_create(struct ovs_cmdl_context *ctx)
     /* Create database file. */
     check_ovsdb_error(ovsdb_log_open(db_file_name, OVSDB_MAGIC,
                                      OVSDB_LOG_CREATE_EXCL, -1, &log));
-    check_ovsdb_error(write_and_free_json(log, json));
+    check_ovsdb_error(ovsdb_log_write_and_free(log, json));
     check_ovsdb_error(ovsdb_log_commit_block(log));
     ovsdb_log_close(log);
 }
@@ -308,7 +317,7 @@ do_create_cluster(struct ovs_cmdl_context *ctx)
     /* Create database file. */
     struct json *snapshot = json_array_create_2(schema_json, data);
     check_ovsdb_error(raft_create_cluster(db_file_name, schema->name,
-                                          local, snapshot));
+                                          local, snapshot, election_timer));
     ovsdb_schema_destroy(schema);
     json_destroy(snapshot);
 }
@@ -348,9 +357,9 @@ write_standalone_db(const char *file_name, const char *comment,
         return error;
     }
 
-    error = write_and_free_json(log, ovsdb_schema_to_json(db->schema));
+    error = ovsdb_log_write_and_free(log, ovsdb_schema_to_json(db->schema));
     if (!error) {
-        error = write_and_free_json(log, ovsdb_to_txn_json(db, comment));
+        error = ovsdb_log_write_and_free(log, ovsdb_to_txn_json(db, comment));
     }
     ovsdb_log_close(log);
 
@@ -396,6 +405,9 @@ compact_or_convert(const char *src_name_, const char *dst_name_,
         ovs_fatal(retval, "%s: failed to lock lockfile", dst_name);
     }
 
+    /* Resulted DB will contain a single transaction without diff anyway. */
+    ovsdb_file_column_diff_disable();
+
     /* Save a copy. */
     struct ovsdb *ovsdb = (new_schema
                            ? ovsdb_file_read_as_schema(src_name, new_schema)
@@ -567,7 +579,9 @@ do_db_has_magic(struct ovs_cmdl_context *ctx, const char *magic)
 
     check_ovsdb_error(ovsdb_log_open(filename, OVSDB_MAGIC"|"RAFT_MAGIC,
                                      OVSDB_LOG_READ_ONLY, -1, &log));
-    if (strcmp(ovsdb_log_get_magic(log), magic)) {
+    int cmp = strcmp(ovsdb_log_get_magic(log), magic);
+    ovsdb_log_close(log);
+    if (cmp) {
         exit(2);
     }
 }
@@ -651,6 +665,8 @@ static void
 print_db_changes(struct shash *tables, struct smap *names,
                  const struct ovsdb_schema *schema)
 {
+    struct json *is_diff = shash_find_data(tables, "_is_diff");
+    bool diff = (is_diff && is_diff->type == JSON_TRUE);
     struct shash_node *n1;
 
     int i = 0;
@@ -685,7 +701,7 @@ print_db_changes(struct shash *tables, struct smap *names,
                 }
             }
 
-            printf("\ttable %s", table);
+            printf("  table %s", table);
 
             if (!old_name) {
                 if (new_name) {
@@ -694,7 +710,8 @@ print_db_changes(struct shash *tables, struct smap *names,
                     printf(" insert row %.8s:\n", row_uuid);
                 }
             } else {
-                printf(" row %s (%.8s):\n", old_name, row_uuid);
+                printf(" row %s (%.8s)%s:\n", old_name, row_uuid,
+                                              diff ? " diff" : "");
             }
 
             if (columns->type == JSON_OBJECT) {
@@ -723,6 +740,7 @@ print_db_changes(struct shash *tables, struct smap *names,
                                 ds_init(&s);
                                 ovsdb_datum_to_string(&datum, type, &s);
                                 value_string = ds_steal_cstr(&s);
+                                ovsdb_datum_destroy(&datum, type);
                             } else {
                                 ovsdb_error_destroy(error);
                             }
@@ -730,7 +748,7 @@ print_db_changes(struct shash *tables, struct smap *names,
                         if (!value_string) {
                             value_string = json_to_string(value, JSSF_SORT);
                         }
-                        printf("\t\t%s=%s\n", column, value_string);
+                        printf("    %s=%s\n", column, value_string);
                         free(value_string);
                     }
                 }
@@ -743,7 +761,7 @@ print_db_changes(struct shash *tables, struct smap *names,
                                     xmemdup0(row_uuid, 8));
                 }
             } else if (columns->type == JSON_NULL) {
-                printf("\t\tdelete row\n");
+                printf("    delete row\n");
                 smap_remove(names, row_uuid);
             }
 
@@ -901,7 +919,8 @@ print_raft_header(const struct raft_header *h,
         if (!uuid_is_zero(&h->snap.eid)) {
             printf(" prev_eid: %04x\n", uuid_prefix(&h->snap.eid, 4));
         }
-        print_data("prev_", h->snap.data, schemap, names);
+        print_data("prev_", raft_entry_get_parsed_data(&h->snap),
+                            schemap, names);
     }
 }
 
@@ -950,6 +969,95 @@ print_raft_record(const struct raft_record *r,
     }
 }
 
+static void
+raft_header_to_standalone_log(const struct raft_header *h,
+                              struct ovsdb_log *db_log_data)
+{
+    if (h->snap_index) {
+        const struct json *data = raft_entry_get_parsed_data(&h->snap);
+
+        if (!data || json_array(data)->n != 2) {
+            ovs_fatal(0, "Incorrect raft header data array length");
+        }
+
+        struct json_array *pa = json_array(data);
+        struct json *schema_json = pa->elems[0];
+        struct ovsdb_error *error = NULL;
+
+        if (schema_json->type != JSON_NULL) {
+            struct ovsdb_schema *schema;
+            check_ovsdb_error(ovsdb_schema_from_json(schema_json, &schema));
+            ovsdb_schema_destroy(schema);
+            error = ovsdb_log_write(db_log_data, schema_json);
+        }
+
+        if (!error) {
+            struct json *data_json = pa->elems[1];
+            if (!data_json || data_json->type != JSON_OBJECT) {
+                ovs_fatal(0, "Invalid raft header data");
+            }
+            if (data_json->type != JSON_NULL) {
+                error = ovsdb_log_write(db_log_data, data_json);
+            }
+        }
+        check_ovsdb_error(error);
+    }
+}
+
+static void
+raft_record_to_standalone_log(const char *db_file_name,
+                              const struct raft_record *r,
+                              struct ovsdb_log *db_log_data)
+{
+    if (r->type == RAFT_REC_ENTRY) {
+        if (!r->entry.data) {
+            return;
+        }
+        struct json_array *pa = json_array(r->entry.data);
+
+        if (pa->n != 2) {
+            ovs_fatal(0, "Incorrect raft record array length");
+        }
+
+        struct json *schema_json = pa->elems[0];
+        struct json *data_json = pa->elems[1];
+
+        if (schema_json->type != JSON_NULL) {
+            /* This is a database conversion record.  Reset the log and
+             * write the new schema. */
+            struct ovsdb_schema *schema;
+
+            check_ovsdb_error(ovsdb_schema_from_json(schema_json, &schema));
+
+            if (data_json->type == JSON_NULL) {
+                /* We have a conversion request with no data.  There is no
+                 * other way as to read back what we have and convert. */
+                struct ovsdb *old_db, *new_db;
+
+                check_ovsdb_error(ovsdb_log_commit_block(db_log_data));
+
+                old_db = ovsdb_file_read(db_file_name, false);
+                check_ovsdb_error(ovsdb_convert(old_db, schema, &new_db));
+                ovsdb_destroy(old_db);
+
+                pa->elems[1] = ovsdb_to_txn_json(
+                                    new_db, "converted by ovsdb-tool");
+                ovsdb_destroy(new_db);
+
+                json_destroy(data_json);
+                data_json = pa->elems[1];
+            }
+
+            ovsdb_schema_destroy(schema);
+            check_ovsdb_error(ovsdb_log_reset(db_log_data));
+            check_ovsdb_error(ovsdb_log_write(db_log_data, schema_json));
+        }
+        if (data_json->type != JSON_NULL) {
+            check_ovsdb_error(ovsdb_log_write(db_log_data, data_json));
+        }
+    }
+}
+
 static void
 do_show_log_cluster(struct ovsdb_log *log)
 {
@@ -985,6 +1093,7 @@ do_show_log_cluster(struct ovsdb_log *log)
             free(s);
         }
 
+        json_destroy(json);
         putchar('\n');
     }
 
@@ -1009,6 +1118,7 @@ do_show_log(struct ovs_cmdl_context *ctx)
 }
 
 struct server {
+    struct ovsdb_log *log;
     const char *filename;
     const char *nickname;
 
@@ -1060,7 +1170,7 @@ get_server_name(const struct cluster *c, const struct uuid *sid,
                 char buf[SID_LEN + 1], size_t bufsize)
 {
     for (size_t i = 0; i < c->n_servers; i++) {
-        struct server *s = &c->servers[c->n_servers];
+        struct server *s = &c->servers[i];
         if (uuid_equals(&s->header.sid, sid)) {
             return s->filename;
         }
@@ -1172,19 +1282,26 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
         struct server *s = &c.servers[c.n_servers];
         s->filename = ctx->argv[i];
 
-        struct ovsdb_log *log;
         check_ovsdb_error(ovsdb_log_open(s->filename, RAFT_MAGIC,
-                                         OVSDB_LOG_READ_ONLY, -1, &log));
+                                         OVSDB_LOG_READ_ONLY, -1, &s->log));
 
         struct json *json;
-        check_ovsdb_error(ovsdb_log_read(log, &json));
+        check_ovsdb_error(ovsdb_log_read(s->log, &json));
         check_ovsdb_error(raft_header_from_json(&s->header, json));
         json_destroy(json);
 
         if (s->header.joining) {
             printf("%s has not joined the cluster, omitting\n", s->filename);
+            ovsdb_log_close(s->log);
             continue;
         }
+        for (size_t j = 0; j < c.n_servers; j++) {
+            if (uuid_equals(&s->header.sid, &c.servers[j].header.sid)) {
+                ovs_fatal(0, "Duplicate server ID "SID_FMT" in %s and %s.",
+                          SID_ARGS(&s->header.sid),
+                          s->filename, c.servers[j].filename);
+            }
+        }
         if (c.n_servers > 0) {
             struct server *s0 = &c.servers[0];
             if (!uuid_equals(&s0->header.cid, &s->header.cid)) {
@@ -1199,6 +1316,10 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
                           s->filename, s->header.name);
             }
         }
+        c.n_servers++;
+    }
+
+    for (struct server *s = c.servers; s < &c.servers[c.n_servers]; s++) {
         s->snap = &s->header.snap;
         s->log_start = s->log_end = s->header.snap_index + 1;
 
@@ -1217,7 +1338,9 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
                 s->records = x2nrealloc(s->records, &allocated_records,
                                         sizeof *s->records);
             }
-            check_ovsdb_error(ovsdb_log_read(log, &json));
+
+            struct json *json;
+            check_ovsdb_error(ovsdb_log_read(s->log, &json));
             if (!json) {
                 break;
             }
@@ -1238,7 +1361,6 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
                 max_term = term;
             }
 
-
             switch (r->type) {
             case RAFT_REC_ENTRY:
                 if (r->entry.index < commit_index) {
@@ -1289,7 +1411,7 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
                 }
                 struct raft_entry *e = &s->entries[log_idx];
                 e->term = r->term;
-                e->data = r->entry.data;
+                raft_entry_set_parsed_data_nocopy(e, r->entry.data);
                 e->eid = r->entry.eid;
                 e->servers = r->entry.servers;
                 break;
@@ -1367,9 +1489,8 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
             }
         }
 
-        ovsdb_log_close(log);
-
-        c.n_servers++;
+        ovsdb_log_close(s->log);
+        s->log = NULL;
     }
 
     /* Check the Leader Completeness property from Figure 3.2: If a log entry
@@ -1381,7 +1502,8 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
     struct commit *commit = NULL;
     for (uint64_t term = min_term; term <= max_term; term++) {
         struct leader *leader = find_leader(&c, term);
-        if (leader && commit && commit->index >= leader->log_end) {
+        if (leader && leader->log_end
+            && commit && commit->index >= leader->log_end) {
             ovs_fatal(0, "leader %s for term %"PRIu64" has log entries only "
                       "up to index %"PRIu64", but index %"PRIu64" was "
                       "committed in a previous term (e.g. by %s)",
@@ -1424,16 +1546,58 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
                 if (!must_equal || raft_entry_equals(ae, be)) {
                     continue;
                 }
-                char *as = json_to_string(raft_entry_to_json(ae), JSSF_SORT);
-                char *bs = json_to_string(raft_entry_to_json(be), JSSF_SORT);
+                struct json *jae = raft_entry_to_json(ae);
+                struct json *jbe = raft_entry_to_json(be);
+                char *as = json_to_string(jae, JSSF_SORT);
+                char *bs = json_to_string(jbe, JSSF_SORT);
                 ovs_fatal(0, "log entries with index %"PRIu64" differ:\n"
                           "%s has %s\n"
                           "%s has %s",
                           idx, a->filename, as, b->filename, bs);
+                json_destroy(jae);
+                json_destroy(jbe);
             }
         }
     }
 
+    /* Check for db consistency:
+     * The serverid must be in the servers list.
+     */
+
+    for (struct server *s = c.servers; s < &c.servers[c.n_servers]; s++) {
+        struct shash *servers_obj = json_object(s->snap->servers);
+        char *server_id = xasprintf(SID_FMT, SID_ARGS(&s->header.sid));
+        bool found = false;
+        const struct shash_node *node;
+
+        SHASH_FOR_EACH (node, servers_obj) {
+            if (!strncmp(server_id, node->name, SID_LEN)) {
+                found = true;
+            }
+        }
+
+        if (!found) {
+            for (struct raft_entry *e = s->entries;
+                 e < &s->entries[s->log_end - s->log_start]; e++) {
+                if (e->servers == NULL) {
+                    continue;
+                }
+                struct shash *log_servers_obj = json_object(e->servers);
+                SHASH_FOR_EACH (node, log_servers_obj) {
+                    if (!strncmp(server_id, node->name, SID_LEN)) {
+                        found = true;
+                    }
+                }
+            }
+        }
+
+        if (!found) {
+            ovs_fatal(0, "%s: server %s not found in server list",
+                      s->filename, server_id);
+        }
+        free(server_id);
+    }
+
     /* Clean up. */
 
     for (size_t i = 0; i < c.n_servers; i++) {
@@ -1450,15 +1614,14 @@ do_check_cluster(struct ovs_cmdl_context *ctx)
     }
     free(c.servers);
 
-    struct commit *next_commit;
-    HMAP_FOR_EACH_SAFE (commit, next_commit, hmap_node, &c.commits) {
+    HMAP_FOR_EACH_SAFE (commit, hmap_node, &c.commits) {
         hmap_remove(&c.commits, &commit->hmap_node);
         free(commit);
     }
     hmap_destroy(&c.commits);
 
-    struct leader *leader, *next_leader;
-    HMAP_FOR_EACH_SAFE (leader, next_leader, hmap_node, &c.leaders) {
+    struct leader *leader;
+    HMAP_FOR_EACH_SAFE (leader, hmap_node, &c.leaders) {
         hmap_remove(&c.leaders, &leader->hmap_node);
         free(leader);
     }
@@ -1506,6 +1669,53 @@ do_compare_versions(struct ovs_cmdl_context *ctx)
     exit(result ? 0 : 2);
 }
 
+static void
+do_convert_to_standalone(const char *db_file_name,
+                         struct ovsdb_log *log, struct ovsdb_log *db_log_data)
+{
+    for (unsigned int i = 0; ; i++) {
+        struct json *json;
+        check_ovsdb_error(ovsdb_log_read(log, &json));
+        if (!json) {
+            break;
+        }
+
+        if (i == 0) {
+            struct raft_header h;
+            check_ovsdb_error(raft_header_from_json(&h, json));
+            raft_header_to_standalone_log(&h, db_log_data);
+            raft_header_uninit(&h);
+        } else {
+            struct raft_record r;
+            check_ovsdb_error(raft_record_from_json(&r, json));
+            raft_record_to_standalone_log(db_file_name, &r, db_log_data);
+            raft_record_uninit(&r);
+        }
+        json_destroy(json);
+    }
+}
+
+static void
+do_cluster_standalone(struct ovs_cmdl_context *ctx)
+{
+    const char *db_file_name = ctx->argv[1];
+    const char *cluster_db_file_name = ctx->argv[2];
+    struct ovsdb_log *log;
+    struct ovsdb_log *db_log_data;
+
+    check_ovsdb_error(ovsdb_log_open(cluster_db_file_name,
+                                     OVSDB_MAGIC"|"RAFT_MAGIC,
+                                     OVSDB_LOG_READ_ONLY, -1, &log));
+    check_ovsdb_error(ovsdb_log_open(db_file_name, OVSDB_MAGIC,
+                                     OVSDB_LOG_CREATE_EXCL, -1, &db_log_data));
+    if (strcmp(ovsdb_log_get_magic(log), RAFT_MAGIC) != 0) {
+        ovs_fatal(0, "Database is not clustered db.\n");
+    }
+    do_convert_to_standalone(db_file_name, log, db_log_data);
+    check_ovsdb_error(ovsdb_log_commit_block(db_log_data));
+    ovsdb_log_close(db_log_data);
+    ovsdb_log_close(log);
+}
 
 static void
 do_help(struct ovs_cmdl_context *ctx OVS_UNUSED)
@@ -1545,7 +1755,9 @@ static const struct ovs_cmdl_command all_commands[] = {
     { "compare-versions", "a op b", 3, 3, do_compare_versions, OVS_RO },
     { "help", NULL, 0, INT_MAX, do_help, OVS_RO },
     { "list-commands", NULL, 0, INT_MAX, do_list_commands, OVS_RO },
-    { NULL, NULL, 0, 0, NULL, OVS_RO },
+    { "cluster-to-standalone", "db clusterdb", 2, 2,
+    do_cluster_standalone, OVS_RW },
+    { NULL, NULL, 2, 2, NULL, OVS_RO },
 };
 
 static const struct ovs_cmdl_command *get_all_commands(void)
