@@ -1,4 +1,4 @@
-/* Copyright (c) 2009, 2010, 2011, 2012, 2013 Nicira, Inc.
+/* Copyright (c) 2009, 2010, 2011, 2012, 2013, 2017 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,13 +20,13 @@
 #include "openvswitch/hmap.h"
 #include "openvswitch/list.h"
 #include "openvswitch/shash.h"
+#include "openvswitch/uuid.h"
 
 struct json;
 struct ovsdb_log;
 struct ovsdb_session;
 struct ovsdb_txn;
 struct simap;
-struct uuid;
 
 /* Database schema. */
 struct ovsdb_schema {
@@ -45,54 +45,86 @@ void ovsdb_schema_destroy(struct ovsdb_schema *);
 struct ovsdb_error *ovsdb_schema_from_file(const char *file_name,
                                            struct ovsdb_schema **)
     OVS_WARN_UNUSED_RESULT;
-struct ovsdb_error *ovsdb_schema_from_json(struct json *,
+struct ovsdb_error *ovsdb_schema_from_json(const struct json *,
                                            struct ovsdb_schema **)
     OVS_WARN_UNUSED_RESULT;
 struct json *ovsdb_schema_to_json(const struct ovsdb_schema *);
 
 bool ovsdb_schema_equal(const struct ovsdb_schema *,
                         const struct ovsdb_schema *);
+
+struct ovsdb_error *ovsdb_schema_check_for_ephemeral_columns(
+    const struct ovsdb_schema *) OVS_WARN_UNUSED_RESULT;
+void ovsdb_schema_persist_ephemeral_columns(struct ovsdb_schema *,
+                                            const char *filename);
+
+struct ovsdb_version {
+    unsigned int x;
+    unsigned int y;
+    unsigned int z;
+};
+bool ovsdb_parse_version(const char *, struct ovsdb_version *);
+bool ovsdb_is_valid_version(const char *);
 
 /* Database. */
+struct ovsdb_txn_history_node {
+    struct ovs_list node; /* Element in struct ovsdb's txn_history list */
+    struct ovsdb_txn *txn;
+};
+
 struct ovsdb {
+    char *name;
     struct ovsdb_schema *schema;
-    struct ovs_list replicas;   /* Contains "struct ovsdb_replica"s. */
+    struct ovsdb_storage *storage; /* If nonnull, log for transactions. */
+    struct uuid prereq;
+    struct ovs_list monitors;   /* Contains "struct ovsdb_monitor"s. */
     struct shash tables;        /* Contains "struct ovsdb_table *"s. */
 
     /* Triggers. */
     struct ovs_list triggers;   /* Contains "struct ovsdb_trigger"s. */
     bool run_triggers;
+    bool run_triggers_now;
+
+    struct ovsdb_table *rbac_role;
+
+    /* History trasanctions for incremental monitor transfer. */
+    bool need_txn_history;     /* Need to maintain history of transactions. */
+    unsigned int n_txn_history; /* Current number of history transactions. */
+    unsigned int n_txn_history_atoms; /* Total number of atoms in history. */
+    struct ovs_list txn_history; /* Contains "struct ovsdb_txn_history_node. */
+
+    size_t n_atoms;  /* Total number of ovsdb atoms in the database. */
+
+    /* Relay mode. */
+    bool is_relay;  /* True, if database is in relay mode. */
+    /* List that holds transactions waiting to be forwarded to the server. */
+    struct ovs_list txn_forward_new;
+    /* Hash map for transactions that are already sent and waits for reply. */
+    struct hmap txn_forward_sent;
 };
 
-struct ovsdb *ovsdb_create(struct ovsdb_schema *);
+struct ovsdb *ovsdb_create(struct ovsdb_schema *, struct ovsdb_storage *);
 void ovsdb_destroy(struct ovsdb *);
 
 void ovsdb_get_memory_usage(const struct ovsdb *, struct simap *usage);
 
 struct ovsdb_table *ovsdb_get_table(const struct ovsdb *, const char *);
 
+struct ovsdb_txn *ovsdb_execute_compose(
+    struct ovsdb *, const struct ovsdb_session *, const struct json *params,
+    bool read_only, const char *role, const char *id,
+    long long int elapsed_msec, long long int *timeout_msec,
+    bool *durable, bool *forwarding_needed, struct json **);
+
 struct json *ovsdb_execute(struct ovsdb *, const struct ovsdb_session *,
                            const struct json *params, bool read_only,
+                           const char *role, const char *id,
                            long long int elapsed_msec,
                            long long int *timeout_msec);
-
-/* Database replication. */
 
-struct ovsdb_replica {
-    struct ovs_list node;       /* Element in "struct ovsdb" replicas list. */
-    const struct ovsdb_replica_class *class;
-};
+struct ovsdb_error *ovsdb_snapshot(struct ovsdb *, bool trim_memory)
+    OVS_WARN_UNUSED_RESULT;
 
-struct ovsdb_replica_class {
-    struct ovsdb_error *(*commit)(struct ovsdb_replica *,
-                                  const struct ovsdb_txn *, bool durable);
-    void (*destroy)(struct ovsdb_replica *);
-};
-
-void ovsdb_replica_init(struct ovsdb_replica *,
-                        const struct ovsdb_replica_class *);
-
-void ovsdb_add_replica(struct ovsdb *, struct ovsdb_replica *);
-void ovsdb_remove_replica(struct ovsdb *, struct ovsdb_replica *);
+void ovsdb_replace(struct ovsdb *dst, struct ovsdb *src);
 
 #endif /* ovsdb/ovsdb.h */
