@@ -25,6 +25,7 @@
 #include "ovsdb-parser.h"
 #include "ovsdb-types.h"
 #include "row.h"
+#include "transaction.h"
 
 static void
 add_column(struct ovsdb_table_schema *ts, struct ovsdb_column *column)
@@ -152,7 +153,7 @@ ovsdb_table_schema_from_json(const struct json *json, const char *name,
             return ovsdb_syntax_error(json, NULL,
                                       "maxRows must be at least 1");
         }
-        n_max_rows = max_rows->u.integer;
+        n_max_rows = max_rows->integer;
     } else {
         n_max_rows = UINT_MAX;
     }
@@ -187,12 +188,12 @@ ovsdb_table_schema_from_json(const struct json *json, const char *name,
     if (indexes) {
         size_t i;
 
-        ts->indexes = xmalloc(indexes->u.array.n * sizeof *ts->indexes);
-        for (i = 0; i < indexes->u.array.n; i++) {
+        ts->indexes = xmalloc(indexes->array.n * sizeof *ts->indexes);
+        for (i = 0; i < indexes->array.n; i++) {
             struct ovsdb_column_set *index = &ts->indexes[i];
             size_t j;
 
-            error = ovsdb_column_set_from_json(indexes->u.array.elems[i],
+            error = ovsdb_column_set_from_json(indexes->array.elems[i],
                                                ts, index);
             if (error) {
                 goto error;
@@ -308,10 +309,10 @@ void
 ovsdb_table_destroy(struct ovsdb_table *table)
 {
     if (table) {
-        struct ovsdb_row *row, *next;
+        struct ovsdb_row *row;
         size_t i;
 
-        HMAP_FOR_EACH_SAFE (row, next, hmap_node, &table->rows) {
+        HMAP_FOR_EACH_SAFE (row, hmap_node, &table->rows) {
             ovsdb_row_destroy(row);
         }
         hmap_destroy(&table->rows);
@@ -339,3 +340,73 @@ ovsdb_table_get_row(const struct ovsdb_table *table, const struct uuid *uuid)
 
     return NULL;
 }
+
+struct ovsdb_error *
+ovsdb_table_execute_insert(struct ovsdb_txn *txn, const struct uuid *row_uuid,
+                           struct ovsdb_table *table, struct json *json_row)
+{
+    const struct ovsdb_row *old_row = ovsdb_table_get_row(table, row_uuid);
+    if (old_row) {
+        return ovsdb_error(
+                    "consistency violation",
+                    "cannot delete missing row "UUID_FMT" from table %s",
+                    UUID_ARGS(row_uuid), table->schema->name);
+    }
+
+    struct ovsdb_row *row = ovsdb_row_create(table);
+
+    struct ovsdb_error *error = ovsdb_row_from_json(row, json_row,
+                                                    NULL, NULL, false);
+    if (!error) {
+        *ovsdb_row_get_uuid_rw(row) = *row_uuid;
+        ovsdb_txn_row_insert(txn, row);
+    } else {
+        ovsdb_row_destroy(row);
+    }
+
+    return error;
+}
+
+struct ovsdb_error *
+ovsdb_table_execute_delete(struct ovsdb_txn *txn, const struct uuid *row_uuid,
+                           struct ovsdb_table *table)
+{
+    const struct ovsdb_row *row = ovsdb_table_get_row(table, row_uuid);
+    if (!row) {
+        return ovsdb_error(
+                    "consistency violation",
+                    "cannot delete missing row "UUID_FMT" from table %s",
+                    UUID_ARGS(row_uuid), table->schema->name);
+    }
+
+    ovsdb_txn_row_delete(txn, row);
+    return NULL;
+}
+
+struct ovsdb_error *
+ovsdb_table_execute_update(struct ovsdb_txn *txn, const struct uuid *row_uuid,
+                           struct ovsdb_table *table, struct json *json_row,
+                           bool xor)
+{
+    const struct ovsdb_row *row = ovsdb_table_get_row(table, row_uuid);
+    if (!row) {
+        return ovsdb_error(
+                    "consistency violation",
+                    "cannot modify missing row "UUID_FMT" from table %s",
+                    UUID_ARGS(row_uuid), table->schema->name);
+    }
+
+    struct ovsdb_column_set columns = OVSDB_COLUMN_SET_INITIALIZER;
+    struct ovsdb_row *update = ovsdb_row_create(table);
+    struct ovsdb_error *error = ovsdb_row_from_json(update, json_row,
+                                                    NULL, &columns, xor);
+
+    if (!error && (xor || !ovsdb_row_equal_columns(row, update, &columns))) {
+        error = ovsdb_row_update_columns(ovsdb_txn_row_modify(txn, row),
+                                         update, &columns, xor);
+    }
+
+    ovsdb_column_set_destroy(&columns);
+    ovsdb_row_destroy(update);
+    return error;
+}
