@@ -314,7 +314,8 @@ update_row_ref_count(struct ovsdb_txn *txn, struct ovsdb_txn_row *r)
         const struct ovsdb_column *column = node->data;
         struct ovsdb_error *error;
 
-        if (bitmap_is_set(r->changed, column->index)) {
+        if (bitmap_is_set(r->changed, column->index)
+            && ovsdb_type_has_strong_refs(&column->type)) {
             if (r->old && !r->new) {
                 error = ovsdb_txn_adjust_row_refs(
                             txn, r->old, column,
@@ -671,6 +672,10 @@ assess_weak_refs(struct ovsdb_txn *txn, struct ovsdb_txn_row *txn_row)
         unsigned int orig_n;
         bool zero = false;
 
+        if (!ovsdb_type_has_weak_refs(&column->type)) {
+            continue;
+        }
+
         orig_n = datum->n;
 
         /* Collecting all key-value pairs that references deleted rows. */
@@ -692,13 +697,17 @@ assess_weak_refs(struct ovsdb_txn *txn, struct ovsdb_txn_row *txn_row)
         ovsdb_datum_destroy(&deleted_refs, &column->type);
 
         /* Generating the difference between old and new data. */
-        if (txn_row->old) {
-            ovsdb_datum_added_removed(&added, &removed,
-                                      &txn_row->old->fields[column->index],
-                                      datum, &column->type);
-        } else {
-            ovsdb_datum_init_empty(&removed);
-            ovsdb_datum_clone(&added, datum, &column->type);
+        ovsdb_datum_init_empty(&added);
+        ovsdb_datum_init_empty(&removed);
+        if (datum->n != orig_n
+            || bitmap_is_set(txn_row->changed, column->index)) {
+            if (txn_row->old) {
+                ovsdb_datum_added_removed(&added, &removed,
+                                          &txn_row->old->fields[column->index],
+                                          datum, &column->type);
+            } else {
+                ovsdb_datum_clone(&added, datum, &column->type);
+            }
         }
 
         /* Checking added data and creating new references. */
