@@ -1,4 +1,4 @@
-# Copyright (c) 2010, 2011 Nicira Networks
+# Copyright (c) 2010, 2011, 2012 Nicira, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,6 +14,7 @@
 
 import os
 
+import ovs.util
 import ovs.vlog
 
 # Values returned by Reconnect.run()
@@ -21,7 +22,7 @@ CONNECT = 'connect'
 DISCONNECT = 'disconnect'
 PROBE = 'probe'
 
-EOF = -1
+EOF = ovs.util.EOF
 vlog = ovs.vlog.Vlog("reconnect")
 
 
@@ -43,7 +44,7 @@ class Reconnect(object):
         is_connected = False
 
         @staticmethod
-        def deadline(fsm):
+        def deadline(fsm, now):
             return None
 
         @staticmethod
@@ -55,7 +56,7 @@ class Reconnect(object):
         is_connected = False
 
         @staticmethod
-        def deadline(fsm):
+        def deadline(fsm, now):
             return None
 
         @staticmethod
@@ -67,7 +68,7 @@ class Reconnect(object):
         is_connected = False
 
         @staticmethod
-        def deadline(fsm):
+        def deadline(fsm, now):
             return fsm.state_entered + fsm.backoff
 
         @staticmethod
@@ -79,7 +80,7 @@ class Reconnect(object):
         is_connected = False
 
         @staticmethod
-        def deadline(fsm):
+        def deadline(fsm, now):
             return fsm.state_entered + max(1000, fsm.backoff)
 
         @staticmethod
@@ -91,17 +92,31 @@ class Reconnect(object):
         is_connected = True
 
         @staticmethod
-        def deadline(fsm):
+        def deadline(fsm, now):
             if fsm.probe_interval:
-                base = max(fsm.last_received, fsm.state_entered)
-                return base + fsm.probe_interval
+                base = max(fsm.last_activity, fsm.state_entered)
+                expiration = base + fsm.probe_interval
+                if (now < expiration or
+                    fsm.last_receive_attempt is None or
+                    fsm.last_receive_attempt >= expiration):
+                    # We still have time before the expiration or the time has
+                    # already passed and there was no activity.  In the first
+                    # case we need to wait for the expiration, in the second -
+                    # we're already past the deadline. */
+                    return expiration
+                else:
+                    # Time has already passed, but we didn't attempt to receive
+                    # anything.  We need to wake up and try to receive even if
+                    # nothing is pending, so we can update the expiration time
+                    # or transition to a different state.
+                    return now + 1
             return None
 
         @staticmethod
         def run(fsm, now):
             vlog.dbg("%s: idle %d ms, sending inactivity probe"
                      % (fsm.name,
-                        now - max(fsm.last_received, fsm.state_entered)))
+                        now - max(fsm.last_activity, fsm.state_entered)))
             fsm._transition(now, Reconnect.Idle)
             return PROBE
 
@@ -110,8 +125,16 @@ class Reconnect(object):
         is_connected = True
 
         @staticmethod
-        def deadline(fsm):
-            return fsm.state_entered + fsm.probe_interval
+        def deadline(fsm, now):
+            if fsm.probe_interval:
+                expiration = fsm.state_entered + fsm.probe_interval
+                if (now < expiration or
+                    fsm.last_receive_attempt is None or
+                    fsm.last_receive_attempt >= expiration):
+                    return expiration
+                else:
+                    return now + 1
+            return None
 
         @staticmethod
         def run(fsm, now):
@@ -125,7 +148,7 @@ class Reconnect(object):
         is_connected = False
 
         @staticmethod
-        def deadline(fsm):
+        def deadline(fsm, now):
             return fsm.state_entered
 
         @staticmethod
@@ -147,10 +170,12 @@ class Reconnect(object):
         self.state = Reconnect.Void
         self.state_entered = now
         self.backoff = 0
-        self.last_received = now
+        self.last_activity = now
         self.last_connected = None
         self.last_disconnected = None
+        self.last_receive_attempt = now
         self.max_tries = None
+        self.backoff_free_tries = 0
 
         self.creation_time = now
         self.n_attempted_connections = 0
@@ -201,8 +226,8 @@ class Reconnect(object):
         """Returns the "probe interval" in milliseconds.  If this is zero, it
         disables the connection keepalive feature.  If it is nonzero, then if
         the interval passes while the FSM is connected and without
-        self.received() being called, self.run() returns ovs.reconnect.PROBE.
-        If the interval passes again without self.received() being called,
+        self.activity() being called, self.run() returns ovs.reconnect.PROBE.
+        If the interval passes again without self.activity() being called,
         self.run() returns ovs.reconnect.DISCONNECT."""
         return self.probe_interval
 
@@ -237,15 +262,21 @@ class Reconnect(object):
 
         if (self.state == Reconnect.Backoff and
             self.backoff > self.max_backoff):
-                self.backoff = self.max_backoff
+            self.backoff = self.max_backoff
+
+    def set_backoff_free_tries(self, backoff_free_tries):
+        """Sets the number of connection attempts that will be made without
+        backoff to 'backoff_free_tries'.  Values 0 and 1 both
+        represent a single attempt."""
+        self.backoff_free_tries = backoff_free_tries
 
     def set_probe_interval(self, probe_interval):
         """Sets the "probe interval" to 'probe_interval', in milliseconds.  If
         this is zero, it disables the connection keepalive feature.  If it is
         nonzero, then if the interval passes while this FSM is connected and
-        without self.received() being called, self.run() returns
+        without self.activity() being called, self.run() returns
         ovs.reconnect.PROBE.  If the interval passes again without
-        self.received() being called, self.run() returns
+        self.activity() being called, self.run() returns
         ovs.reconnect.DISCONNECT.
 
         If 'probe_interval' is nonzero, then it will be forced to a value of at
@@ -334,7 +365,10 @@ class Reconnect(object):
                 else:
                     self.info_level("%s: error listening for connections"
                                     % self.name)
-            else:
+            elif self.state == Reconnect.Reconnect:
+                self.info_level("%s: connection closed by client"
+                                % self.name)
+            elif self.backoff < self.max_backoff:
                 if self.passive:
                     type_ = "listen"
                 else:
@@ -349,9 +383,16 @@ class Reconnect(object):
             if (self.state in (Reconnect.Active, Reconnect.Idle)):
                 self.last_disconnected = now
 
+            if not self.__may_retry():
+                self._transition(now, Reconnect.Void)
+                return
+
             # Back off
-            if (self.state in (Reconnect.Active, Reconnect.Idle) and
-                (self.last_received - self.last_connected >= self.backoff or
+            if self.backoff_free_tries > 1:
+                self.backoff_free_tries -= 1
+                self.backoff = 0
+            elif (self.state in (Reconnect.Active, Reconnect.Idle) and
+                (self.last_activity - self.last_connected >= self.backoff or
                  self.passive)):
                 if self.passive:
                     self.backoff = 0
@@ -360,23 +401,26 @@ class Reconnect(object):
             else:
                 if self.backoff < self.min_backoff:
                     self.backoff = self.min_backoff
-                elif self.backoff >= self.max_backoff / 2:
-                    self.backoff = self.max_backoff
-                else:
+                elif self.backoff < self.max_backoff / 2:
                     self.backoff *= 2
-
-                if self.passive:
-                    self.info_level("%s: waiting %.3g seconds before trying "
-                                    "to listen again"
-                                    % (self.name, self.backoff / 1000.0))
+                    if self.passive:
+                        action = "trying to listen again"
+                    else:
+                        action = "reconnect"
+                    self.info_level("%s: waiting %.3g seconds before %s"
+                                    % (self.name, self.backoff / 1000.0,
+                                       action))
                 else:
-                    self.info_level("%s: waiting %.3g seconds before reconnect"
-                                    % (self.name, self.backoff / 1000.0))
-
-            if self.__may_retry():
-                self._transition(now, Reconnect.Backoff)
-            else:
-                self._transition(now, Reconnect.Void)
+                    if self.backoff < self.max_backoff:
+                        if self.passive:
+                            action = "try to listen"
+                        else:
+                            action = "reconnect"
+                        self.info_level("%s: continuing to %s in the "
+                                        "background but suppressing further "
+                                        "logging" % (self.name, action))
+                    self.backoff = self.max_backoff
+            self._transition(now, Reconnect.Backoff)
 
     def connecting(self, now):
         """Tell this FSM that a connection or listening attempt is in progress.
@@ -387,7 +431,7 @@ class Reconnect(object):
         if self.state != Reconnect.ConnectInProgress:
             if self.passive:
                 self.info_level("%s: listening..." % self.name)
-            else:
+            elif self.backoff < self.max_backoff:
                 self.info_level("%s: connecting..." % self.name)
             self._transition(now, Reconnect.ConnectInProgress)
 
@@ -423,7 +467,7 @@ class Reconnect(object):
         """Tell this FSM that the connection was successful.
 
         The FSM will start the probe interval timer, which is reset by
-        self.received().  If the timer expires, a probe will be sent (by
+        self.activity().  If the timer expires, a probe will be sent (by
         returning ovs.reconnect.PROBE from self.run().  If the timer expires
         again without being reset, the connection will be aborted (by returning
         ovs.reconnect.DISCONNECT from self.run()."""
@@ -441,12 +485,23 @@ class Reconnect(object):
         self.connecting(now)
         self.disconnected(now, error)
 
-    def received(self, now):
-        """Tell this FSM that some data was received.  This resets the probe
-        interval timer, so that the connection is known not to be idle."""
+    def activity(self, now):
+        """Tell this FSM that some activity occurred on the connection.  This
+        resets the probe interval timer, so that the connection is known not to
+        be idle."""
         if self.state != Reconnect.Active:
             self._transition(now, Reconnect.Active)
-        self.last_received = now
+        self.last_activity = now
+
+    def receive_attempted(self, now):
+        """Tell 'fsm' that some attempt to receive data on the connection was
+        made at 'now'.  The FSM only allows probe interval timer to expire when
+        some attempt to receive data on the connection was received after the
+        time when it should have expired.  This helps in the case where there's
+        a long delay in the poll loop and then reconnect_run() executes before
+        the code to try to receive anything from the remote runs.  (To disable
+        this feature, pass None for 'now'.)"""
+        self.last_receive_attempt = now
 
     def _transition(self, now, state):
         if self.state == Reconnect.ConnectInProgress:
@@ -503,7 +558,9 @@ class Reconnect(object):
               connection is indeed in working order.  (This will only be
               returned if the "probe interval" is nonzero--see
               self.set_probe_interval())."""
-        if now >= self.state.deadline(self):
+
+        deadline = self.state.deadline(self, now)
+        if deadline is not None and now >= deadline:
             return self.state.run(self, now)
         else:
             return None
@@ -512,14 +569,14 @@ class Reconnect(object):
         """Causes the next call to poller.block() to wake up when self.run()
         should be called."""
         timeout = self.timeout(now)
-        if timeout >= 0:
+        if timeout is not None and timeout >= 0:
             poller.timer_wait(timeout)
 
     def timeout(self, now):
         """Returns the number of milliseconds after which self.run() should be
         called if nothing else notable happens in the meantime, or None if this
         is currently unnecessary."""
-        deadline = self.state.deadline(self)
+        deadline = self.state.deadline(self, now)
         if deadline is not None:
             remaining = deadline - now
             return max(0, remaining)
@@ -556,7 +613,7 @@ class Reconnect(object):
         stats.creation_time = self.creation_time
         stats.last_connected = self.last_connected
         stats.last_disconnected = self.last_disconnected
-        stats.last_received = self.last_received
+        stats.last_activity = self.last_activity
         stats.backoff = self.backoff
         stats.seqno = self.seqno
         stats.is_connected = self.is_connected()
