@@ -1,4 +1,4 @@
-# Copyright (c) 2010, 2012 Nicira, Inc.
+# Copyright (c) 2010, 2012, 2014, 2015 Nicira, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,7 +16,6 @@ import errno
 import os
 import os.path
 import random
-import select
 import socket
 import sys
 
@@ -24,6 +23,15 @@ import ovs.fatal_signal
 import ovs.poller
 import ovs.vlog
 
+try:
+    import ssl
+except ImportError:
+    ssl = None
+
+if sys.platform == 'win32':
+    import ovs.winutils as winutils
+    import win32file
+
 vlog = ovs.vlog.Vlog("socket_util")
 
 
@@ -33,14 +41,14 @@ def make_short_name(long_name):
     long_name = os.path.abspath(long_name)
     long_dirname = os.path.dirname(long_name)
     tmpdir = os.getenv('TMPDIR', '/tmp')
-    for x in xrange(0, 1000):
+    for x in range(0, 1000):
         link_name = \
             '%s/ovs-un-py-%d-%d' % (tmpdir, random.randint(0, 10000), x)
         try:
             os.symlink(long_dirname, link_name)
             ovs.fatal_signal.add_file_to_unlink(link_name)
             return os.path.join(link_name, os.path.basename(long_name))
-        except OSError, e:
+        except OSError as e:
             if e.errno != errno.EEXIST:
                 break
     raise Exception("Failed to create temporary symlink")
@@ -65,7 +73,7 @@ def make_unix_socket(style, nonblock, bind_path, connect_path, short=False):
 
     try:
         sock = socket.socket(socket.AF_UNIX, style)
-    except socket.error, e:
+    except socket.error as e:
         return get_exception_errno(e), None
 
     try:
@@ -75,7 +83,7 @@ def make_unix_socket(style, nonblock, bind_path, connect_path, short=False):
             # Delete bind_path but ignore ENOENT.
             try:
                 os.unlink(bind_path)
-            except OSError, e:
+            except OSError as e:
                 if e.errno != errno.ENOENT:
                     return e.errno, None
 
@@ -83,20 +91,17 @@ def make_unix_socket(style, nonblock, bind_path, connect_path, short=False):
             sock.bind(bind_path)
 
             try:
-                if sys.hexversion >= 0x02060000:
-                    os.fchmod(sock.fileno(), 0700)
-                else:
-                    os.chmod("/dev/fd/%d" % sock.fileno(), 0700)
-            except OSError, e:
+                os.fchmod(sock.fileno(), 0o700)
+            except OSError:
                 pass
         if connect_path is not None:
             try:
                 sock.connect(connect_path)
-            except socket.error, e:
+            except socket.error as e:
                 if get_exception_errno(e) != errno.EINPROGRESS:
                     raise
         return 0, sock
-    except socket.error, e:
+    except socket.error as e:
         sock.close()
         if (bind_path is not None and
             os.path.exists(bind_path)):
@@ -113,22 +118,26 @@ def make_unix_socket(style, nonblock, bind_path, connect_path, short=False):
                 dirname = os.path.dirname(connect_path)
                 basename = os.path.basename(connect_path)
                 try:
-                    connect_dirfd = os.open(dirname, os.O_DIRECTORY | os.O_RDONLY)
-                except OSError, err:
+                    connect_dirfd = os.open(dirname,
+                                            os.O_DIRECTORY | os.O_RDONLY)
+                except OSError as err:
                     return get_exception_errno(err), None
-                short_connect_path = "/proc/self/fd/%d/%s" % (connect_dirfd, basename)
+                short_connect_path = "/proc/self/fd/%d/%s" % (connect_dirfd,
+                                                              basename)
 
             if bind_path is not None:
                 dirname = os.path.dirname(bind_path)
                 basename = os.path.basename(bind_path)
                 try:
                     bind_dirfd = os.open(dirname, os.O_DIRECTORY | os.O_RDONLY)
-                except OSError, err:
+                except OSError as err:
                     return get_exception_errno(err), None
-                short_bind_path = "/proc/self/fd/%d/%s" % (bind_dirfd, basename)
+                short_bind_path = "/proc/self/fd/%d/%s" % (bind_dirfd,
+                                                           basename)
 
             try:
-                return make_unix_socket(style, nonblock, short_bind_path, short_connect_path)
+                return make_unix_socket(style, nonblock, short_bind_path,
+                                        short_connect_path)
             finally:
                 if connect_dirfd is not None:
                     os.close(connect_dirfd)
@@ -155,21 +164,37 @@ def make_unix_socket(style, nonblock, bind_path, connect_path, short=False):
 
 
 def check_connection_completion(sock):
-    p = ovs.poller.SelectPoll()
-    p.register(sock, ovs.poller.POLLOUT)
+    if sys.platform == "win32":
+        p = ovs.poller.SelectPoll()
+        event = winutils.get_new_event(None, False, True, None)
+        # Receive notification of readiness for writing, of completed
+        # connection or multipoint join operation, and of socket closure.
+        win32file.WSAEventSelect(sock, event,
+                                 win32file.FD_WRITE |
+                                 win32file.FD_CONNECT |
+                                 win32file.FD_CLOSE)
+        p.register(event, ovs.poller.POLLOUT)
+    else:
+        p = ovs.poller.get_system_poll()
+        p.register(sock, ovs.poller.POLLOUT)
     pfds = p.poll(0)
     if len(pfds) == 1:
         revents = pfds[0][1]
-        if revents & ovs.poller.POLLERR:
+        if revents & ovs.poller.POLLERR or revents & ovs.poller.POLLHUP:
             try:
                 # The following should raise an exception.
-                socket.send("\0", socket.MSG_DONTWAIT)
+                if ssl and isinstance(sock, ssl.SSLSocket):
+                    # SSL wrapped socket does not allow
+                    # non-zero optional flag.
+                    sock.send("\0".encode())
+                else:
+                    sock.send("\0".encode(), socket.MSG_DONTWAIT)
 
                 # (Here's where we end up if it didn't.)
                 # XXX rate-limit
                 vlog.err("poll return POLLERR but send succeeded")
                 return errno.EPROTO
-            except socket.error, e:
+            except socket.error as e:
                 return get_exception_errno(e)
         else:
             return 0
@@ -177,39 +202,74 @@ def check_connection_completion(sock):
         return errno.EAGAIN
 
 
+def is_valid_ipv4_address(address):
+    try:
+        socket.inet_pton(socket.AF_INET, address)
+    except AttributeError:
+        try:
+            socket.inet_aton(address)
+        except socket.error:
+            return False
+    except socket.error:
+        return False
+
+    return True
+
+
 def inet_parse_active(target, default_port):
     address = target.split(":")
-    host_name = address[0]
+    if len(address) >= 2:
+        host_name = ":".join(address[0:-1]).lstrip('[').rstrip(']')
+        port = int(address[-1])
+    else:
+        if default_port:
+            port = default_port
+        else:
+            raise ValueError("%s: port number must be specified" % target)
+        host_name = address[0]
     if not host_name:
         raise ValueError("%s: bad peer name format" % target)
-    if len(address) >= 2:
-        port = int(address[1])
-    elif default_port:
-        port = default_port
-    else:
-        raise ValueError("%s: port number must be specified" % target)
     return (host_name, port)
 
 
+def inet_create_socket_active(style, address):
+    try:
+        is_addr_inet = is_valid_ipv4_address(address[0])
+        if is_addr_inet:
+            sock = socket.socket(socket.AF_INET, style, 0)
+            family = socket.AF_INET
+        else:
+            sock = socket.socket(socket.AF_INET6, style, 0)
+            family = socket.AF_INET6
+    except socket.error as e:
+        return get_exception_errno(e), None
+
+    return family, sock
+
+
+def inet_connect_active(sock, address, family, dscp):
+    try:
+        set_nonblocking(sock)
+        set_dscp(sock, family, dscp)
+        error = sock.connect_ex(address)
+        if error not in (0, errno.EINPROGRESS, errno.EWOULDBLOCK):
+            sock.close()
+            return error
+        return 0
+    except socket.error as e:
+        sock.close()
+        return get_exception_errno(e)
+
+
 def inet_open_active(style, target, default_port, dscp):
     address = inet_parse_active(target, default_port)
-    try:
-        sock = socket.socket(socket.AF_INET, style, 0)
-    except socket.error, e:
-        return get_exception_errno(e), None
-
-    try:
-        set_nonblocking(sock)
-        set_dscp(sock, dscp)
-        try:
-            sock.connect(address)
-        except socket.error, e:
-            if get_exception_errno(e) != errno.EINPROGRESS:
-                raise
-        return 0, sock
-    except socket.error, e:
-        sock.close()
-        return get_exception_errno(e), None
+    family, sock = inet_create_socket_active(style, address)
+    if sock is None:
+        return family, sock
+    error = inet_connect_active(sock, address, family, dscp)
+    if error:
+        return error, None
+    return 0, sock
 
 
 def get_exception_errno(e):
@@ -217,7 +277,7 @@ def get_exception_errno(e):
     exception is documented as having two completely different forms of
     arguments: either a string or a (errno, string) tuple.  We only want the
     errno."""
-    if type(e.args) == tuple:
+    if isinstance(e.args, tuple):
         return e.args[0]
     else:
         return errno.EPROTO
@@ -233,9 +293,12 @@ def get_null_fd():
     global null_fd
     if null_fd < 0:
         try:
-            null_fd = os.open("/dev/null", os.O_RDWR)
-        except OSError, e:
-            vlog.err("could not open /dev/null: %s" % os.strerror(e.errno))
+            # os.devnull ensures compatibility with Windows, returns
+            # '/dev/null' for Unix and 'nul' for Windows
+            null_fd = os.open(os.devnull, os.O_RDWR)
+        except OSError as e:
+            vlog.err("could not open %s: %s" % (os.devnull,
+                                                os.strerror(e.errno)))
             return -e.errno
     return null_fd
 
@@ -248,6 +311,8 @@ def write_fully(fd, buf):
     bytes_written = 0
     if len(buf) == 0:
         return 0, 0
+    if not isinstance(buf, bytes):
+        buf = bytes(buf, 'utf-8')
     while True:
         try:
             retval = os.write(fd, buf)
@@ -260,20 +325,26 @@ def write_fully(fd, buf):
             else:
                 bytes_written += retval
                 buf = buf[:retval]
-        except OSError, e:
+        except OSError as e:
             return e.errno, bytes_written
 
 
 def set_nonblocking(sock):
     try:
         sock.setblocking(0)
-    except socket.error, e:
+    except socket.error as e:
         vlog.err("could not set nonblocking mode on socket: %s"
                  % os.strerror(get_exception_errno(e)))
 
 
-def set_dscp(sock, dscp):
+def set_dscp(sock, family, dscp):
     if dscp > 63:
         raise ValueError("Invalid dscp %d" % dscp)
+
     val = dscp << 2
-    sock.setsockopt(socket.IPPROTO_IP, socket.IP_TOS, val)
+    if family == socket.AF_INET:
+        sock.setsockopt(socket.IPPROTO_IP, socket.IP_TOS, val)
+    elif family == socket.AF_INET6:
+        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_TCLASS, val)
+    else:
+        raise ValueError('Invalid family %d' % family)
