@@ -37,7 +37,7 @@ vlog.init(None)
 
 
 def unbox_json(json):
-    if type(json) == list and len(json) == 1:
+    if type(json) is list and len(json) == 1:
         return json[0]
     else:
         return json
@@ -232,75 +232,87 @@ def get_singleton_table_printable_row(row):
     return "name=%s" % row.name
 
 
-def print_row(table, row, step, contents):
-    s = "%03d: table %s: %s " % (step, table, contents)
-    s += get_simple_printable_row_string(row, ["uuid"])
+def print_row(table, row, step, contents, terse):
+    if terse:
+        s = "%03d: table %s" % (step, table)
+    else:
+        s = "%03d: table %s: %s " % (step, table, contents)
+        s += get_simple_printable_row_string(row, ["uuid"])
     print(s)
 
 
-def print_idl(idl, step):
+def print_idl(idl, step, terse=False):
     n = 0
     if "simple" in idl.tables:
         simple = idl.tables["simple"].rows
         for row in simple.values():
             print_row("simple", row, step,
-                      get_simple_table_printable_row(row))
+                      get_simple_table_printable_row(row),
+                      terse)
             n += 1
 
     if "simple2" in idl.tables:
         simple2 = idl.tables["simple2"].rows
         for row in simple2.values():
             print_row("simple2", row, step,
-                      get_simple2_table_printable_row(row))
+                      get_simple2_table_printable_row(row),
+                      terse)
             n += 1
 
     if "simple3" in idl.tables:
         simple3 = idl.tables["simple3"].rows
         for row in simple3.values():
             print_row("simple3", row, step,
-                      get_simple3_table_printable_row(row))
+                      get_simple3_table_printable_row(row),
+                      terse)
             n += 1
 
     if "simple4" in idl.tables:
         simple4 = idl.tables["simple4"].rows
         for row in simple4.values():
             print_row("simple4", row, step,
-                      get_simple4_table_printable_row(row))
+                      get_simple4_table_printable_row(row),
+                      terse)
             n += 1
 
     if "simple5" in idl.tables:
         simple5 = idl.tables["simple5"].rows
         for row in simple5.values():
             print_row("simple5", row, step,
-                      get_simple5_table_printable_row(row))
+                      get_simple5_table_printable_row(row),
+                      terse)
             n += 1
 
     if "simple6" in idl.tables:
         simple6 = idl.tables["simple6"].rows
         for row in simple6.values():
             print_row("simple6", row, step,
-                      get_simple6_table_printable_row(row))
+                      get_simple6_table_printable_row(row),
+                      terse)
             n += 1
 
     if "link1" in idl.tables:
         l1 = idl.tables["link1"].rows
         for row in l1.values():
             print_row("link1", row, step,
-                      get_link1_table_printable_row(row))
+                      get_link1_table_printable_row(row),
+                      terse)
             n += 1
 
     if "link2" in idl.tables:
         l2 = idl.tables["link2"].rows
         for row in l2.values():
             print_row("link2", row, step,
-                      get_link2_table_printable_row(row))
+                      get_link2_table_printable_row(row),
+                      terse)
             n += 1
 
     if "singleton" in idl.tables:
         sng = idl.tables["singleton"].rows
         for row in sng.values():
             print_row("singleton", row, step,
-                      get_singleton_table_printable_row(row))
+                      get_singleton_table_printable_row(row),
+                      terse)
             n += 1
 
     if not n:
@@ -313,9 +325,9 @@ def substitute_uuids(json, symtab):
         symbol = symtab.get(json)
         if symbol:
             return str(symbol)
-    elif type(json) == list:
+    elif type(json) is list:
         return [substitute_uuids(element, symtab) for element in json]
-    elif type(json) == dict:
+    elif type(json) is dict:
         d = {}
         for key, value in json.items():
             d[key] = substitute_uuids(value, symtab)
@@ -329,10 +341,10 @@ def parse_uuids(json, symtab):
         name = "#%d#" % len(symtab)
         sys.stderr.write("%s = %s\n" % (name, json))
         symtab[name] = json
-    elif type(json) == list:
+    elif type(json) is list:
         for element in json:
             parse_uuids(element, symtab)
-    elif type(json) == dict:
+    elif type(json) is dict:
         for value in json.values():
             parse_uuids(value, symtab)
 
@@ -604,18 +616,32 @@ def idl_set(idl, commands, step):
     sys.stdout.flush()
 
 
-def update_condition(idl, commands):
+def update_condition(idl, commands, step):
+    next_cond_seqno = 0
     commands = commands[len("condition "):].split(";")
     for command in commands:
         command = command.split(" ")
-        if(len(command) != 2):
+        if len(command) != 2:
             sys.stderr.write("Error parsing condition %s\n" % command)
             sys.exit(1)
 
         table = command[0]
         cond = ovs.json.from_string(command[1])
 
-        idl.cond_change(table, cond)
+        next_seqno = idl.cond_change(table, cond)
+        if idl.cond_seqno == next_seqno:
+            sys.stdout.write("%03d: %s: conditions unchanged\n" %
+                             (step, table))
+        else:
+            sys.stdout.write("%03d: %s: change conditions\n" %
+                             (step, table))
+        sys.stdout.flush()
+
+        assert next_seqno == idl.cond_change(table, cond), \
+            "condition expected seqno changed"
+        next_cond_seqno = max(next_cond_seqno, next_seqno)
+
+    return next_cond_seqno
 
 
 def do_idl(schema_file, remote, *commands):
@@ -651,7 +677,7 @@ def do_idl(schema_file, remote, *commands):
         commands = commands[1:]
     else:
         schema_helper.register_all()
-    idl = ovs.db.idl.Idl(remote, schema_helper)
+    idl = ovs.db.idl.Idl(remote, schema_helper, leader_only=False)
     if "simple3" in idl.tables:
         idl.index_create("simple3", "simple3_by_name")
 
@@ -672,6 +698,7 @@ def do_idl(schema_file, remote, *commands):
     else:
         rpc = None
 
+    next_cond_seqno = 0
     symtab = {}
     seqno = 0
     step = 0
@@ -695,27 +722,48 @@ def do_idl(schema_file, remote, *commands):
 
     commands = list(commands)
     if len(commands) >= 1 and "condition" in commands[0]:
-        update_condition(idl, commands.pop(0))
-        sys.stdout.write("%03d: change conditions\n" % step)
-        sys.stdout.flush()
+        next_cond_seqno = update_condition(idl, commands.pop(0), step)
         step += 1
 
     for command in commands:
+        terse = False
+        if command.startswith("?"):
+            # We're only interested in terse table contents.
+            terse = True
+            command = command[1:]
+
         if command.startswith("+"):
             # The previous transaction didn't change anything.
             command = command[1:]
-        else:
-            # Wait for update.
-            while idl.change_seqno == seqno and not idl.run():
+        elif command.startswith("^"):
+            # Wait for condition change to be acked by the server.
+            command = command[1:]
+            while idl.cond_seqno != next_cond_seqno and not idl.run():
                 rpc.run()
 
                 poller = ovs.poller.Poller()
                 idl.wait(poller)
                 rpc.wait(poller)
                 poller.block()
+        else:
+            # Wait for update.
+            while True:
+                while idl.change_seqno == seqno and not idl.run():
+                    rpc.run()
 
-            print_idl(idl, step)
-            step += 1
+                    poller = ovs.poller.Poller()
+                    idl.wait(poller)
+                    rpc.wait(poller)
+                    poller.block()
+
+                print_idl(idl, step, terse)
+                step += 1
+
+                # Run IDL forever in case of a simple monitor, otherwise
+                # break and execute the command.
+                seqno = idl.change_seqno
+                if command != "monitor":
+                    break
 
         seqno = idl.change_seqno
 
@@ -725,9 +773,7 @@ def do_idl(schema_file, remote, *commands):
             step += 1
             idl.force_reconnect()
         elif "condition" in command:
-            update_condition(idl, command)
-            sys.stdout.write("%03d: change conditions\n" % step)
-            sys.stdout.flush()
+            next_cond_seqno = update_condition(idl, command, step)
             step += 1
         elif not command.startswith("["):
             idl_set(idl, command, step)
@@ -994,14 +1040,14 @@ def main(argv):
         sys.exit(1)
 
     func, n_args = commands[command_name]
-    if type(n_args) == tuple:
+    if type(n_args) is tuple:
         if len(args) < n_args[0]:
             sys.stderr.write("%s: \"%s\" requires at least %d arguments but "
                              "only %d provided\n"
                              % (ovs.util.PROGRAM_NAME, command_name,
                                 n_args[0], len(args)))
             sys.exit(1)
-    elif type(n_args) == int:
+    elif type(n_args) is int:
         if len(args) != n_args:
             sys.stderr.write("%s: \"%s\" requires %d arguments but %d "
                              "provided\n"
