@@ -34,11 +34,9 @@
 #include "dummy.h"
 #include "fatal-signal.h"
 #include "memory.h"
-#include "metrics.h"
 #include "netdev.h"
 #include "openflow/openflow.h"
 #include "ovsdb-idl.h"
-#include "ovs-doca.h"
 #include "ovs-rcu.h"
 #include "ovs-router.h"
 #include "ovs-thread.h"
@@ -112,8 +110,6 @@ main(int argc, char *argv[])
     unixctl_command_register("exit", "[--cleanup]", 0, 1,
                              ovs_vswitchd_exit, NULL);
 
-    metrics_unixctl_register("ovs_vswitchd");
-
     bridge_init(remote);
     free(remote);
 
@@ -212,7 +208,6 @@ parse_options(int argc, char *argv[], char **unixctl_pathp)
         case 'V':
             ovs_print_version(0, 0);
             print_dpdk_version();
-            print_doca_version();
             exit(EXIT_SUCCESS);
 
         case OPT_MLOCKALL:
